/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323
 
    Comments: Swyx H.323 specific stuff 

  --------------------------------------------------------------------------

   $Archive:: /H323/include/h323lib.h                                      $
    $Author:: Findeisen                                                    $
      $Date:: 7.05.02 16:34                                                $
   $Modtime:: 7.05.02 16:28                                                $
  $Revision:: 10                                                           $
  $Workfile:: h323lib.h                                                    $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef H323LIB_H
#define H323LIB_H


/** Initialise H323Lib.
*/
extern void H323LibInit(bool bUsingSwtcl = false);

/** Exit H323Lib.
*/
extern void H323LibExit();


/// Module including Swyx H.323 specific stuff for e.g. tracing
class H323Module
{
  public:
    /**
      Constructor.
      
      @param bUsingSwtcl  Boolean indicating usage of Swyx Telephony Class Library
    */
    H323Module(bool bUsingSwtcl = false);
    virtual ~H323Module();
    
    /// Initialize the tracing for all trace modules
    virtual void InitTracing();

    /// Destroy the tracing for all trace modules
    virtual void DestroyTracing();

  /**@name Trace module IDs */
  //@{
    /// Codec trace module ID
    unsigned m_CodecTraceId;

    /// H.225 client trace module ID
    unsigned m_H225TraceId;

    /// H.225 gatekeeper trace module ID
    unsigned m_GkTraceId;

    /// H.245 trace module ID  
    unsigned m_H245TraceId;
    
    /// H.323 trace module ID
    unsigned m_H323TraceId;
    
    /// RAS trace module ID
    unsigned m_RasTraceId;      

    /// RTP trace module ID    
    unsigned m_RtpTraceId;
  //@}

    /// Whether using Swyx Telephony Class Library or not
    bool m_bUsingSwtcl;
};

#endif // H323LIB_H
