/*---------------------------------------------------------------------------

   Swyx Solutions GmbH
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Solutions GmbH are
  Copyright (C) 2000-2005 Swyx Solutions GmbH.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323 PDUs
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/include/h323pdu.h                 $
    $Author:: Martin.hueser                                                $
      $Date:: 20.12.05 18:34                                               $
   $Modtime:: 16.12.05 16:35                                               $
  $Revision:: 4                                                            $
  $Workfile:: h323pdu.h                                                    $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __H323PDU_H
#define __H323PDU_H

#include <stlib.h>
#include "h323.h"
#include "q931.h"
#include "h225.h"
#include "h245.h"


/**Wrapper class for the H323 signalling channel.
 */
class H323SignalPDU : public H225_H323_UserInformation
{
  public:
    H323SignalPDU();

    H225_Facility_UUIE * BuildFacility(
      const H323Connection & connection,
      bool empty
    );
    H225_Facility_UUIE * BuildFacility(
      const H323Connection & connection,
      bool empty,
      const Tools::SByteVector & facilityIE
    );
    H225_CallProceeding_UUIE & BuildCallProceeding(
      const H323Connection & connection,
      int callRef,
      const H323GloballyUniqueID & callIdentifier
    );
    H225_CallProceeding_UUIE & BuildCallProceeding(
      const H323Connection & connection,
      int callRef,
      const H323GloballyUniqueID & callIdentifier,
      const Tools::SIPSocket::Address & h245Address,
      WORD port
    );
    H225_Alerting_UUIE & BuildAlerting(
      const H323Connection & connection,
      int callRef,
      const H323GloballyUniqueID & callIdentifier,
      const Q931::ProgressIndicator & ProgressInd
    );
    H225_Alerting_UUIE & BuildAlerting(
      const H323Connection & connection,
      int callRef,
      const H323GloballyUniqueID & callIdentifier,
      const Q931::ProgressIndicator & ProgressInd,
      const Tools::SIPSocket::Address & h245Address,
      WORD port
    );
    H225_ReleaseComplete_UUIE & BuildReleaseComplete(
      const H323Connection & connection,
      const H225_ReleaseCompleteReason & reason,
      const Q931::Cause &cause
    );
    H225_Connect_UUIE & BuildConnect(
      const H323Connection & connection,
      int callRef,
      const H323GloballyUniqueID & callIdentifier,
      const H323GloballyUniqueID & conferenceId
    );
    H225_Connect_UUIE & BuildConnect(
      const H323Connection & connection,
      int callRef,
      const H323GloballyUniqueID & callIdentifier,
      const H323GloballyUniqueID & conferenceId,
      const Tools::SIPSocket::Address & h245Address,
      WORD port
    );
    H225_Setup_UUIE & BuildSetup(
      const H323Connection & connection,
      int callRef,
      const Tools::SString & origAlias,
      const Tools::SString & callingPartyNumber,
      const Tools::SString & calledPartyNumber,
      const Tools::SString & destAlias,
      const Tools::SString & srcAlias,
      bool canOverlapSend = false,
      bool bSuppressNumber = false,
      bool bSendingComplete = false
    );
    H225_Setup_UUIE & BuildSetup(
      const H323Connection & connection,
      int callRef,
      const Tools::SString & origAlias,
      const Tools::SString & destAlias,
      const Tools::SIPSocket::Address & destAddr,
      WORD port
    );

    void BuildSetupAck(int callRef);

    H225_Progress_UUIE & BuildProgress(const H323Connection          & connection,
                                       const Q931::ProgressIndicator & ProgressInd,
                                       const Q931::Cause             & cause);

    void BuildNotify(const H323Connection              & connection,
                     const Q931::NotificationIndicator & NotificationInd);

    H225_Information_UUIE & BuildInformation(
      int callRef,
      bool fromDest,
      const Tools::SString & calledPartyNumber,
      const H323GloballyUniqueID & callIdentifier
    );

    void BuildStatus(int callRef, bool fromDest);

    void SetDisplayName(const Tools::SString & name) { m_DisplayName = name; }

    void PrintOn(std::ostream & strm) const;
    bool Read(H323Transport & transport);
    bool Read(H323Transport & transport, SByteVector& rawData);
    bool Write(H323Transport & transport);

    const Q931 & GetQ931() const { return q931pdu; }

  protected:
    void BuildQ931();

    // Even though we generally deal with the H323 protocol (H225) it is
    // actually contained within a field of the Q931 protocol.
    Q931 q931pdu;

    Tools::SString m_DisplayName;
};


/**Wrapper class for the H323 control channel.
 */
class H323ControlPDU : public H245_MultimediaSystemControlMessage
{
  public:
    H245_RequestMessage    & Build(H245_RequestMessage   ::Choices request);
    H245_ResponseMessage   & Build(H245_ResponseMessage  ::Choices response);
    H245_CommandMessage    & Build(H245_CommandMessage   ::Choices command);
    H245_IndicationMessage & Build(H245_IndicationMessage::Choices indication);

    H245_MasterSlaveDetermination & BuildMasterSlaveDetermination(
      unsigned terminalType,
      unsigned statusDeterminationNumber
    );
    H245_MasterSlaveDeterminationAck & BuildMasterSlaveDeterminationAck(
      bool isMaster
    );
    H245_MasterSlaveDeterminationReject & BuildMasterSlaveDeterminationReject(
      unsigned cause
    );

    H245_TerminalCapabilitySet & BuildTerminalCapabilitySet(
      const H323Connection & connection,
      unsigned sequenceNumber
    );
    H245_TerminalCapabilitySet & BuildEmptyTerminalCapabilitySet(
      unsigned sequenceNumber
    );
    H245_TerminalCapabilitySetAck & BuildTerminalCapabilitySetAck(
      unsigned sequenceNumber
    );
    H245_TerminalCapabilitySetReject & BuildTerminalCapabilitySetReject(
      unsigned sequenceNumber,
      unsigned cause
    );

    H245_OpenLogicalChannel & BuildOpenLogicalChannel(
      unsigned forwardLogicalChannelNumber
    );
    H245_RequestChannelClose & BuildRequestChannelClose(
      unsigned channelNumber,
      unsigned reason
    );
    H245_CloseLogicalChannel & BuildCloseLogicalChannel(
      unsigned channelNumber
    );
    H245_OpenLogicalChannelAck & BuildOpenLogicalChannelAck(
      unsigned channelNumber
    );
    H245_OpenLogicalChannelReject & BuildOpenLogicalChannelReject(
      unsigned channelNumber,
      unsigned cause
    );
    H245_OpenLogicalChannelConfirm & BuildOpenLogicalChannelConfirm(
      unsigned channelNumber
    );
    H245_CloseLogicalChannelAck & BuildCloseLogicalChannelAck(
      unsigned channelNumber
    );
    H245_RequestChannelCloseAck & BuildRequestChannelCloseAck(
      unsigned channelNumber
    );
    H245_RequestChannelCloseReject & BuildRequestChannelCloseReject(
      unsigned channelNumber
    );
    H245_RequestChannelCloseRelease & BuildRequestChannelCloseRelease(
      unsigned channelNumber
    );

    H245_RequestMode & BuildRequestMode(
      unsigned sequenceNumber
    );
    H245_RequestModeAck & BuildRequestModeAck(
      unsigned sequenceNumber,
      unsigned response
    );
    H245_RequestModeReject & BuildRequestModeReject(
      unsigned sequenceNumber,
      unsigned cause
    );

    H245_RoundTripDelayRequest & BuildRoundTripDelayRequest(
      unsigned sequenceNumber
    );
    H245_RoundTripDelayResponse & BuildRoundTripDelayResponse(
      unsigned sequenceNumber
    );

    H245_UserInputIndication & BuildUserInputIndication(
      const Tools::SString & value
    );

    H245_UserInputIndication & BuildUserInputIndication(
      const Tools::SString & tone,    /// DTMF tone code
      unsigned duration,              /// Duration of tone in milliseconds
      unsigned logicalChannel,        /// Logical channel number for RTP sync.
      unsigned rtpTimestamp           /// RTP timestamp in logical channel sync.
    );

    H245_FunctionNotUnderstood & BuildFunctionNotUnderstood(
      const H323ControlPDU & pdu
    );

    H245_EndSessionCommand & BuildEndSessionCommand(
      unsigned reason
    );

    static bool GetMikey(const H245_TerminalCapabilitySet & termCapSet, SByteVector & rMikey);
};


class H323RasPDU : public H225_RasMessage
{
  public:
    H225_GatekeeperRequest      & BuildGatekeeperRequest(unsigned seqNum);
    H225_GatekeeperConfirm      & BuildGatekeeperConfirm(unsigned seqNum);
    H225_GatekeeperReject       & BuildGatekeeperReject(unsigned seqNum,
                                                        unsigned reason = H225_GatekeeperRejectReason::e_undefinedReason);
    H225_RegistrationRequest    & BuildRegistrationRequest(unsigned seqNum);
    H225_RegistrationConfirm    & BuildRegistrationConfirm(unsigned seqNum);
    H225_RegistrationReject     & BuildRegistrationReject(unsigned seqNum,
                                                          unsigned reason);
    H225_UnregistrationRequest  & BuildUnregistrationRequest(unsigned seqNum);
    H225_UnregistrationConfirm  & BuildUnregistrationConfirm(unsigned seqNum);
    H225_UnregistrationReject   & BuildUnregistrationReject(unsigned seqNum,
                                                            unsigned reason);
    H225_LocationRequest        & BuildLocationRequest(unsigned seqNum);
    H225_LocationConfirm        & BuildLocationConfirm(unsigned seqNum);
    H225_LocationReject         & BuildLocationReject(unsigned seqNum,
                                                      unsigned reason);
    H225_AdmissionRequest       & BuildAdmissionRequest(unsigned seqNum);
    H225_AdmissionConfirm       & BuildAdmissionConfirm(unsigned seqNum,
                                                        unsigned bandWidth);
    H225_AdmissionReject        & BuildAdmissionReject(unsigned seqNum,
                                                       unsigned reason);
    H225_DisengageRequest       & BuildDisengageRequest(unsigned seqNum);
    H225_DisengageConfirm       & BuildDisengageConfirm(unsigned seqNum);
    H225_DisengageReject        & BuildDisengageReject(unsigned seqNum,
                                                       unsigned reason);
    H225_BandwidthRequest       & BuildBandwidthRequest(unsigned seqNum);
    H225_BandwidthConfirm       & BuildBandwidthConfirm(unsigned seqNum, unsigned bandWidth);
    H225_BandwidthReject        & BuildBandwidthReject(unsigned seqNum, unsigned reason);
    H225_InfoRequestResponse    & BuildInfoRequestResponse(unsigned seqNum);
    H225_UnknownMessageResponse & BuildUnknownMessageResponse(unsigned seqNum);

    enum H323RasPDUError
    {
        /// Successful completion
        Success,
        /// Failure in transport object
        TransportFailure,
        /// Failure in decode of ASN stream
        DecodeFailure,
        /// Failure in encode of ASN stream
        EncodeFailure,
        /// Failure in transport object which can be ignored
        DiscardLastAction,
        MaxErr
    };

    H323RasPDUError Read(H323Transport & transport);
    H323RasPDUError Write(H323Transport & transport) const;
};


void H323SetAliasAddress(const Tools::SString & name, H225_AliasAddress & alias);
Tools::SString H323GetAliasAddressString(const H225_AliasAddress & alias);

Tools::SString H323GetAliasAddressString(const H225_AliasAddress & alias,
                                         H225_AliasAddress::Choices tag);

void H323GetAliasAddressStringVector(
    const H225_ArrayOf_AliasAddress & rAliasArray,
    H225_AliasAddress::Choices tag,
    Tools::SStringVector & rAliases);

H225_TransportAddress H323GetTransportAddress(
    const H225_ArrayOf_TransportAddress & AddressArray,
    H225_TransportAddress::Choices tag);

#endif // __H323PDU_H
