/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Gatekeeper Server
 
    Comments: Registration of endpoints at the gatekeeper

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/GkRegistration.cpp         $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 6.01.06 11:15                                                $
   $Modtime:: 6.01.06 11:13                                                $
  $Revision:: 14                                                           $
  $Workfile:: GkRegistration.cpp                                           $
$NoKeywords: $
----------------------------------------------------------------------------*/

#include <stlib.h>

#include "h323.h"
#include "h323pdu.h"
#include "gkserver.h"
#include "GkRegistration.h"

using namespace std;
using namespace Tools;

//-------------------------------------------------------------------

H323GkAliases::H323GkAliases(const Tools::SStringVector & rE164Aliases,
                             const Tools::SStringVector & rH323Aliases) :
    m_E164Aliases(rE164Aliases), 
    m_H323Aliases(rH323Aliases)
{
}

//-------------------------------------------------------------------

void H323GkAliases::PrintOn(ostream & rStream) const
{
    rStream << "E164Aliases: ";
    SStringVector::const_iterator Iter;
    for (Iter = m_E164Aliases.begin(); Iter != m_E164Aliases.end(); ++Iter)
        rStream << "'" << (*Iter).c_str() << "', ";
    rStream << " H323Aliases: ";
    for (Iter = m_H323Aliases.begin(); Iter != m_H323Aliases.end(); ++Iter)
        rStream << "'" << (*Iter).c_str() << "', ";
    rStream << " reg. H323Alias: '" << m_RegisteredH323Alias << "'";
    rStream << " reg. E164Aliases: ";
    for (Iter = m_RegisteredE164Aliases.begin(); Iter != m_RegisteredE164Aliases.end(); ++Iter)
        rStream << "'" << (*Iter).c_str() << "', ";
    rStream << " reg. Mixed Aliases: ";
    for (Iter = m_RegisteredMixedAliases.begin(); Iter != m_RegisteredMixedAliases.end(); ++Iter)
        rStream << "'" << (*Iter).c_str() << "', ";
}


/* ----------------------------------------------------------------------- */
/* lessthanH323RemoteEndpoint - determins correct timer list order         */
/* ----------------------------------------------------------------------- */
SINLINE bool lessthanH323RemoteEndpoint::operator()(const H323RemoteEndpoint* ep1, const H323RemoteEndpoint* ep2) const
{
	return (ep1->m_TimeToLive < ep2->m_TimeToLive) || 
            (ep1->m_TimeToLive == ep2->m_TimeToLive && ep1 < ep2 );
}


//===================================================================

// we're storing the this pointer only. Therefore the warning can be ignored
#pragma warning(disable:4355)
H323RemoteEndpoint::H323RemoteEndpoint(
    H323GkServer &rGKServer,
    const Tools::SIpAddress & rIpAddress,
    const H323TransportAddress & rCallSignalAddress,
    const H323TransportAddress & rRasAddress,
    const Tools::SStringVector & rE164Aliases,
    const Tools::SStringVector & rH323Aliases,
    Tools::STimeInterval TimeToLive,
    const H323EndPoint::TerminalTypes & rTerminalType,
    const Tools::SWString &rszVersion,
    int   iSupportedCodecs,
    bool  bRequiresDelayedDisconnect,
	int   iLineId,
	int   iNumOfChannels ) :
        m_rGKServer(rGKServer),
        m_IpAddress(rIpAddress),
        m_CallSignalAddress(rCallSignalAddress),
        m_RasAddress(rRasAddress),
        m_H323GkAliases(rE164Aliases, rH323Aliases),
        m_TerminalType(rTerminalType),
        m_szVersion(rszVersion),
	    m_iLineId(iLineId),
	    m_iNumOfChannels(iNumOfChannels),
        m_TimeToLive(0),
        m_uTimeToLiveInSeconds(0),
        m_RegistrationStatus(RegistrationTimerStopped),
        m_bRequiresDelayedDisconnect(bRequiresDelayedDisconnect),
        m_iSupportedCodecs(iSupportedCodecs),
        m_uSequenceNumber(0),
        m_uRequestRetryNumber(0),
        m_pUserData(0)
#ifdef _DEBUG
      , m_ProcessingSync(1,1,SString(SString::Printf, "H323RemoteEndpoint::ProcessingSync[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#else
      , m_ProcessingSync(1,1)
#endif
#pragma warning(default:4355)
{
    static DWORD UniqueId = 0;
    
    SDEFFUNC("H323RemoteEP::H323RemoteEP");

    // set absolute time of inital registration for this EP
    m_TimeOfRegistration = Tools::STime();

    // is registration time to live supported by this EP?
    if ( TimeToLive > 0 )
    {
        // set initial tick value for registration timeout check!
        m_TimeToLive = STimer::Tick() + TimeToLive + STimeInterval(0, 10);
        // set time to live in seconds for RCF
        m_uTimeToLiveInSeconds = TimeToLive.GetSeconds();
        // set absolute time when registration expires
        m_TimeOfRegistrationExpiration = m_TimeOfRegistration + TimeToLive + STimeInterval(0, 10);
    }

    // override m_IpAddress if we have a valid m_CallSignalAddress!
    SIPSocket::Address Addr;
    WORD Port;
    m_CallSignalAddress.GetIpAndPort(Addr, Port);
    if ( Port )
    {
        m_IpAddress.m_Address = Addr;
        m_IpAddress.m_Port    = Port;
    }

    m_EndpointIdentifier = SProcess::Current().GetManufacturer() + SString(SString::Unsigned, ++UniqueId);

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "(" << m_EndpointIdentifier << ", " << m_IpAddress << ")" );
}

//-------------------------------------------------------------------

H323RemoteEndpoint::~H323RemoteEndpoint()
{
    SDEFFUNC("H323RemoteEP::~H323RemoteEP");

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "(" << m_EndpointIdentifier << ")" );
}

//-------------------------------------------------------------------

void H323RemoteEndpoint::PrintOn(ostream & rStream) const
{
    static Tools::SString szCodecTxt[NoOfSupportedCodecs] =
    {
        "G711a",
        "G711u",
        "Speex",
        "comp",
        "G729",
        "GSM",
        "T38"
    };

    try
    {
        rStream << m_EndpointIdentifier << "\n" << STrace::MultiLinePrefix;
        rStream << "  Time of registration      " << m_TimeOfRegistration << "\n" << STrace::MultiLinePrefix;
        rStream << "  TimeToLive in seconds:    " << m_uTimeToLiveInSeconds << "\n" << STrace::MultiLinePrefix;
        rStream << "  Registration expires:     " << m_TimeOfRegistrationExpiration << "\n" << STrace::MultiLinePrefix;
        if ( m_RegistrationStatus == H323RemoteEndpoint::RegistrationTimerStarted )
            rStream << "  Registration status:      active" << "\n" << STrace::MultiLinePrefix;
        else if ( m_RegistrationStatus == H323RemoteEndpoint::RegistrationTimerStopped )
            rStream << "  Registration status:      UNREGISTERED" << "\n" << STrace::MultiLinePrefix;
        else
            rStream << "  Registration status:      EXPIRED" << "\n" << STrace::MultiLinePrefix;
        rStream << "  IpAddress:                " << m_IpAddress << "\n" << STrace::MultiLinePrefix;
        rStream << "  CallSignalAddress:        " << m_CallSignalAddress << "\n" << STrace::MultiLinePrefix;
        rStream << "  RasAddress:               " << m_RasAddress << "\n" << STrace::MultiLinePrefix;
        if ( m_iLineId > 0 )
            rStream << "  Line Id:                  " << m_iLineId;
        else if ( m_iNumOfChannels > 0 )
            rStream << "  Number of fax channels:   " << m_iNumOfChannels;
        rStream << "\n" << STrace::MultiLinePrefix;
        rStream << "  RRQ H.323 Aliases:        ";
        SStringVector::const_iterator Iter;
        for (Iter = m_H323GkAliases.m_H323Aliases.begin(); Iter != m_H323GkAliases.m_H323Aliases.end(); ++Iter)
            rStream << "'" << (*Iter).c_str() << "' ";
        rStream << "\n" << STrace::MultiLinePrefix;
        rStream << "  RRQ E.164 Aliases:        ";
        for (Iter = m_H323GkAliases.m_E164Aliases.begin(); Iter != m_H323GkAliases.m_E164Aliases.end(); ++Iter)
            rStream << "'" << (*Iter).c_str() << "' ";
        rStream << "\n" << STrace::MultiLinePrefix;
        rStream << "  Registered H323Alias:     '" << m_H323GkAliases.m_RegisteredH323Alias << "'" << "\n" << STrace::MultiLinePrefix;
        rStream << "  Registered E164Aliases:   ";
        for (Iter = m_H323GkAliases.m_RegisteredE164Aliases.begin(); Iter != m_H323GkAliases.m_RegisteredE164Aliases.end(); ++Iter)
            rStream << "'" << (*Iter).c_str() << "' ";
        rStream << "\n" << STrace::MultiLinePrefix;
        rStream << "  Registered Mixed Aliases: ";
        for (Iter = m_H323GkAliases.m_RegisteredMixedAliases.begin(); Iter != m_H323GkAliases.m_RegisteredMixedAliases.end(); ++Iter)
            rStream << "'" << (*Iter).c_str() << "' ";
        rStream << "\n" << STrace::MultiLinePrefix;
        rStream << "  TerminalType:             " << m_TerminalType << "\n" << STrace::MultiLinePrefix;
        rStream << "  Supported codes           ";
        int iIdx = 0;
        for (int iLoop = 1; iLoop < H323RemoteEndpoint::CodecAll; iLoop = iLoop << 1)
        {
            if (iLoop & m_iSupportedCodecs)
                rStream << szCodecTxt[iIdx] << ", ";
            ++iIdx;
        }
        rStream << "\n" << STrace::MultiLinePrefix;

    }
    catch(...)
    {
        rStream << "### Exception thrown on printing of remote endpoint ###";
    }
}

//-------------------------------------------------------------------

bool H323RemoteEndpoint::FindMixedAlias(const Tools::SString & rAlias)
{
    SDEFFUNC("H323RemoteEP::FindMixedAlias");

    Tools::SString szLowerAlias = rAlias.ToLower();

    SStringVector::iterator it;
    for (it = m_H323GkAliases.m_RegisteredMixedAliases.begin();
         it != m_H323GkAliases.m_RegisteredMixedAliases.end(); ++it)
    {
        if ((*it).ToLower() == szLowerAlias)
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() Found mixed alias '" << rAlias << "'");
            return true;
        }
    }
    return false;
}

//-------------------------------------------------------------------

// Notification function called by SInternalTimerlist::Process() call the notification function
void H323RemoteEndpoint::OnRegistrationExpired( Tools::SThreadPool&, int )
{ 
    SDEFFUNC("H323RemoteEP::OnRegExpired");

    H323RegisteredEPList &rH323RegisteredEPList =  m_rGKServer.GetRegisteredEndpoints();
    SMutex& rRegisteredEPListMutex = rH323RegisteredEPList.GetRegTimerListMutex();

    bool bDeleteEndpoint = rH323RegisteredEPList.HandleRasRegistrationTimeout( this );

    rRegisteredEPListMutex.Wait();

    if( bDeleteEndpoint )
    {
    	// timer marked to be deleted. We remove it from timerlist and delete it
    	STRACE2(STrace::Debug, g_pH323Module->m_GkTraceId, this, 0, "() Registration timer will be deleted");

        if( m_RegistrationStatus == RegistrationTimerStarted )
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0, "() Registration timer must be stopped, when ToBeDeleted flag is set -> timer will be deleted");

            // if the timer is still running, i.e. it is still in the list - remove it
            rH323RegisteredEPList.RemoveEPFromRegTimerList( this, false );
            m_RegistrationStatus = RegistrationTimerStopped;
        }
    }
    m_ProcessingSync.Signal();  // remark: the value is set in a proteted area (by TimerListMutex)

    rRegisteredEPListMutex.Signal();

    DecRef(__FILE__, __LINE__); // remove reference for work item -> EP may be deleted NOW!
}


//===================================================================

H323RegisteredEPList::H323RegisteredEPList(H323GkServer & rGatekeeper,
                                   bool AutoDelete) :
        m_rGatekeeper(rGatekeeper),
        m_AutoDelete(AutoDelete),
        m_pRegTimeoutThreadPool(0)
{
    SDEFFUNC("H323RegEPList::H323RegEPList");

    m_pRegTimeoutThreadPool = new SThreadPool();
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "()");
}

//-------------------------------------------------------------------

H323RegisteredEPList::~H323RegisteredEPList()
{
    SDEFFUNC("H323RegEPList::~H323RegEPList");

    H323RemoteEndpointHashmap::iterator it;

    m_EndpointsCritSect.Wait();
    while ((it = m_Endpoints.begin()) != m_Endpoints.end())
    {
        H323RemoteEndpoint *pEndpoint = it->second;
        m_EndpointsCritSect.Signal();
        // release endpoint at tcl-gatekeeper
        if ( !SBAD_READPTR(pEndpoint) )
        {
            if ( pEndpoint->m_pUserData != 0 )
            {
                m_rGatekeeper.OnReleaseRemoteEndpoint(
                    pEndpoint->m_IpAddress,
                    pEndpoint->m_H323GkAliases,
                    pEndpoint->m_pUserData,
                    pEndpoint->m_TerminalType);
            }
            // remove endpoint from registration timeout list
            RemoveEPFromRegTimerList(pEndpoint);
            // remove endpoint reference from timer list
            pEndpoint->DecRef(__FILE__, __LINE__);
            pEndpoint = 0;
        }
        m_EndpointsCritSect.Wait();
        m_Endpoints.erase(it);
    }
    m_EndpointsCritSect.Signal();

    // in case of something went wrong with registration timer list handling...
    if ( m_EndpointsRegTimerList.size() ) 
    {
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0, "() there are "
                                << m_EndpointsRegTimerList.size() << " EP(s) still in registration timer list!");

        H323RemoteEndpointRegTimerList::iterator listPos = m_EndpointsRegTimerList.begin();
        while( listPos != m_EndpointsRegTimerList.end() )
        {
            // this stops the registrations timer and removes it from the EP registraion timer list.
            RemoveEPFromRegTimerList(*listPos, false);
            // reset iterator which is now invalid!
            listPos = m_EndpointsRegTimerList.begin();
        }
        m_EndpointsRegTimerList.clear();
    }
    
    // Finally destruct timer list thread pool
    if ( m_pRegTimeoutThreadPool )
    {
        delete m_pRegTimeoutThreadPool;
        m_pRegTimeoutThreadPool = 0;
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "()");
}

//-------------------------------------------------------------------

void H323RegisteredEPList::PrintOn(std::ostream & rStream) const
{
    rStream << "Registered Endpoints:" << "\n" << STrace::MultiLinePrefix;

    H323RemoteEndpointHashmap::const_iterator Iter;
    for (Iter = m_Endpoints.begin(); Iter != m_Endpoints.end(); ++Iter)
    {
        H323RemoteEndpoint *pEp = (*Iter).second;
        if (!SBAD_READPTR(pEp))
            rStream << *(pEp);
    }
}

//-------------------------------------------------------------------

H323GkSrv::Err H323RegisteredEPList::Insert(const Tools::SIpAddress & rIpAddress,
                                            H323RemoteEndpoint      * pEndpoint)
{
    SDEFFUNC("H323RegisteredEPList::Insert");

    H323GkSrv::Err err = H323GkSrv::Failure;

    if (SBAD_READPTR(pEndpoint))
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() endpoint invalid!");
        return H323GkSrv::Failure;
    }

    if (   pEndpoint->m_H323GkAliases.m_H323Aliases.empty()
        && pEndpoint->m_H323GkAliases.m_E164Aliases.empty() )
    {
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() No H.323 or E.164 aliases found for endpoint " << pEndpoint->m_IpAddress);
        return H323GkSrv::Failure;
    }

    // Try to register this endpoint at server
    err = m_rGatekeeper.OnRegisterRemoteEndpoint(rIpAddress,
                                                 pEndpoint->m_H323GkAliases,
                                                 &pEndpoint->m_pUserData,
                                                 pEndpoint->m_TerminalType,
                                                 pEndpoint->m_iSupportedCodecs,
                                                 pEndpoint->m_szVersion,
                                                 pEndpoint->m_iLineId,
                                                 pEndpoint->m_iNumOfChannels);
    if (err != H323GkSrv::Success)
    {
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() Registration of endpoint " << pEndpoint->m_IpAddress << " with aliases ("
            << pEndpoint->m_H323GkAliases
            << ") and endpointIdentifier '" << pEndpoint->m_EndpointIdentifier
            << "' to gatekeeper failed!");
        return err;
    }

    // check if endpoint identifier is already present in endpoint list and matches endpoint object
    m_EndpointsCritSect.Wait();
    H323RemoteEndpointHashmap::iterator pos = m_Endpoints.find( pEndpoint->m_EndpointIdentifier );
    if (  pos != m_Endpoints.end() 
         && (*pos).second == pEndpoint )
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() endpoint " << pEndpoint->m_IpAddress << " with alias '"
            << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias
            << "' and endpointIdentifier '" << pEndpoint->m_EndpointIdentifier
            << "' is already in hasmap!");
    }
    else
    {
        pair<H323RemoteEndpointHashmap::iterator, bool> Answer;
        Answer = m_Endpoints.insert(pair<const SString, H323RemoteEndpoint *>(
                                    pEndpoint->m_EndpointIdentifier, pEndpoint));
        if (Answer.second == false)
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Insertion of endpoint " << pEndpoint->m_IpAddress << " with alias '"
                << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias
                << "' and endpointIdentifier '" << pEndpoint->m_EndpointIdentifier
                << "' in hasmap failed!");
            m_EndpointsCritSect.Signal();
            return H323GkSrv::Failure;
        }
        pEndpoint->IncRef(__FILE__, __LINE__);
    }
    m_EndpointsCritSect.Signal();
    
    // time to live value included in RRQ?
    if ( pEndpoint->m_uTimeToLiveInSeconds > 0 )
    {
        // add endpoint to registration timeout list to follow up keep alive RRQ status
        AddEPToRegTimerList(pEndpoint);
    }
    
    return H323GkSrv::Success;
}

//-------------------------------------------------------------------

bool H323RegisteredEPList::Erase(const Tools::SString & rEndpointIdentifier)
{
    SDEFFUNC("H323RegisteredEPList::Erase");

    bool Answer = false;
    m_EndpointsCritSect.Wait();
    H323RemoteEndpointHashmap::iterator it = m_Endpoints.find(rEndpointIdentifier);
    if ( it != m_Endpoints.end() )
    {
        Answer = true;
        H323RemoteEndpoint *pEndpoint = it->second;
        m_Endpoints.erase(it);
        m_EndpointsCritSect.Signal();

        if (SBAD_READPTR(pEndpoint))
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() endpoint invalid!");
        }
        else
        {

            STRACE2_INFO2( g_pH323Module->m_GkTraceId, this, 0, "("
                << pEndpoint << ") remove EP from registration timer list");

            if ( pEndpoint->m_pUserData != 0 )
            {
                m_rGatekeeper.OnReleaseRemoteEndpoint(
                    pEndpoint->m_IpAddress,
                    pEndpoint->m_H323GkAliases,
                    pEndpoint->m_pUserData,
                    pEndpoint->m_TerminalType);
            }
            // remove endpoint from registration timeout list
            RemoveEPFromRegTimerList(pEndpoint);
            // remove endpoint reference from timer list
            pEndpoint->DecRef(__FILE__, __LINE__);
        }
    }

    return Answer;
}

//-------------------------------------------------------------------

bool H323RegisteredEPList::AddEPToRegTimerList(H323RemoteEndpoint * pEndpoint)
{
    SDEFFUNC("H323RegEPList::AddEPRegTimerList");

    // someone tricking us?
    if (SBAD_READPTR(pEndpoint))
        return false;

    m_EPRegTimerListMutex.Wait();

    pair<H323RemoteEndpointRegTimerList::iterator, bool> Answer;
    Answer = m_EndpointsRegTimerList.insert( pEndpoint );
    if ( Answer.second == true )   // true = inserted!
    {
        STRACE2_INFO2( g_pH323Module->m_GkTraceId, this, 0, "("
            << pEndpoint << ") inserted EP in registration timer list");
    
        // add reference to endpoint
        pEndpoint->IncRef(__FILE__, __LINE__);
        // update status if required
        if ( pEndpoint->m_RegistrationStatus == H323RemoteEndpoint::RegistrationTimerStopped )
            pEndpoint->m_RegistrationStatus = H323RemoteEndpoint::RegistrationTimerStarted; // State is set here due to the protection by #m_EPRegTimerListMutex#
    }
    else
    {
        STRACE2_ERROR( g_pH323Module->m_GkTraceId, this, 0, "("
            << pEndpoint << ") insertion of EP in registration timer list FAILED!");
    }

    m_EPRegTimerListMutex.Signal();

    if ( Answer.second == true )   // true = inserted!
    {
        // We have added a new EP registration timer, signal it to the GK server thread 
        m_rGatekeeper.SignalRegistrationTimerListChange();
    }

    return Answer.second;
}

//-------------------------------------------------------------------

bool H323RegisteredEPList::UpdateEPRegTimer(H323RemoteEndpoint * pEndpoint, Tools::STimeInterval TimeToLive)
{
    SDEFFUNC("H323RegEPList::UpdateEPRegTimer");

    // someone tricking us?
    if (SBAD_READPTR(pEndpoint))
        return false;

    m_EPRegTimerListMutex.Wait();

    pair<H323RemoteEndpointRegTimerList::iterator, bool> Answer;
    Answer.second = false;

    bool bErased = (m_EndpointsRegTimerList.erase( pEndpoint ) == 1);
    if ( bErased )
    {
        // set new tick count for registration expiry as current tick count plus registration TTL as of RRQ PLUS 10 sec reserve!
        pEndpoint->m_TimeToLive = STimer::Tick() + TimeToLive + STimeInterval(0, 10);
        // set absolute time when registration expires
        pEndpoint->m_TimeOfRegistrationExpiration = Tools::STime() + TimeToLive + STimeInterval(0, 10);
        
        // is endpoint registered to upper layer gatekeeper?
        if ( pEndpoint->m_RegistrationStatus == H323RemoteEndpoint::RegistrationTimerExpired )
        {
            // currently NOT! -> try to register this endpoint at server
            if ( m_rGatekeeper.OnRegisterRemoteEndpoint(pEndpoint->m_IpAddress,
                                                        pEndpoint->m_H323GkAliases,
                                                        &pEndpoint->m_pUserData,
                                                        pEndpoint->m_TerminalType,
                                                        pEndpoint->m_iSupportedCodecs,
                                                        pEndpoint->m_szVersion,
                                                        pEndpoint->m_iLineId,
                                                        pEndpoint->m_iNumOfChannels) != H323GkSrv::Success )
            {
                STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                    "() Registration of endpoint " << pEndpoint->m_IpAddress << " with aliases ("
                    << pEndpoint->m_H323GkAliases
                    << ") and endpointIdentifier '" << pEndpoint->m_EndpointIdentifier
                    << "' to gatekeeper failed!");

                m_EPRegTimerListMutex.Signal();
                return false;
            }
        }

        Answer = m_EndpointsRegTimerList.insert( pEndpoint );
        if ( Answer.second == true )   // true = inserted!
        {
            STRACE2_INFO2( g_pH323Module->m_GkTraceId, this, 0, "("
                << pEndpoint << ") updated EP in registration timer list");
            // update status because endpoint back to play :-)
            pEndpoint->m_RegistrationStatus = H323RemoteEndpoint::RegistrationTimerStarted;
        }
        else
        {
            STRACE2_ERROR( g_pH323Module->m_GkTraceId, this, 0, "("
                << pEndpoint << ") update of EP in registration timer list FAILED!");
        }
    }
    else
    {
        STRACE2_WARN( g_pH323Module->m_GkTraceId, this, 0, "("
            << pEndpoint << ") EP in registration timer list NOT found!");
    }

    m_EPRegTimerListMutex.Signal();

    if ( Answer.second == true )   // true = inserted!
    {
        // We have updated a EP registration timer, signal it to the GK server thread 
        m_rGatekeeper.SignalRegistrationTimerListChange();
    }

    return Answer.second;
}

//-------------------------------------------------------------------

bool H323RegisteredEPList::RemoveEPFromRegTimerList(H323RemoteEndpoint * pEndpoint, bool bSignalListChange )
{
    SDEFFUNC("H323RegEPList::RemEPRegTimerList");

    // someone tricking us?
    if (SBAD_READPTR(pEndpoint))
        return false;

    m_EPRegTimerListMutex.Wait();

    bool bErased = (m_EndpointsRegTimerList.erase( pEndpoint ) == 1);
    if ( bErased )
    {
        STRACE2_INFO2( g_pH323Module->m_GkTraceId, this, 0, "("
            << pEndpoint << ") removed EP from registration timer list");
    
        // remove endpoint reference (may be deleted NOW!!!)
        pEndpoint->DecRef(__FILE__, __LINE__);

        // We have removed a EP registration timer, signal it to the GK server thread 
        if ( bSignalListChange )
        {
             // State is set here due to the protection by #m_EPRegTimerListMutex#
            pEndpoint->m_RegistrationStatus = H323RemoteEndpoint::RegistrationTimerStopped;
            
            m_rGatekeeper.SignalRegistrationTimerListChange();
        }
    
        m_EPRegTimerListMutex.Signal();
    }
    else
    {
        m_EPRegTimerListMutex.Signal();

        STRACE2_WARN( g_pH323Module->m_GkTraceId, this, 0, "("
            << pEndpoint << ") NOT found in EP registration timer list");
    }

    return bErased;
}

//-------------------------------------------------------------------

bool H323RegisteredEPList::HandleRasRegistrationTimeout(H323RemoteEndpoint * pEndpoint )
{
    SDEFFUNC("H323RegEPList::HdlRasRegTimeout");
    
    STRACE2( STrace::Debug, g_pH323Module->m_GkTraceId, this, 0, "(" << pEndpoint << ")" );

    return m_rGatekeeper.HandleRasRegistrationTimeout( pEndpoint );
}

//-------------------------------------------------------------------

STimeInterval H323RegisteredEPList::ProcessRegTimerList()
{
    SDEFFUNC("H323RegEPList::ProcRegTimerList");

    STRACE2( STrace::Debug, g_pH323Module->m_GkTraceId, this, 0, "()" );

    // SInternal timer list is an ordered (multi)set. To find all registration timers
    // having timed out it sufficient to take all timers from the beginning until
    // we get a timer which is not timed out yet.
    // All the rest of the list will time out even later!

    m_EPRegTimerListMutex.Wait();
    
    H323RemoteEndpointRegTimerList::iterator listPos;
    for( listPos = m_EndpointsRegTimerList.begin(); 
         listPos!=m_EndpointsRegTimerList.end() && (*listPos)->m_TimeToLive <= STimer::Tick();
         /* iterator is changed within in the loop */ )
    {
        // CheckForRestart, inserts the timer into the list if the timer is continuously running. 
        // So the list will change and listPos gets invalid.
        H323RemoteEndpoint* pCurrentEndpoint = (*listPos);

        if ( !SBAD_READPTR(pCurrentEndpoint) )
        {
            STRACE2_ERROR( g_pH323Module->m_GkTraceId, this, 0, "("
                << pCurrentEndpoint->m_EndpointIdentifier << ") EP registration timed out");

            // skip all timers whose timeout function is still running
            if( pCurrentEndpoint->m_ProcessingSync.WillBlock() )
            {
                // timeout function of this timer is still in process
                // and we are going to start it once again.
                STRACE2(STrace::Fatal, g_pH323Module->m_GkTraceId, this, 0, 
                    "##### Timeout Congestion detected: H323RemoteEndpoint " 
		            << pCurrentEndpoint->m_EndpointIdentifier << " #####");

                listPos++;
            }
            else
            {
                // insert into the timeout list 
                STRACE2_INFO3( g_pH323Module->m_GkTraceId, this, 0, 
                    "() creating new work item, EP= " 
			        << pCurrentEndpoint);

                // add reference for work item
                pCurrentEndpoint->IncRef(__FILE__, __LINE__);

                // indicate to any other thread that the timer is calling the timeout function
                // signal is called when timeout function is finished in STimer::OnTimeout( , )
                pCurrentEndpoint->m_ProcessingSync.Wait();

                // remove timer from the list. If it must restarted it has to be reinserted again
                // with a new timeout time value
                RemoveEPFromRegTimerList(pCurrentEndpoint, false); // false= do not notify myself about timer list change :-)
            
                // release Mutex as long as we start the timeout functions
                m_EPRegTimerListMutex.Signal();

                // queue a timeout thread using the SThreadPool
                m_pRegTimeoutThreadPool->QueueWorkItem(SCREATE_NOTIFIER2(pCurrentEndpoint, H323RemoteEndpoint::OnRegistrationExpired),
                                                       SThreadPool::ExecLongFunction);

                m_EPRegTimerListMutex.Wait();

                // refresh the iterator after removing timer from the list
                listPos = m_EndpointsRegTimerList.begin();
            }
        }
        else
        {
            // INVALID endpoint!
            STRACE2_ERROR( g_pH323Module->m_GkTraceId, this, 0, 
                "##### Timeout on invalid H323RemoteEndpoint " << hex << pCurrentEndpoint << " #####");

            // remove timer from the list. If it must restarted it has to be reinserted again
            // with a new timeout time value
            RemoveEPFromRegTimerList(pCurrentEndpoint, false); // false= do not notify myself about timer list change :-)

            // refresh the iterator after removing timer from the list
            listPos = m_EndpointsRegTimerList.begin();
        }
    }

    
    STimeInterval minTimeLeft = SMaxTimeInterval;
  
    // determine the time the next timer will time out
    // listPos is now either the first timer which will time out or
    // the end of the list
    H323RemoteEndpointRegTimerList::iterator MinListPos = listPos;
    if( MinListPos != m_EndpointsRegTimerList.end() )
    {
        minTimeLeft = (*MinListPos)->m_TimeToLive - STimer::Tick();

        if( minTimeLeft < 0 )
        {
            minTimeLeft = 1000;
            STRACE2_WARN( g_pH323Module->m_GkTraceId, this, 0, "() negative timeout interval calculated!");
        }
    }

    m_EPRegTimerListMutex.Signal();

    return minTimeLeft;
}

//-------------------------------------------------------------------

H323RemoteEndpoint * H323RegisteredEPList::FindEndpointByTsap(const Tools::SIpAddress & rIpAddress)
{
    SDEFFUNC("H323RegEPList::FindEpByTsap");

    H323RemoteEndpointHashmap::iterator it;

    m_EndpointsCritSect.Wait();
    for (it = m_Endpoints.begin(); it != m_Endpoints.end(); ++it)
    {
        if (   (*it).second
            && ((*it).second->m_IpAddress == rIpAddress) )
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Found endpoint by TSAP " << rIpAddress);
            m_EndpointsCritSect.Signal();
            return (*it).second;
        }
    }
    m_EndpointsCritSect.Signal();

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Endpoint not found by TSAP " << rIpAddress);
    return 0;
}

//-------------------------------------------------------------------

H323RemoteEndpoint * H323RegisteredEPList::FindEndpointByAlias(const Tools::SString    & rAlias,
                                                               const Tools::SIpAddress & rIpAddress,
                                                               bool bMatchIpAndPort,
                                                               bool bSkipExpiredEPs)
{
    SDEFFUNC("H323RegEPList::FindEpByAlias");

    H323RemoteEndpoint * pMatchingEndpoint = 0;

    if ( bMatchIpAndPort )
    {
        STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0, "() Search endpoint by alias '"
            << rAlias << "' and " << rIpAddress
            << ", skip EXPIRED endpoints= " << (bSkipExpiredEPs ? "YES" : "no") );
    }
    else
    {
        STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,"() Search endpoint by alias '"
            << rAlias << "', skip EXPIRED endpoints= " << (bSkipExpiredEPs ? "YES" : "no") );
    }

    H323RemoteEndpointHashmap::iterator it;
    // 1st: find endpoint(s) with matching alias AND IP address AND port
    m_EndpointsCritSect.Wait();
    for (it = m_Endpoints.begin(); it != m_Endpoints.end(); ++it)
    {
        if ( !SBAD_READPTR((*it).second) )
        {
            if (   bSkipExpiredEPs == true
                && (*it).second->m_RegistrationStatus == H323RemoteEndpoint::RegistrationTimerExpired )
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Skip EXPIRED endpoint with id '"<< (*it).second->m_EndpointIdentifier << "'" );
                continue;
            }

            if (   (*it).second->m_H323GkAliases.m_RegisteredH323Alias.ToLower() == rAlias.ToLower()
                || (*it).second->FindMixedAlias(rAlias) == true )
            {
                if (   bMatchIpAndPort
                    && (*it).second->m_IpAddress == rIpAddress )
                {
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                        "() Found matching endpoint with alias '"<< rAlias << "' and " << (*it).second->m_IpAddress );
                    pMatchingEndpoint = (*it).second;
                    break;
                }
                else if ( !bMatchIpAndPort )
                {
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                        "() Found matching endpoint with alias '"<< rAlias << "'" );
                    pMatchingEndpoint = (*it).second;
                    break;
                }
            }
        }
    }
    m_EndpointsCritSect.Signal();

    if ( pMatchingEndpoint == 0 )
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() NO endpoint found!" );
    }
    return pMatchingEndpoint;
}

//-------------------------------------------------------------------

H323RemoteEndpointVector H323RegisteredEPList::FindEndpointsByAlias(const Tools::SString & rAlias)
{
    SDEFFUNC("H323RegEPList::FindEpsByAlias");

    H323RemoteEndpointVector matchingEPs;

    H323RemoteEndpointHashmap::iterator it;
    // find endpoint(s) with matching alias
    m_EndpointsCritSect.Wait();
    for (it = m_Endpoints.begin(); it != m_Endpoints.end(); ++it)
    {
        if ( !SBAD_READPTR((*it).second) )
        {
            if (   (*it).second->m_H323GkAliases.m_RegisteredH323Alias.ToLower() == rAlias.ToLower()
                || (*it).second->FindMixedAlias(rAlias) == true )
            {
                (*it).second->IncRef(__FILE__, __LINE__);
                matchingEPs.push_back((*it).second);
            }
        }
    }
    m_EndpointsCritSect.Signal();

    if ( matchingEPs.empty() )
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() NO endpoint found!" );
    }
    else
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Found " << matchingEPs.size() << " endpoint(s) for alias '"<< rAlias << "'" );
    }
    return matchingEPs;
}

//-------------------------------------------------------------------

H323RemoteEndpoint * H323RegisteredEPList::FindEndpointByAliasAndLineId(const Tools::SString & rAlias,
                                                                        int iLineID,
                                                                        const Tools::SIpAddress & rIpAddress,
                                                                        bool bMatchIpAndPort)
{
    SDEFFUNC("H323RegEPList::FindEpByAliasLneId");

    H323RemoteEndpoint * pMatchingEndpoint = 0;

    if ( iLineID > 0 )
    {
        STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
            "() Search endpoint by alias '" << rAlias << "'and lineId= " << iLineID);
    }
    else
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() NO valid line Id");
        return 0;
    }

    H323RemoteEndpointHashmap::iterator it;
    // 1st: find endpoint(s) with matching alias AND IP address AND port
    m_EndpointsCritSect.Wait();
    for (it = m_Endpoints.begin(); it != m_Endpoints.end(); ++it)
    {
        if ( !SBAD_READPTR((*it).second) )
        {
            if (   (    (*it).second->m_H323GkAliases.m_RegisteredH323Alias.ToLower() == rAlias.ToLower()
                     && (*it).second->m_iLineId == iLineID )
                || ((*it).second->FindMixedAlias(rAlias) == true) )
            {
                if ( bMatchIpAndPort && (*it).second->m_IpAddress == rIpAddress )
                {
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                        "() Found matching endpoint with alias '" << rAlias
                        << "', lineId= " << iLineID << " and " << (*it).second->m_IpAddress );
                    pMatchingEndpoint = (*it).second;
                    break;
                }
                else if ( !bMatchIpAndPort )
                {
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                        "() Found matching endpoint with alias '" << rAlias
                        << "' and lineId= " << iLineID);
                    pMatchingEndpoint = (*it).second;
                    break;
                }
            }
        }
    }
    m_EndpointsCritSect.Signal();

    if ( pMatchingEndpoint == 0 )
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() NO endpoint found!" );
    }
    return pMatchingEndpoint;
}

//-------------------------------------------------------------------

H323RemoteEndpoint * H323RegisteredEPList::FindEndpointByAliases(const Tools::SStringVector & rAliases,
                                                                 const Tools::SIpAddress    & rIpAddress,
                                                                 bool bMatchIpAndPort,
                                                                 bool bSkipExpiredEPs)
{
    H323RemoteEndpoint * pEndpoint;
    for(size_t ii = 0; ii < rAliases.size(); ii++)
    {
        pEndpoint = FindEndpointByAlias(rAliases[ii], rIpAddress, bMatchIpAndPort, bSkipExpiredEPs);
        if (pEndpoint != 0)
            return pEndpoint;
    }
    return 0;
}

//-------------------------------------------------------------------

H323RemoteEndpoint * H323RegisteredEPList::FindEndpointByEndpointId(const Tools::SString & rEndpointIdentifier)
{
    SDEFFUNC("H323RegEPList::FindEpByEpId");

    H323RemoteEndpointHashmap::iterator it;

    m_EndpointsCritSect.Wait();
    if ((it = m_Endpoints.find(rEndpointIdentifier)) != m_Endpoints.end())
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Found endpoint by endpointIdentifier '" << rEndpointIdentifier << "'");
        m_EndpointsCritSect.Signal();
        return it->second;
    }
    m_EndpointsCritSect.Signal();
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Endpoint not found by endpointIdentifier '" << rEndpointIdentifier << "'");
    return 0;
}

//-------------------------------------------------------------------

WORD H323RegisteredEPList::FindPortByCallSignalingAddr(const Tools::SIpAddress & rIpAddress)
{
    Tools::SIPSocket::Address IPAddr;
    WORD wPort;

    H323RemoteEndpointHashmap::iterator it;

    m_EndpointsCritSect.Wait();
    for ( it = m_Endpoints.begin(); it != m_Endpoints.end(); ++it )
    {
        if (   (*it).second
            && ((*it).second->m_IpAddress == rIpAddress) )
        {
            (*it).second->m_CallSignalAddress.GetIpAndPort( IPAddr, wPort );
            if (IPAddr == rIpAddress.m_Address)
            {
                m_EndpointsCritSect.Signal();
                return wPort;
            }
        }
    }
    m_EndpointsCritSect.Signal();

    return 0;
}

//-------------------------------------------------------------------

WORD H323RegisteredEPList::FindPortByH323Alias(const Tools::SString    & rH323Alias,
                                               const Tools::SIpAddress & rIpAddress)
{
    SDEFFUNC("H323Regist::FindPortByH323Alias");

    Tools::SIPSocket::Address IPAddr;
    WORD wPort;

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Searching for signalling port of endpoint with alias '" << rH323Alias
        << "' in list (size: " << m_Endpoints.size() << ")...");

    H323RemoteEndpoint * pEndpoint = FindEndpointByAlias(rH323Alias, rIpAddress, true);
    if (!SBAD_READPTR(pEndpoint))
    {
        pEndpoint->m_CallSignalAddress.GetIpAndPort(IPAddr, wPort);
        STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
            "() Found entry: " << IPAddr << ", port " << wPort);
        return wPort;
    }
    return 0;
}

//-------------------------------------------------------------------

WORD H323RegisteredEPList::FindPortByH323AliasAndLineId(const Tools::SString    & rH323Alias,
                                                        int iLineID,
                                                        const Tools::SIpAddress & rIpAddress)
{
    SDEFFUNC("H323Regist::FindPortByH323AliasLId");

    Tools::SIPSocket::Address IPAddr;
    WORD wPort;

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Searching for signalling port of endpoint with alias '" << rH323Alias
        << "', lineId= " << iLineID << " in list (size: " << m_Endpoints.size() << ")...");

    H323RemoteEndpoint * pEndpoint = FindEndpointByAliasAndLineId(rH323Alias, iLineID, rIpAddress, true);
    if (!SBAD_READPTR(pEndpoint))
    {
        pEndpoint->m_CallSignalAddress.GetIpAndPort(IPAddr, wPort);
        STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
            "() Found entry: " << IPAddr << ", port " << wPort);
        return wPort;
    }
    return 0;
}

//-------------------------------------------------------------------

bool H323RegisteredEPList::FindIsRequiresDelayedDisconnectEPByH323Alias(const Tools::SString    & rH323Alias,
                                                                        const Tools::SIpAddress & rIpAddress)
{
    SDEFFUNC("H323Reg::FindDDiscEPByH323Alias");

    Tools::SIPSocket::Address IPAddr;

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Checking for special EP which requires 'delayed disconnect' with alias '" << rH323Alias
        << "' in list (size: " << m_Endpoints.size() << ")...");

    H323RemoteEndpoint * pEndpoint = FindEndpointByAlias(rH323Alias, rIpAddress, true);
    if (!SBAD_READPTR(pEndpoint))
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() is special EP: " << (pEndpoint->m_bRequiresDelayedDisconnect ? "YES" : "no"));
        return pEndpoint->m_bRequiresDelayedDisconnect;
    }
    return false;
}

//-------------------------------------------------------------------

Tools::SString H323RegisteredEPList::FindH323AliasByCallIdentifier(const H323GloballyUniqueID & rCallIdentifier,
                                                                   const Tools::SStringVector & rE164Aliases)
{
    H323RemoteEndpointHashmap::iterator it;
    
    m_EndpointsCritSect.Wait();
    for ( it = m_Endpoints.begin(); it != m_Endpoints.end(); ++it )
    {
        if ( (*it).second )
        {
            // search in call identifier list
            H323CallIdentifierList::iterator CallIdListIter;
            for( CallIdListIter =  (*it).second->m_H323CallIdentifier.begin();
                 CallIdListIter != (*it).second->m_H323CallIdentifier.end();
                 ++CallIdListIter )
            {
                if ( (*CallIdListIter) == rCallIdentifier )
                {
                    // Check for usage of mixed aliases
                    if (!(*it).second->m_H323GkAliases.m_RegisteredMixedAliases.empty())
                    {
                        // Search for matching mixed alias
                        SStringVector::const_iterator IterH;
                        SStringVector::const_iterator IterE;
                        SStringVector::const_iterator IterM;
                        SStringVector                 szVectorMixed = (*it).second->m_H323GkAliases.m_RegisteredMixedAliases;
                        SString                       szMixedAlias;
                        for (IterH = (*it).second->m_H323GkAliases.m_H323Aliases.begin(); IterH != (*it).second->m_H323GkAliases.m_H323Aliases.end(); ++IterH)
                        {
                            for (IterE = rE164Aliases.begin(); IterE != rE164Aliases.end(); ++IterE)
                            {
                                for (IterM = szVectorMixed.begin(); IterM != szVectorMixed.end(); ++IterM)
                                {
                                    szMixedAlias = (*IterH) + (*IterE);
                                    if (szMixedAlias == (*IterM))
                                    {
                                        m_EndpointsCritSect.Signal();
                                        return szMixedAlias;
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        m_EndpointsCritSect.Signal();
                        return (*it).second->m_H323GkAliases.m_RegisteredH323Alias;
                    }
                }
            }
        }
    }
    m_EndpointsCritSect.Signal();

    return Tools::SString();
}

