/*---------------------------------------------------------------------------

   Swyx Solutions GmbH
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Solutions GmbH are
  Copyright (C) 2000-2005 Swyx Solutions GmbH.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Swyx non standard data helper methods
 
    Comments:

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxGate v5.0.x.x.root/H323/source/SwyxExtHelper.cp $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 25.10.05 16:18                                               $
   $Modtime:: 25.10.05 16:08                                               $
  $Revision:: 5                                                            $
  $Workfile:: SwyxExtHelper.cpp                                            $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>

#include "SwyxExt.h"
#include "SwyxExtHelper.h"

using namespace std;
using namespace Tools;

bool SwyxH225NonStandardData::BuildSwH225NonStandardData(Tools::SByteVector & rawData)
{
    SDEFFUNC("SwyxH225NSData::BuildH225NSData");

    SwyxExt_Data SwyxData;
    SwyxData.m_version = iSwyxH225NonStandardDataVersion;

    switch ( m_NSDataChoice )
    {
        case SwyxH225NonStandardData::CallInfo :
        {
            SwyxData.m_data_message_body.SetTag(SwyxExt_Data_data_message_body::e_callInfo);
            SwyxExt_CallInfo & callInfo = (SwyxExt_CallInfo &)SwyxData.m_data_message_body;
            switch(m_uFaxType)
            {
            case SwyxH225NonStandardData::UnknownFT :
                STRACE2_INFO(m_uTraceId, this, GetConnID(), "() Fax type: No fax");
                callInfo.m_faxType.SetTag(SwyxExt_FaxType::e_noFax);
                break;
            case SwyxH225NonStandardData::T30 :
                STRACE2_INFO(m_uTraceId, this, GetConnID(), "() Fax type: Normal fax");
                callInfo.m_faxType.SetTag(SwyxExt_FaxType::e_normalFax);
                break;
            case SwyxH225NonStandardData::T38 :
                STRACE2_INFO(m_uTraceId, this, GetConnID(), "() Fax type: T.38 fax");
                callInfo.m_faxType.SetTag(SwyxExt_FaxType::e_t38Fax);
                break;
            default:
                STRACE2_ERROR(m_uTraceId, this, GetConnID(), "() Fax type: <unknown>");
                return false;
            }
            if ( m_uLineId > 0 )
            {
                callInfo.IncludeOptionalField( SwyxExt_CallInfo::e_lineId );
                callInfo.m_lineId = m_uLineId;
            }
        }
        break;

        case SwyxH225NonStandardData::FaxChannels :
        {
            SwyxData.m_data_message_body.SetTag(SwyxExt_Data_data_message_body::e_faxChannelInfo);
            SwyxExt_FaxChannelInfo & faxChannelInfo = (SwyxExt_FaxChannelInfo &)SwyxData.m_data_message_body;
            if ( m_uFaxChannels > 0 )
            {
                faxChannelInfo.m_numOfChannels = m_uFaxChannels;
            }
        }
        break;

        default :
            STRACE2_ERROR(m_uTraceId, this, GetConnID(), "()  NonStandardData type: <unknown>");
            return false;
    }

    // Add (optional) endpoint vendor identifier (including product and version ID) if vendor ID member is set
    if (m_H225VendorIdentifier.m_vendor.m_manufacturerCode != 0)
    {
        SwyxData.IncludeOptionalField(SwyxExt_Data::e_endpointVendor);
        SwyxData.m_endpointVendor = m_H225VendorIdentifier;
    }

    // Encode Swyx extension data
    SPerStream strm;
    SwyxData.Encode(strm);
    strm.CompleteEncoding();

    if (STrace::CanTrace(g_pH323Module->m_H225TraceId, STrace::Debug))
    {
        STRACE2_INFO(m_uTraceId, this, GetConnID(), "() Building nonStandardData:\n"
            << STrace::MultiLinePrefix << "  nonStandardData\n" << STrace::MultiLinePrefix
            << "  "    << setprecision(2) << setfill(' ') << SwyxData);
    }

    rawData = strm;
    return true;
} // End of BuildSwH225NonStandardData

//-----------------------------------------------------------------------------

bool SwyxH225NonStandardData::HandleSwH225NonStandardData( const Tools::SByteVector & rawData )
{
    SDEFFUNC("SwyxH225NSData::HdlH225NSData");

    SPerStream                  strm;
    SwyxExt_Data SwyxData;

    if (!rawData.empty())
    {
        strm = rawData;
        try
        {
            if (!SwyxData.Decode(strm))
            {
                STRACE2_ERROR(m_uTraceId, this, GetConnID(),
                    "() Read error: PER decode failure if nonStandardData!");
                return false;
            }
            if (STrace::CanTrace(g_pH323Module->m_H225TraceId, STrace::Debug))
            {
                STRACE2_INFO(m_uTraceId, this, GetConnID(),
                    "() Handling nonStandardData:\n" << STrace::MultiLinePrefix
                    << "  nonStandardData\n" << STrace::MultiLinePrefix
                    << "  "    << setprecision(2) << setfill(' ') << SwyxData);
            }

            if (SwyxData.m_version != iSwyxH225NonStandardDataVersion)
            {
                STRACE2_ERROR(m_uTraceId, this, GetConnID(),
                    "() Unknown nonStandardData version: " << SwyxData.m_version  );
                return false;
            }

            if ( SwyxData.m_data_message_body.GetTag() == SwyxExt_Data_data_message_body::e_callInfo )
            {
                m_NSDataChoice = SwyxH225NonStandardData::CallInfo;

                const SwyxExt_CallInfo & callInfo = SwyxData.m_data_message_body;
                switch(callInfo.m_faxType.GetTag())
                {
                case SwyxExt_FaxType::e_noFax:
                    m_uFaxType = UnknownFT;
                    STRACE2_INFO(m_uTraceId, this, GetConnID(), "() Fax type: No fax");
                    break;
                case SwyxExt_FaxType::e_normalFax:
                    m_uFaxType = T30;
                    STRACE2_INFO(m_uTraceId, this, GetConnID(), "() Fax type: Normal fax");
                    break;
                case SwyxExt_FaxType::e_t38Fax:
                    m_uFaxType = T38;
                    STRACE2_INFO(m_uTraceId, this, GetConnID(), "() Fax type: T.38 fax");
                    break;
                default:
                    STRACE2_ERROR(m_uTraceId, this, GetConnID(), "() Fax type: <unknown>");
                    return false;
                }
                if ( callInfo.HasOptionalField(SwyxExt_CallInfo::e_lineId) )
                    m_uLineId = callInfo.m_lineId;
            }
            else if ( SwyxData.m_data_message_body.GetTag() == SwyxExt_Data_data_message_body::e_faxChannelInfo )
            {
                m_NSDataChoice = SwyxH225NonStandardData::FaxChannels;

                const SwyxExt_FaxChannelInfo & channelInfo = SwyxData.m_data_message_body;
                m_uFaxChannels = channelInfo.m_numOfChannels;
                STRACE2_INFO(m_uTraceId, this, GetConnID(), "() Fax channels: " << m_uFaxChannels);
            }

            if (SwyxData.HasOptionalField(SwyxExt_Data::e_endpointVendor))
            {
                m_H225VendorIdentifier = SwyxData.m_endpointVendor;
            }
        }

        catch(...)
        {
            STRACE2_FATAL(m_uTraceId, this, GetConnID(),
                "### Exception in nonStandardData ASN decode thrown! ###");
            return false;
        }
    }
    return true;

} // End of HandleSwH225NonStandardData
