/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Gatekeeper client
 
    Comments: Gatekeeper client protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxGate v5.0.x.x.root/H323/source/gkclient.cpp     $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 16.11.05 17:45                                               $
   $Modtime:: 16.11.05 17:45                                               $
  $Revision:: 12                                                           $
  $Workfile:: gkclient.cpp                                                 $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "gkclient.h"
#include "h323pdu.h"
#include "SwyxExt.h"

using namespace std;
using namespace Tools;

class H225RasThread : public SThread
{
public:
    H225RasThread(H323Gatekeeper * gatekeeper,
                  H323EndPoint & endpoint);
    
protected:
    void Main();
    
    H323Gatekeeper * gatekeeper;
};


#define new SNEW


/////////////////////////////////////////////////////////////////////////////

H225RasThread::H225RasThread(H323Gatekeeper * gk,
                             H323EndPoint   & endpoint)
 : SThread(endpoint.GetRasThreadStackSize(), NoAutoDeleteThread, SThread::HighPriority)
{
    if ( !SBAD_READPTR(gk) )
    {
        gatekeeper = gk;
        gatekeeper->IncRef(__FILE__, __LINE__);
    }
    Resume();
}


void H225RasThread::Main()
{
    SDEFFUNC("H225RasThread::Main");
    
    if ( !SBAD_READPTR(gatekeeper) )
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Started RAS thread");

        gatekeeper->HandleRasChannel();
        gatekeeper->SetRASChannelInvalid();

        gatekeeper->DecRef(__FILE__, __LINE__);
        gatekeeper = 0;

        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Stopped RAS thread");
    }
    else
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() Failed to start RAS thread due to invalid gatekeeper!");
    }
}


/////////////////////////////////////////////////////////////////////////////
// GK client (H323 endpoint) specific data.

H323Gatekeeper::Client::Client( H323Gatekeeper    * gk,
                                H323EndPoint      & ep,
			                    ClientDataStruct  & cd )
 : endpoint(ep),
   gatekeeper(0),
   m_ClientData(cd),
   m_GKRegStatus(gk_RegistrationStatusUnknown),
   m_bRegisteredToMainGK(false),
   m_uNumberOfRRQAttempts(0)
{
    SDEFFUNC("H323Gatekeeper::Client::Client");

    if ( !SBAD_READPTR(gk) )
    {
        gatekeeper = gk;
        gatekeeper->IncRef(__FILE__, __LINE__);
    }
   
    m_RegTimer.SetNotifier( SCREATE_NOTIFIER( OnRegistrationExpiry ) );

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0, "(" << gk << ")");
}

//-----------------------------------------------------------------------------

H323Gatekeeper::Client::~Client()
{
    SDEFFUNC("H323Gatekeeper::Client::~Client");

    if ( !SBAD_READPTR(gatekeeper) )
    {
        gatekeeper->DecRef(__FILE__, __LINE__);
        gatekeeper = 0;
    }

    // if case of...
    m_RegTimer.Stop();

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0, "()");
}

//-----------------------------------------------------------------------------

void H323Gatekeeper::Client::OnRegistrationExpiry( STimer & Timer, int iContinous )
{
    SDEFFUNC("H323Gatekeeper::Client::OnRegExp");
    
    // Check if RAS thread is ok
    if (gatekeeper->IsRASChannelInvalid())
    {
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() RAS channel invalid -> starting GK register thread, endpoint id="
            << m_szEndpointIdentifier);
        
        // If thread is already running it will return!
        gatekeeper->StartGkRegisterThread();
        return;
    }
    
    if ( m_GKRegStatus == gk_RegistrationStatusOK )
    {
        m_GKRegStatus = gk_RegistrationStatusInKeepAlive; 
        
        STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
            "() Registration TTL expired -> try a lightweight RRQ for endpoint id=" << m_szEndpointIdentifier);
        
        bool bRegisterResult = false;
        if ( m_ClientData.clientType == Client::Gateway )
            bRegisterResult = gatekeeper->RegisterClientAgain(m_ClientData.szH323Alias, true, m_ClientData.Data.GatewayData.lineId);
        else
            bRegisterResult = gatekeeper->RegisterClientAgain(m_ClientData.szH323Alias);
        
        if ( bRegisterResult == true )
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() lightweight RRQ successful" );
            return;
        }
    }
    
    // first time registration is lost?
    if ( m_GKRegStatus != gk_RegistrationStatusRegistrationLost )
    {
        m_GKRegStatus = gk_RegistrationStatusRegistrationLost;
        
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() RAS registration lost for endpoint id="
            << m_szEndpointIdentifier);

        // perform all necessary actions to reestablish GK connection in endpoint...
        endpoint.OnGatekeeperRegistrationLost( reinterpret_cast<int>(this) );
    }
}


/////////////////////////////////////////////////////////////////////////////
// GK register thread.

// we're storing the this pointer only. Therefore the warning can be ignored
#pragma warning(disable:4355)
H323Gatekeeper::GKRegisterThread::GKRegisterThread(H323Gatekeeper & gk,
                                                   H323EndPoint   & ep,
                                                   H323Transport  * tp,
                                                   const SStringVector & H323Aliases)
 : SThread(ep.GetRasThreadStackSize(), NoAutoDeleteThread),
   gatekeeper(gk),
   endpoint(ep),
   transport(tp),
   m_H323Aliases(H323Aliases),
   m_bStopReRegistration(false)
#ifdef _DEBUG
 , m_ReRegistrationSyncPoint(Tools::SString(Tools::SString::Printf, "H323Gatekeeper::GKRegisterThread::ReRegistrationSyncPoint[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#endif
#pragma warning(default:4355)
{
    Resume();
}

//-----------------------------------------------------------------------------

H323Gatekeeper::GKRegisterThread::~GKRegisterThread()
{
    if ( !m_bStopReRegistration )
    {
        m_bStopReRegistration = true;
        // stop reregistration attemps
        m_ReRegistrationSyncPoint.Signal();
        WaitForTermination();
    }
}

//-----------------------------------------------------------------------------

void H323Gatekeeper::GKRegisterThread::Main()
{
    SDEFFUNC("H323Gatekeeper::GKRegThread::Main");
    
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Started GK registration thread");
    
    SIPSocket::Address binding;
    unsigned uStartPort = gatekeeper.GetTransportAddress().m_Port;

    if ( !SBAD_READPTR(transport) )
    {
        transport->CleanUpOnTermination();
        delete transport;
    }
    transport = new H323TransportUDP(endpoint,binding,uStartPort);
    gatekeeper.SetH323Transport(transport);
    
    while ( !m_bStopReRegistration )
    {
        if ( gatekeeper.Discover(gatekeeper.GetGatekeeperName()) )
        {
            gatekeeper.SetRASChannelValid();
            
            if( m_bStopReRegistration )
                break;
            
			for (SStringVector::size_type ii = 0; ii < m_H323Aliases.size(); ii++)
            {
                if (!m_H323Aliases[ii].empty())
                {
                    Tools::SStringVector H323Alias = m_H323Aliases[ii].Tokenize("|");
                    int iLineId = 0;
                    if (H323Alias.size() > 1 && !H323Alias[1].empty())
                        iLineId = H323Alias[1].AsInteger();
                    
                    if ( !gatekeeper.RegistrationRequest(H323Alias[0], false, iLineId, 0) )
                    {
                        // give UP!
                        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                            "() Client '" << m_H323Aliases[ii] << "' FAILED to register!");
                    }
                }
            }
            break;
        }
        
        if( m_bStopReRegistration )
            break;
        
        // repeat discovery/registration cycle until succeeding (timeout in seconds!)
        m_ReRegistrationSyncPoint.Wait( endpoint.GetReRegisterTimeout() * 1000 );
        
        if ( !SBAD_READPTR(transport) )
        {
            transport->CleanUpOnTermination();
            delete transport;
        }
        transport = new H323TransportUDP(endpoint,binding,uStartPort);
        gatekeeper.SetH323Transport(transport);
    }
    
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Stopped GK registration thread");
}



/////////////////////////////////////////////////////////////////////////////
// we're storing the this pointer only. Therefore the warning can be ignored
#pragma warning(disable:4355)
H323Gatekeeper::H323Gatekeeper(H323EndPoint & ep,const Tools::SIpAddress & transportAddress)
 : endpoint(ep)
#ifdef _DEBUG
 , m_GkMutex(Tools::SString(Tools::SString::Printf, "H323Gatekeeper::GkMutex[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#endif
#pragma warning(default:4355)
{
    SDEFFUNC("H323Gatekeeper::H323Gatekeeper");

    m_uNextRASReqSequenceNumber    = 0;
    m_uLastRASReqSequenceNumber    = 0;
    m_uLastRASReqTag               = 0;
    m_uAllocatedBandwidth          = 0;
    
    m_bGatekeeperDiscoveryComplete = false;
    m_bRASChannelInvalid           = false;
    m_bIsAlternativeGK             = false;
    m_bShutdownGatekeeper          = false;

    m_pGKRegisterThread            = 0;
    
    // calculate MAXIMUM RAS request timeout as: timeout in seconds for RAS request * number of retries + 5 second spare time
    m_MaxMakeReqestTimeout   = STimeInterval( 0, (endpoint.GetGatekeeperRequestRetries() * endpoint.GetGatekeeperRequestTimeout().GetSeconds()) + 5);
    
    // create transport object from transport address given...
    transport = new H323TransportUDP( ep,
        transportAddress.m_Address,
        transportAddress.m_Port );
    
    m_TransportAddress   = transportAddress;
    m_LastTransportError = 0;
    

    m_RASReqExpiryCheckTimer.SetNotifier( SCREATE_NOTIFIER( OnRASReqExpiryCheck ) );
    // add reference to ourself because we will start the RAS request expiry check timer!
    IncRef(__FILE__, __LINE__);
    m_RASReqExpiryCheckTimer = 1000;

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() master RAS request timeout " << m_MaxMakeReqestTimeout.GetSeconds() << "sec" );
}

//-----------------------------------------------------------------------------

H323Gatekeeper::~H323Gatekeeper()
{
    SDEFFUNC("H323Gatekeeper::~H323Gatekeeper");

    // just in case of...
    if (transport != NULL)
    {
        transport->CleanUpOnTermination();
        delete transport;
        transport = 0;
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "()");
}

//-----------------------------------------------------------------------------

void H323Gatekeeper::PrintOn(ostream & strm) const
{
    if (!m_szGatekeeperIdentifier.empty())
        strm << m_szGatekeeperIdentifier << "@";
    if (transport != NULL)
    {
        strm << transport->GetRemoteAddressString();
    }
}

//-----------------------------------------------------------------------------

void H323Gatekeeper::Shutdown()
{
    SDEFFUNC("H323Gatekeeper::Shutdown");

    // set shutdown here, to stop all ongoing activities
    m_bShutdownGatekeeper = true;

    if ( m_RASReqInProgressMap.empty() )
    {
        if ( m_RASReqExpiryCheckTimer.Stop() )
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0, "() stopped RAS request expiry timer");
            DecRef(__FILE__, __LINE__);
        }
    }
    else // uh uh - we have a pending RAS request to terminate!
    {
        STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0, "() terminate pending RAS request(s)...");

        m_GkMutex.Wait();
        
        RASReqInProgress *pRASReqInProgress = 0;
        unsigned uRASReqSequenceNumber = 0;

        RASReqInProgressMap::iterator pos = m_RASReqInProgressMap.begin();
        while ( pos != m_RASReqInProgressMap.end() )
        {
            // get values from map element...
            uRASReqSequenceNumber = (*pos).first;
            pRASReqInProgress = (*pos).second;
            // ...and increment iterator (may be reset after erase() later on!)
            ++pos;

            if (!SBAD_READPTR(pRASReqInProgress))
            {
                STRACE2_INFO3(g_pH323Module->m_RasTraceId, this, 0, "() RAS request '"
                    << pRASReqInProgress->m_rPendingReq.GetTagName() << "' , seqNum="
                    << uRASReqSequenceNumber << " will be teminated!");
                        
                // max number of retries exceeded -> remove request from list!
                m_RASReqInProgressMap.erase(uRASReqSequenceNumber);
                // reset iterator position
                pos = m_RASReqInProgressMap.begin();
                // unblock pending request with timeout
                pRASReqInProgress->m_RASReqResult = RASReqInProgress::RequestTimeout;
                pRASReqInProgress->m_PendingReqSync.Signal();
            }
        }
        m_GkMutex.Signal();
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "()");
}

//-----------------------------------------------------------------------------

void H323Gatekeeper::StartGkRegisterThread()
{
    SDEFFUNC("H323Gatekeeper::StartGkRegThread");
    
    SWaitAndSignal wait(m_GkMutex);
    
    if (m_pGKRegisterThread)
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Gk reregister thread already running.");
        return;
    }
    
    // Get alias of all clients in map.
    SStringVector Aliases;
    ClientsMap::iterator iter;
    for (iter = m_Clients.begin(); iter != m_Clients.end(); ++iter)
    {
        if ((*iter).second)
        {
            Aliases.push_back((*iter).first);
        }
    }
    m_pGKRegisterThread = new H323Gatekeeper::GKRegisterThread( *this, endpoint, transport, Aliases );
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::AreAllClientsRegistered()
{
    SDEFFUNC("H323Gatekeeper::AreAllClientsReg");
    
    // Check if all clients are registerd again
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Check if all clients are registered...");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator iter;
    for (iter = m_Clients.begin(); iter != m_Clients.end(); ++iter)
    {
        SString szClient = (*iter).first;
        Client *pClient  = (*iter).second;
        if (!SBAD_READPTR(pClient))
        {
            if (!pClient->IsRegistered())
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Client with alias '" << szClient << "' NOT registered!");
                return false;
            }
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Client with alias '" << szClient << "' is registered!");
        }
        else
        {
            m_Clients.erase(iter);
            // iterator is now invalid -> reset to first element in map!
            iter = m_Clients.begin();
        }
    }
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::UnregisterAllClients(int reason)
{
    SDEFFUNC("H323GK::UnregisterAllClients");

    m_GkMutex.Wait();
    // Perform URQ on all clients in map.
    ClientsMap::iterator iter = m_Clients.begin();
    while ( iter != m_Clients.end() )
    {
        Client *pClient  = (*iter).second;
        if ( !SBAD_READPTR(pClient) )
        {
            if ( pClient->IsRegistered() )
            {
                // stop registration timer in any case!
                pClient->m_RegTimer.Stop();

                if ( pClient->m_ClientData.clientType == Client::Gateway )
                {
                    m_GkMutex.Signal();
                    UnregistrationRequest( pClient->m_ClientData.szH323Alias,
                                        pClient->m_ClientData.Data.GatewayData.lineId,
                                        reason );
                    m_GkMutex.Wait();
                }
                else
                {
                    m_GkMutex.Signal();
                    UnregistrationRequest( pClient->m_ClientData.szH323Alias, 0, reason );
                    m_GkMutex.Wait();
                }
            }
            // remove reference for map and...
            pClient->DecRef(__FILE__, __LINE__);
            pClient = 0;
        }
        // ...remove client from map in any case!
        m_Clients.erase(iter);
        // iterator is now invalid -> reset to first element in map!
        iter = m_Clients.begin();
    }
    m_GkMutex.Signal();

    // set shutdown GK flag
    Shutdown();

    if ( !SBAD_READPTR(m_pGKRegisterThread) )
    {
        delete m_pGKRegisterThread;
        m_pGKRegisterThread = 0;
    }
    
    m_GkMutex.Wait();
    if ( !SBAD_READPTR(transport) )
    {
        transport->CleanUpOnTermination();
        delete transport;
        transport = 0;
    }
    m_GkMutex.Signal();

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() All client(s) unregistered!");

    return true;
}

//-----------------------------------------------------------------------------

H323Gatekeeper::Client * H323Gatekeeper::FindClientByAliasAndLineId(const Tools::SString & rAlias, int iLineId)
{
    SDEFFUNC("H323Gk::FindClientByAliasLineId");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator it;

    // if we have no valid line id we use the alias only!
    if ( iLineId == 0 )
        return FindClientByAlias(rAlias);

    SString szSearchAlias;
    szSearchAlias.sprintf("%s|%d", rAlias.c_str(), iLineId);
    for (it = m_Clients.begin(); it != m_Clients.end(); ++it)
    {
        if (szSearchAlias.ToLower() == (*it).first.ToLower())
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() Found client by alias '" << rAlias << "' and line Id " << iLineId);
            return it->second;
        }
    }
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Client not found by alias '" << rAlias << "' and line Id " << iLineId);
    return 0;
}

//-----------------------------------------------------------------------------

H323Gatekeeper::Client * H323Gatekeeper::FindClientByAlias(const Tools::SString & rAlias)
{
    SDEFFUNC("H323Gatekeeper::FindClientByAlias");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator it;
    for (it = m_Clients.begin(); it != m_Clients.end(); ++it)
    {
        if (rAlias.ToLower() == (*it).first.ToLower())
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() Found client by alias '" << rAlias << "'");
            return it->second;
        }
    }
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Client not found by alias '" << rAlias << "'");
    return 0;
}

//-----------------------------------------------------------------------------

H323Gatekeeper::Client * H323Gatekeeper::FindClientBySeqNum(unsigned seqNum)
{
    SDEFFUNC("H323Gatekeeper::FindClientBySeqNu");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator it;
    for (it = m_Clients.begin(); it != m_Clients.end(); ++it)
    {
        if (   (*it).second
            && ((*it).second->m_uLastReqSequenceNumber == seqNum) )
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() Found client by seqNum " << seqNum);
            return it->second;
        }
    }
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Client not found by seqNum" << seqNum);
    return 0;
}

//-----------------------------------------------------------------------------

H323Gatekeeper::Client * H323Gatekeeper::FindClientByEndpointId(const Tools::SString & rEndpointId)
{
    SDEFFUNC("H323Gatekeeper::FindClientByEpId");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator it;
    for (it = m_Clients.begin(); it != m_Clients.end(); ++it)
    {
        if (   (*it).second
            && ((*it).second->m_szEndpointIdentifier == rEndpointId) )
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() Found client by endpointIdentifier '" << rEndpointId << "'");
            return it->second;
        }
    }
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Client not found by endpointIdentifier '" << rEndpointId << "'");
    return 0;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::RegisterClientAgain(const Tools::SString & szH323Alias,
                                         bool                   bIsLightWeightRRQ,
                                         int                    iLineId,
                                         bool                   bSetStatusToUnregistered )
{
    SDEFFUNC("H323Gatekeeper::RegClientAgain");
    
    m_GkMutex.Wait();
    // Find client by H.323 alias to check/set 'unregistered' flag
    Client * pClient = FindClientByAliasAndLineId( szH323Alias, iLineId );
    if (SBAD_READPTR(pClient))
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() client with alias '" << szH323Alias << "' NOT found!");
        m_GkMutex.Signal();
        return false;
    }
    // should we change the registration status AND is registration lost for first time ?
    if (   bSetStatusToUnregistered == true
        && pClient->m_GKRegStatus != gk_RegistrationStatusRegistrationLost )

    {
        // yep! -> set registration status to 'unregistered', inform endpoint and write eventlog entry!
        pClient->m_GKRegStatus = gk_RegistrationStatusRegistrationLost;
    }
    m_GkMutex.Signal();
  
    bool bRegReqResult = false;
    if ( pClient->m_ClientData.clientType == Client::Gateway )
    {
        bRegReqResult = RegistrationRequest( pClient->m_ClientData.szH323Alias,
                                             bIsLightWeightRRQ,
                                             pClient->m_ClientData.Data.GatewayData.lineId,
                                             0 );
    }
    else if ( pClient->m_ClientData.clientType == Client::FaxServer )
    {
        bRegReqResult = RegistrationRequest( pClient->m_ClientData.szH323Alias,
                                             bIsLightWeightRRQ,
                                             0,
                                             pClient->m_ClientData.Data.FaxServerData.numOfChannels );
    }
    else
    {
        bRegReqResult = RegistrationRequest( pClient->m_ClientData.szH323Alias,
                                             bIsLightWeightRRQ,
                                             0,
                                             0 );
    }
                              
    if ( bRegReqResult == false ) // perform RRQ
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() RRQ failed for client with alias '" << szH323Alias << "'");
        return false;
    }
    return true;
}

//-----------------------------------------------------------------------------

void H323Gatekeeper::HandleRasChannel()
{
    SDEFFUNC("H323Gatekeeper::HandleRasChannel");
    
    unsigned uCountDiscardedActions = 0;

    transport->SetReadTimeout(SMaxTimeInterval);
    H323RasPDU response;
    
    while(transport->IsOpen())
    {
        switch(response.Read(*transport))
        {
        case H323RasPDU::Success:
            {
                dynamic_cast<H323TransportUDP *>(transport)->
                    GetLastReceiveAddress(m_LastReceiveIpAddress.m_Address,
                    m_LastReceiveIpAddress.m_Port);
                
                if ( HandleRasPDU(response) == false )
                {
                    STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                        "() HandleRasPDU() failed!");
                }
                else if ( response.GetTag() == H225_RasMessage::e_registrationConfirm &&
                          m_pGKRegisterThread )
                {
                    // Check if all clients are registerd again
                    if (AreAllClientsRegistered() == true)
                    {
                        // Stop and delete the GK reregistration thread
                        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                            "() Stop and delete reregister thread...");
                        delete m_pGKRegisterThread;
                        m_pGKRegisterThread = 0;
                    }
                }
            }
            break;
            
        case H323RasPDU::TransportFailure:
            {
                if (transport->GetErrorNumber() != 0)
                {
                    STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                        "() Read on RAS channel failed: error=0x"
                        << hex << transport->GetErrorNumber() << dec << ", " << transport->GetErrorText());
                }
                return;
            }
            break;

        case H323RasPDU::DiscardLastAction:
            {
                // This can occure and is not handled as an error, because
                // when gatekeeper is down and we send e.g. a RRQ we get
                // an icmp-message from the machine running gatekeeper.
                // This will result in a failed read-operation on the socket,
                // but here we will ignore it and everything works fine!
                if ( ++uCountDiscardedActions == 1 )
                {
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                        "() Ignoring last error on RAS transport channel!");
                }
                else if ( uCountDiscardedActions > 10 )
                    uCountDiscardedActions = 0;
                continue;
            }
            break;
            
        case H323RasPDU::DecodeFailure:
            {
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() Decode error - discarding RAS message!");
                continue;
            }
            break;
            
        default:
            {
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() Unknown error reading RAS message - ignoring it!");
                continue;
            }
            break;
        }
    }
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::HandleRasPDU(const H323RasPDU & response)
{
    SDEFFUNC("H323Gatekeeper::HandleRasPDU");
    
    // Tracing
    if (STrace::CanTrace(g_pH323Module->m_GkTraceId, STrace::Info))
    {
        SString szMessage = response.GetTagName();
        szMessage[0] = ::toupper(szMessage[0]);
        
        if (m_LastReceiveIpAddress == SIpAddress())
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS " << szMessage);
        }
        else
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS " << szMessage << " from " << m_LastReceiveIpAddress);
        }
    }
    
    switch (response.GetTag())
    {
    case H225_RasMessage::e_gatekeeperRequest :
        return OnReceiveGatekeeperRequest(response);
        
    case H225_RasMessage::e_gatekeeperConfirm :
        return OnReceiveGatekeeperConfirm(response);
        
    case H225_RasMessage::e_gatekeeperReject :
        return OnReceiveGatekeeperReject(response);
        
    case H225_RasMessage::e_registrationRequest :
        return OnReceiveRegistrationRequest(response);
        
    case H225_RasMessage::e_registrationConfirm :
        return OnReceiveRegistrationConfirm(response);
        
    case H225_RasMessage::e_registrationReject :
        return OnReceiveRegistrationReject(response);
        
    case H225_RasMessage::e_unregistrationRequest :
        return OnReceiveUnregistrationRequest(response);
        
    case H225_RasMessage::e_unregistrationConfirm :
        return OnReceiveUnregistrationConfirm(response);
        
    case H225_RasMessage::e_unregistrationReject :
        return OnReceiveUnregistrationReject(response);
        
    case H225_RasMessage::e_admissionRequest :
        return OnReceiveAdmissionRequest(response);
        
    case H225_RasMessage::e_admissionConfirm :
        return OnReceiveAdmissionConfirm(response);
        
    case H225_RasMessage::e_admissionReject :
        return OnReceiveAdmissionReject(response);
        
    case H225_RasMessage::e_bandwidthRequest :
        return OnReceiveBandwidthRequest(response);
        
    case H225_RasMessage::e_bandwidthConfirm :
        return OnReceiveBandwidthConfirm(response);
        
    case H225_RasMessage::e_bandwidthReject :
        return OnReceiveBandwidthReject(response);
        
    case H225_RasMessage::e_disengageRequest :
        return OnReceiveDisengageRequest(response);
        
    case H225_RasMessage::e_disengageConfirm :
        return OnReceiveDisengageConfirm(response);
        
    case H225_RasMessage::e_disengageReject :
        return OnReceiveDisengageReject(response);
        
    case H225_RasMessage::e_locationRequest :
        return OnReceiveLocationRequest(response);
        
    case H225_RasMessage::e_locationConfirm :
        return OnReceiveLocationConfirm(response);
        
    case H225_RasMessage::e_locationReject :
        return OnReceiveLocationReject(response);
        
    case H225_RasMessage::e_infoRequest :
        return OnReceiveInfoRequest(response);
        
    case H225_RasMessage::e_infoRequestResponse :
        return OnReceiveInfoRequestResponse(response);
        
    case H225_RasMessage::e_nonStandardMessage :
        return OnReceiveNonStandardMessage(response);
        
    case H225_RasMessage::e_unknownMessageResponse :
        return OnReceiveUnknownMessageResponse(response);
        
    case H225_RasMessage::e_requestInProgress :
        return OnReceiveRequestInProgress(response);
        
    case H225_RasMessage::e_resourcesAvailableIndicate :
        return OnReceiveResourcesAvailableIndicate(response);
        
    case H225_RasMessage::e_resourcesAvailableConfirm :
        return OnReceiveResourcesAvailableConfirm(response);
        
    case H225_RasMessage::e_infoRequestAck :
        return OnReceiveInfoRequestAck(response);
        
    case H225_RasMessage::e_infoRequestNak :
        return OnReceiveInfoRequestNak(response);
        
    default :
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() RECV RAS unknown message from " << m_LastReceiveIpAddress); 
        return OnReceiveUnkown(response);
    }
    
    return false;
}

//-----------------------------------------------------------------------------

unsigned H323Gatekeeper::GetNextSequenceNumber()
{
    SWaitAndSignal lock (m_GkMutex);
    
    ++m_uNextRASReqSequenceNumber;
    if (m_uNextRASReqSequenceNumber > 65535)
        m_uNextRASReqSequenceNumber = 1;
    return m_uNextRASReqSequenceNumber;
}

//-----------------------------------------------------------------------------

H323Gatekeeper::RASReqInProgress *  H323Gatekeeper::MakeRequest(const H323RasPDU & request, const unsigned uSeqNumber)
{
    SDEFFUNC("H323Gatekeeper::MakeRequest");
    
    if ( m_bRASChannelInvalid )
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() RAS channel invalid -> discard request!" );
        return 0;
    }
    
    RASReqInProgress *pRASReqInProgress = new RASReqInProgress(request);
    if (SBAD_READPTR(pRASReqInProgress))
        return 0;

    // Tracing
    if (STrace::CanTrace(g_pH323Module->m_GkTraceId, STrace::Info))
    {
        Tools::SIPSocket::Address ip = dynamic_cast<H323TransportUDP *>(transport)->GetRemoteIpAddress();
        WORD port                    = dynamic_cast<H323TransportUDP *>(transport)->GetRemotePort();
        SString szMessage = request.GetTagName();
        szMessage[0] = ::toupper(szMessage[0]);
        
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS " << szMessage << ", seqNum= " << uSeqNumber << " to " << ip << ":" <<  port);
    }
    
    // lock all other RAS request related activity...
    m_GkMutex.Wait();
    // get retry and timeout values from endpoint
    pRASReqInProgress->m_ReqTimeout = endpoint.GetRasRequestTimeout();
    pRASReqInProgress->m_MaxNumberOfReqRetries = endpoint.GetRasRequestRetries();
    // save request timestamp
    pRASReqInProgress->m_LastReqAttempt = STimer::Tick();
    // try to send the RAS request message and wait for reply...
    m_RASReqInProgressMap.insert(pair< const unsigned, H323Gatekeeper::RASReqInProgress *>(uSeqNumber, pRASReqInProgress));

    if (pRASReqInProgress->m_rPendingReq.Write(*transport) != H323RasPDU::Success)
    {
        // sending RAS message failed -> remove from map with pending requests
        m_RASReqInProgressMap.erase(uSeqNumber);
        delete pRASReqInProgress;
        // save last error on transport level
        m_LastTransportError = transport->GetErrorNumber();
        // Unlock all other RAS request related activity...
        m_GkMutex.Signal();

        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "write to socket failed, error=" << m_LastTransportError );

        return 0;
    }
    // Unlock all other RAS request related activity...
    m_GkMutex.Signal();
    
    return pRASReqInProgress;
}

//-----------------------------------------------------------------------------

void H323Gatekeeper::CleanupReqInProgress(RASReqInProgress * pRASReqInProgress)
{
    SWaitAndSignal lock(m_GkMutex);
    
    RASReqInProgressMap::iterator pos = m_RASReqInProgressMap.begin();
    while ( pos != m_RASReqInProgressMap.end() )
    {
        if (pRASReqInProgress == (*pos).second)
        {
            m_RASReqInProgressMap.erase((*pos).first);
            break;
        }
        ++pos;
    }
    delete pRASReqInProgress;
}

//-----------------------------------------------------------------------------
bool H323Gatekeeper::CheckForResponse(unsigned reqTag, unsigned seqNum)
{
    SDEFFUNC("H323Gatekeeper::CheckForResponse");
    
    SWaitAndSignal lock(m_GkMutex);
    
    if (m_uLastRASReqTag != reqTag)
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "Received message with wrong tag (reqTag=" << reqTag << ") " <<
            " (requestTag=" << m_uLastRASReqTag << ") seqNum=" << seqNum << 
            " lastReqSequenceNumber=" << m_uLastRASReqSequenceNumber);
        
        return false;
    }
    
    if (seqNum == m_uLastRASReqSequenceNumber)
        return true;
    
    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
        "() Received incorrect sequence number, want "
        << m_uLastRASReqSequenceNumber << ", got " << seqNum <<
        " reqTag=" << reqTag);
    
    return false;
}

H323Gatekeeper::RASReqInProgress * H323Gatekeeper::CheckForReqInProgress(unsigned reqTag, unsigned seqNum)
{
    SDEFFUNC("H323Gk::CheckForReqInProgress");
    
    SWaitAndSignal lock (m_GkMutex);
    
    // try to find coresponding request in map
    RASReqInProgress *pRASReqInProgress = 0;
    RASReqInProgressMap::iterator pos = m_RASReqInProgressMap.find(seqNum);
    if ( pos == m_RASReqInProgressMap.end() )
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Received unknown sequence number " << seqNum <<
            " with reqTag=" << reqTag);
        return 0;
    }
    
    pRASReqInProgress = (*pos).second;
    if (!SBAD_READPTR(pRASReqInProgress))
    {
        m_RASReqInProgressMap.erase(seqNum);

        if (pRASReqInProgress->m_rPendingReq.GetTag() != reqTag)
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Received message with wrong tag= " << reqTag
                << " (requestTag=" << pRASReqInProgress->m_rPendingReq.GetTag() << ") seqNum=" << seqNum);

            delete pRASReqInProgress;
            return 0;
        }
        
        // okay we found the req in queue AND it's the correct one...
        STRACE2_INFO3(g_pH323Module->m_RasTraceId, this, 0,
            "() Found req tag= " << reqTag
            << " for seqNum=" << seqNum);
        
        return pRASReqInProgress;
    }

    // THIS should never happen...
    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
        "() Queued request '" << reqTag << "' invalid! sequence number " << seqNum );
    
    return 0;
}

void H323Gatekeeper::OnRASReqExpiryCheck( STimer & Timer, int iContinous )
{
    SDEFFUNC("H323Gk::OnRASReqExpiryCheck");
    
    SWaitAndSignal lock(m_GkMutex);
    
    RASReqInProgress *pRASReqInProgress = 0;
    unsigned uRASReqSequenceNumber = 0;

    RASReqInProgressMap::iterator pos = m_RASReqInProgressMap.begin();
    while ( pos != m_RASReqInProgressMap.end() )
    {
        // get values from map element...
        uRASReqSequenceNumber = (*pos).first;
        pRASReqInProgress = (*pos).second;
        // ...and increment iterator (may be reset after erase() later on!)
        ++pos;

        if (!SBAD_READPTR(pRASReqInProgress))
        {
            if ( pRASReqInProgress->m_LastReqAttempt + endpoint.GetRasRequestTimeout() < STimer::Tick() )
            {
                ++pRASReqInProgress->m_NumberOfReqRetry;

                if ( pRASReqInProgress->m_NumberOfReqRetry > endpoint.GetRasRequestRetries() )
                {
                    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0, "() RAS request '"
                        << pRASReqInProgress->m_rPendingReq.GetTagName() << "' , seqNum="
                        << uRASReqSequenceNumber << " failed with timeout!");
                    
                    // max number of retries exceeded -> remove request from list!
                    m_RASReqInProgressMap.erase(uRASReqSequenceNumber);
                    // reset iterator position
                    pos = m_RASReqInProgressMap.begin();
                    // unblock pending request with timeout
                    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RequestTimeout;
                    pRASReqInProgress->m_PendingReqSync.Signal();
                }
                else
                {
                    STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0, "() Timeout: will resent '"
                        << pRASReqInProgress->m_rPendingReq.GetTagName() << "', seq num= "
                        << uRASReqSequenceNumber << ", retry number "
                        << pRASReqInProgress->m_NumberOfReqRetry);
                
                    // resent stored RAS request
                    if (pRASReqInProgress->m_rPendingReq.Write(*transport) == H323RasPDU::Success)
                    {
                        // save new time stamp
                        pRASReqInProgress->m_LastReqAttempt = STimer::Tick();
                    }
                    else // write() failed -> discard RAS request with timeout
                    {
                        m_LastTransportError = transport->GetErrorNumber();
                        
                        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                            "() write to socket failed, error=" << hex << m_LastTransportError );
                        
                        // failed to repeat request due socket error -> remove request from list!
                        m_RASReqInProgressMap.erase(uRASReqSequenceNumber);
                        // reset iterator position
                        pos = m_RASReqInProgressMap.begin();
                        // unblock pending request with timeout
                        pRASReqInProgress->m_RASReqResult = RASReqInProgress::RequestTimeout;
                        pRASReqInProgress->m_PendingReqSync.Signal();
                    }
                }
            }
        }
    }
    
    // start over with next check (except we're already in shutdown phase!) in about a second...
    if ( m_bShutdownGatekeeper == false )
        m_RASReqExpiryCheckTimer = 1000;
    else // we're done!
    {
        STRACE2_INFO(g_pH323Module->m_RasTraceId, this, 0, "() finish RAS request expiry check");
        //DecRef(__FILE__, __LINE__);
    }
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveRequestInProgress(const H225_RequestInProgress & rip)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_requestInProgress, rip.m_requestSeqNum);
    
    if (pRASReqInProgress == 0)
        return false;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RequestInProgress;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::Discover(const H323TransportAddress & address)
{
    SDEFFUNC("H323Gatekeeper::Discover");
    
    H323RasPDU request;
    H225_GatekeeperRequest & grq = request.BuildGatekeeperRequest(GetNextSequenceNumber());
    endpoint.SetEndpointTypeInfo(grq.m_endpointType);
    if (transport != NULL)
        transport->SetUpTransportPDU(grq.m_rasAddress, true);
    else
        return false;
        
    
    if (endpoint.IsTerminal() || endpoint.IsGateway())
    {
        grq.IncludeOptionalField(H225_GatekeeperRequest::e_endpointAlias);
        if (endpoint.GetE164TerminalAlias().empty())
        {
            if (endpoint.GetH323IdTerminalAlias().empty())
            {
                grq.m_endpointAlias.SetSize(1);
                H323SetAliasAddress(endpoint.GetLocalUserName(), grq.m_endpointAlias[0]);
            }
            else
            {
                grq.m_endpointAlias.SetSize(1);
                H323SetAliasAddress(endpoint.GetH323IdTerminalAlias(), grq.m_endpointAlias[0]);
            }
        }
        else
        {
            if (endpoint.GetH323IdTerminalAlias().empty())
            {
                grq.m_endpointAlias.SetSize(1);
                H323SetAliasAddress(endpoint.GetE164TerminalAlias(), grq.m_endpointAlias[0]);
            }
            else
            {
                grq.m_endpointAlias.SetSize(2);
                H323SetAliasAddress(endpoint.GetE164TerminalAlias(), grq.m_endpointAlias[0]);
                H323SetAliasAddress(endpoint.GetH323IdTerminalAlias(), grq.m_endpointAlias[1]);
            }
        }
    }
    
    OnSendGatekeeperRequest(grq);
    
    m_bGatekeeperDiscoveryComplete = false;
    
    m_uLastRASReqTag            = request.GetTag();
    m_uLastRASReqSequenceNumber = grq.m_requestSeqNum;
    
    m_szGatekeeperName = address;
    
    unsigned uGRQRetries = 0;
    while (!transport->DiscoverGatekeeper(*this, request, address))
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() GRQ failed on request " << uGRQRetries);
        
        if (++uGRQRetries > endpoint.GetGatekeeperRequestRetries())
            return false;
    }
    
    if (!m_bGatekeeperDiscoveryComplete)
        return false;
    
    if (transport->Connect())
        transport->AttachThread(new H225RasThread(this, endpoint));
    
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveGatekeeperConfirm(const H225_GatekeeperConfirm & gcf)
{
    SDEFFUNC("H323Gatekeeper::OnReceiveGCF");
    
    if (!CheckForResponse(H225_RasMessage::e_gatekeeperRequest, gcf.m_requestSeqNum))
        return false;
    
    //  RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_gatekeeperRequest, gcf.m_requestSeqNum);
    //  if (pRASReqInProgress == 0)
    //    return false;
    
    m_LocatedAddress = gcf.m_rasAddress;
    m_LastGKAddress  = gcf.m_rasAddress;
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Gatekeeper discovery found " << m_LocatedAddress);
    
    if ( transport->SetRemoteAddress(m_LastGKAddress) == false )
    {
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() Invalid gatekeeper discovery address: \"" << m_LastGKAddress << '"');
    }
    
    m_bGatekeeperDiscoveryComplete = true;
    
    //  pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    //  pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::RegistrationRequest(const Tools::SString & szH323Alias,
                                         bool                   bIsLightWeightRRQ,
                                         int                    iLineId,
                                         int                    iNumOfChannels,
                                         bool                   bCleanupInCaseOfFailure)
{
    SDEFFUNC("H323Gatekeeper::RegistrationReq");
    
    bool bNewClient = false;
    Client * pClient = FindClientByAliasAndLineId( szH323Alias, iLineId );
    if (SBAD_READPTR(pClient))
    {
        Tools::SString szH323AliasForMap = szH323Alias;
        // No client found, add new one.
        m_GkMutex.Wait();
        
        Client::ClientDataStruct cd;
        // fill in structure
        cd.transport   = 0;
        cd.szH323Alias = szH323Alias;
        if ( iLineId > 0 )
        {
            cd.clientType = Client::Gateway;
            cd.Data.GatewayData.lineId = iLineId;

            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Insertion of new gateway client with alias '" << szH323Alias << "' and line id " << iLineId);

            // in the m_Clients map we store those cliens as <alias>|<line id>!
            Tools::SString szLineIdPart;
            szLineIdPart.sprintf( "|%d", iLineId );
            szH323AliasForMap += szLineIdPart;
        }
        else if ( iNumOfChannels > 0 )
        {
            cd.clientType = Client::FaxServer;
            cd.Data.FaxServerData.numOfChannels = iNumOfChannels;

            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Insertion of new fax client with alias '" << szH323Alias << "'");
        }
        else
        {
            cd.clientType = Client::Unknown;

            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Insertion of new client with alias '" << szH323Alias << "'");
        }

        pClient = new Client(this, endpoint, cd);
        if ( pClient == 0 )
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Failed to create client for alias '" << szH323Alias
                << "'!");
            m_GkMutex.Signal();
            return false;
        }
        
        pair<ClientsMap::iterator, bool> Answer;
        Answer = m_Clients.insert(pair<const SString, Client *>(szH323AliasForMap, pClient));
        if (Answer.second == false)
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Insertion of client with alias '" << szH323AliasForMap
                << "' in map failed!");
            m_GkMutex.Signal();
            return false;
        }
        // add reference for map
        bNewClient = true;
        pClient->IncRef(__FILE__, __LINE__);
        // done!
        m_GkMutex.Signal();
    }
    
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Register client with alias '" << szH323Alias
        << (bIsLightWeightRRQ ? "' (lightweight RRQ) ..." : "' (full RRQ) ..."));
    
    H323RasPDU request;
    H225_RegistrationRequest & rrq = request.BuildRegistrationRequest(GetNextSequenceNumber());
    rrq.m_discoveryComplete = m_bGatekeeperDiscoveryComplete;
    m_bGatekeeperDiscoveryComplete = false;
    
    rrq.m_rasAddress.SetSize(1);
    if (transport != NULL)
        transport->SetUpTransportPDU(rrq.m_rasAddress[0], true);
    else
        return false;
    
    const H323ListenerList & listeners = endpoint.GetListeners();
    if (listeners.empty()) {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Cannot register with Gatekeeper without a H323Listener!");
        return false;
    }
    
	for (H323ListenerList::size_type i = 0; i < listeners.size(); i++)
        listeners[i]->SetUpTransportPDU(rrq.m_callSignalAddress, rrq.m_rasAddress[0]);
    
    endpoint.SetEndpointTypeInfo(rrq.m_terminalType);
    endpoint.SetVendorIdentifierInfo(rrq.m_endpointVendor);
    
    if ( !bIsLightWeightRRQ && (endpoint.IsTerminal() || endpoint.IsGateway()))
    {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_terminalAlias);
        
        if (pClient->m_ClientData.szH323Alias.empty())
        {
            // Backward compability
            if (endpoint.GetE164TerminalAlias().empty())
            {
                if (endpoint.GetH323IdTerminalAlias().empty())
                {
                    rrq.m_terminalAlias.SetSize(1);
                    H323SetAliasAddress(endpoint.GetLocalUserName(), rrq.m_terminalAlias[0]);
                }
                else
                {
                    rrq.m_terminalAlias.SetSize(1);
                    H323SetAliasAddress(endpoint.GetH323IdTerminalAlias(), rrq.m_terminalAlias[0]);
                }
            }
            else
            {
                if (endpoint.GetH323IdTerminalAlias().empty())
                {
                    rrq.m_terminalAlias.SetSize(1);
                    H323SetAliasAddress(endpoint.GetE164TerminalAlias(), rrq.m_terminalAlias[0]);
                }
                else
                {
                    rrq.m_terminalAlias.SetSize(2);
                    H323SetAliasAddress(endpoint.GetE164TerminalAlias(), rrq.m_terminalAlias[0]);
                    H323SetAliasAddress(endpoint.GetH323IdTerminalAlias(), rrq.m_terminalAlias[1]);
                }
            }
        }
        else
        {
            rrq.m_terminalAlias.SetSize(1);
            H323SetAliasAddress(szH323Alias, rrq.m_terminalAlias[0]);
        }
    }
    
    if (!m_szGatekeeperIdentifier.empty()) {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_gatekeeperIdentifier);
        rrq.m_gatekeeperIdentifier = m_szGatekeeperIdentifier;
    }
    
    if (!pClient->m_szEndpointIdentifier.empty()) {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_endpointIdentifier);
        rrq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    }
    
    if ( endpoint.GetRegistrationTTL() > 0 )
    {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_timeToLive);
        rrq.m_timeToLive = endpoint.GetRegistrationTTL(); // set RRQ time to live
    }
    rrq.m_keepAlive = bIsLightWeightRRQ; // true = keep alive with light weight RRQ

    if ( pClient->m_ClientData.clientType == Client::Gateway   ||
         pClient->m_ClientData.clientType == Client::FaxServer   )
    {
        // create Swyx extension structure...
        SwyxH225NonStandardData nsData(g_pH323Module->m_RasTraceId);
        if ( pClient->m_ClientData.clientType == Client::Gateway )
        {
            nsData.m_NSDataChoice = SwyxH225NonStandardData::CallInfo;
            nsData.m_uFaxType     = SwyxH225NonStandardData::UnknownFT;
            nsData.m_uLineId      = pClient->m_ClientData.Data.GatewayData.lineId;
        }
        else if ( pClient->m_ClientData.clientType == Client::FaxServer )
        {
            nsData.m_NSDataChoice = SwyxH225NonStandardData::FaxChannels;
            nsData.m_uFaxChannels = pClient->m_ClientData.Data.FaxServerData.numOfChannels;
        }

        // build and attach NonStandardData to RRQ signal pdu
		SPerStream ExtStrm;
        nsData.BuildSwH225NonStandardData(ExtStrm);
        AttachRRQNonStandardData(ExtStrm, rrq);
    }

    // while RRQ we have to match the RCF/RRJ to client by the sequence number
    pClient->m_uLastReqSequenceNumber = rrq.m_requestSeqNum;

    OnSendRegistrationRequest(rrq);

    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, rrq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if ( pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived )
            {
                bResult = true;
            }
            else if ( pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                STRACE2_INFO(g_pH323Module->m_RasTraceId, this, 0,
                    "() RRQ rejected!");
            }
            else
            {
                STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
                    "() RRQ NOT confirmed!");
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending RRQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    // are we successful on RRQ?
    if ( bResult == false )
    {
        // something failed! -> erase NEW client from map
        if ( bNewClient == true )
        {
            m_GkMutex.Wait();
            ClientsMap::iterator iter;
            for (iter = m_Clients.begin(); iter != m_Clients.end(); ++iter)
            {
                if ((*iter).second == pClient)
                {
                    m_Clients.erase(iter);
                    // remove reference for map
                    pClient->DecRef(__FILE__, __LINE__);
                    pClient = 0;
                    break;
                }
            }
            m_GkMutex.Signal();
        }
        // free all communication ressources if requested!
        if ( bCleanupInCaseOfFailure == true )
        {
            if ( m_pGKRegisterThread )
            {
                delete m_pGKRegisterThread;
                m_pGKRegisterThread = 0;
            }

            // cleanup RAS UDP transport
            if (transport != NULL)
            {
                transport->CleanUpOnTermination();
                m_GkMutex.Wait();
                delete transport;
                transport = 0;
                m_GkMutex.Signal();
            }
        }
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveRegistrationConfirm(const H225_RegistrationConfirm & rcf)
{
    SDEFFUNC("H323Gatekeeper::OnReceiveRCF");
    
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_registrationRequest, rcf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    SwyxH225NonStandardData nsData(g_pH323Module->m_RasTraceId);
    if (rcf.HasOptionalField(H225_RegistrationConfirm::e_nonStandardData))
    {
		SPerStream ExtStrm;
        // get NonStandardData out of RCF
        GetRCFNonStandardData(rcf, ExtStrm);
        // Handle NonStandardData
        nsData.HandleSwH225NonStandardData(ExtStrm);
    }
    // Search for client of this RCF
    Client * pClient = 0;
    // 1. Search for client of RCF terminalAlias if present
    if (rcf.HasOptionalField(H225_RegistrationConfirm::e_terminalAlias))
    {
        SStringVector H323Aliases;
        H323GetAliasAddressStringVector(rcf.m_terminalAlias,
            H225_AliasAddress::e_h323_ID,
            H323Aliases);
        if (!H323Aliases.empty())
        {
			for(SStringVector::size_type ii = 0; ii < H323Aliases.size(); ii++)
            {
                // in case if SwyxGate concatinate alias and line id:
                SString szH323Alias;
                if ( nsData.m_NSDataChoice == SwyxH225NonStandardData::CallInfo )
                    pClient = FindClientByAliasAndLineId(H323Aliases[ii], nsData.m_uLineId);
                else
                    pClient = FindClientByAlias(H323Aliases[ii]);
                if (pClient != 0)
                    goto EndpointFound;
            }
        }
    }
    // 2. Search for client of RCF endpointIdentifier
    pClient = FindClientByEndpointId(rcf.m_endpointIdentifier);
    if (pClient == 0)
    {
        // No client found
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client for this RCF found!");
        
        pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
        pRASReqInProgress->m_PendingReqSync.Signal();
        return false;
    }
    
EndpointFound:
    
    if (rcf.HasOptionalField(H225_RegistrationConfirm::e_gatekeeperIdentifier))
    {
        m_szGatekeeperIdentifier = (SString)rcf.m_gatekeeperIdentifier;
    }
    pClient->m_szEndpointIdentifier = (SString)rcf.m_endpointIdentifier;

    // client previosly unregistered?
    if (   pClient->m_GKRegStatus == gk_RegistrationStatusUnknown
        || pClient->m_GKRegStatus == gk_RegistrationStatusUnregistered
        || pClient->m_GKRegStatus == gk_RegistrationStatusRegistrationLost )
    {
        // reset flag
        if ( pClient->m_ClientData.clientType == Client::Gateway )
        {
            // do all necessary correct action in endpoint...
            endpoint.OnGatekeeperRegistrationDone( reinterpret_cast<int>(pClient) );
        }
    }
    
    pClient->m_GKRegStatus          = gk_RegistrationStatusOK;
    pClient->m_uNumberOfRRQAttempts = 0;
    
    // TTL value present?
    if (rcf.HasOptionalField(H225_RegistrationConfirm::e_timeToLive))
        pClient->m_RegTimer = STimeInterval(0, (long)(rcf.m_timeToLive * 0.75)); // sent keep alive RRQ after 75% expiration time
    else // nope! -> keep this registration for a very looooong time ;-)
        pClient->m_RegTimer = SMaxTimeInterval;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveRegistrationReject(const H225_RegistrationReject & rrj)
{
    SDEFFUNC("H323Gatekeeper::OnReceiveRRJ");
    
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_registrationRequest, rrj.m_requestSeqNum);
    
    if (pRASReqInProgress == 0)
        return false;
    
    SwyxH225NonStandardData nsData(g_pH323Module->m_RasTraceId);
    if (rrj.HasOptionalField(H225_RegistrationReject::e_nonStandardData))
    {
		SPerStream ExtStrm;
        // get NonStandardData out of RRJ
        GetRRJNonStandardData(rrj, ExtStrm);
        // Handle NonStandardData
        nsData.HandleSwH225NonStandardData(ExtStrm);
    }
    // Search for client of this RRJ
    Client * pClient = FindClientBySeqNum(rrj.m_requestSeqNum);
    if (pClient == 0)
    {
        // No client found
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client for this RRJ found!");
        
        pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived;
        pRASReqInProgress->m_PendingReqSync.Signal();
        return false;
    }

    STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
        "() RRQ for client with alias '" << pClient->m_ClientData.szH323Alias
        << "' rejected!");

    // are we registered before? -> only try reregistering after reject if we were registered before!
    if ( pClient->m_GKRegStatus == gk_RegistrationStatusInKeepAlive )
    {    
        // chance status in case of reason==resourceUnavailable which is used by the server
        // gatekeeper to indicate insufficient voice channel licenses
        if (rrj.m_rejectReason.GetTag() == H225_RegistrationRejectReason::e_resourceUnavailable)
            pClient->m_GKRegStatus = gk_RegistrationStatusNotPermitted;
        else
            pClient->m_GKRegStatus = gk_RegistrationStatusRegistrationLost;

        pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived;
        pRASReqInProgress->m_PendingReqSync.Signal();
        // perform all necessary actions to reestablish GK connection in endpoint...
        endpoint.OnGatekeeperRegistrationLost( reinterpret_cast<int>(pClient) );
        return true;
    }

    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::UnregistrationRequest(const Tools::SString & szH323Alias,
                                           int                    iLineId, // used by SwyxGate
                                           int                    reason)
{
    SDEFFUNC("H323Gatekeeper::UnregistrationReq");
    
    Tools::SString szEndpointIdentifier;
    
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No alias -> discard request!");
        return false;
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Unregister client with alias '" << szH323Alias << "' ...");

    Client * pClient = FindClientByAliasAndLineId( szH323Alias, iLineId );
    if (SBAD_READPTR(pClient))
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "' (URQ)");
        return false;
    }
    
    // stop registration timer for this client in any case!
    pClient->m_RegTimer.Stop();

    szEndpointIdentifier = pClient->m_szEndpointIdentifier;
    
    H323RasPDU request;
    H225_UnregistrationRequest & urq = request.BuildUnregistrationRequest(GetNextSequenceNumber());
    
    H225_TransportAddress rasAddress;
    if (transport != NULL)
        transport->SetUpTransportPDU(rasAddress, true);
    else
        return false;
    
    const H323ListenerList & listeners = endpoint.GetListeners();
	for (H323ListenerList::size_type i = 0; i < listeners.size(); i++)
        listeners[i]->SetUpTransportPDU(urq.m_callSignalAddress, rasAddress);
    
    if (endpoint.IsTerminal() || endpoint.IsGateway())
    {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_endpointAlias);
        if (!szH323Alias.empty())
        {
            urq.m_endpointAlias.SetSize(1);
            H323SetAliasAddress(szH323Alias, urq.m_endpointAlias[0]);
        }
    }
    
    if (!m_szGatekeeperIdentifier.empty()) {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_gatekeeperIdentifier);
        urq.m_gatekeeperIdentifier = m_szGatekeeperIdentifier;
    }
    
    if (!szEndpointIdentifier.empty())
    {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_endpointIdentifier);
        urq.m_endpointIdentifier = szEndpointIdentifier;
    }
    
    if (reason >= 0) {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_reason);
        urq.m_reason = reason;
    }

    if ( pClient->m_ClientData.clientType == Client::Gateway )
    {
        // create Swyx extension structure...
        SwyxH225NonStandardData nsData(g_pH323Module->m_RasTraceId);
        nsData.m_NSDataChoice = SwyxH225NonStandardData::CallInfo;
        nsData.m_uFaxType     = SwyxH225NonStandardData::UnknownFT;
        nsData.m_uLineId      = pClient->m_ClientData.Data.GatewayData.lineId;

        // build and attach NonStandardData to RRQ signal pdu
		SPerStream ExtStrm;
        nsData.BuildSwH225NonStandardData(ExtStrm);
        AttachURQNonStandardData(ExtStrm, urq);
    }
    
    OnSendUnregistrationRequest(urq);

    // while URQ we have to match the UCF to client by the sequence number
    pClient->m_uLastReqSequenceNumber = urq.m_requestSeqNum;
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, urq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                bResult = true;
            }
            else
            {
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() URQ failed!");
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending URQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveUnregistrationConfirm(const H225_UnregistrationConfirm & ucf)
{
    SDEFFUNC("H323Gatekeeper::OnReceiveUCF");
    
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_unregistrationRequest, ucf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    // Search for client of this UCF
    Client * pClient = 0;
    pClient = FindClientBySeqNum(ucf.m_requestSeqNum);
    if (pClient == 0)
    {
        // No client found
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client for this UCF found!");
        
        pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
        pRASReqInProgress->m_PendingReqSync.Signal();
        return false;
    }
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() URQ successful -> EP is now unregistered!");
    
    pClient->m_GKRegStatus = gk_RegistrationStatusUnregistered;
    pClient->m_szEndpointIdentifier.clear();
    pClient->m_uLastReqSequenceNumber = 0;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveUnregistrationRequest(const H225_UnregistrationRequest & urq)
{
    SDEFFUNC("H323Gatekeeper::OnReceiveURQ");
    
    Client * pClient = 0;
    // 1. Search for client of URQ endpointIdentifier if present
    if (urq.HasOptionalField(H225_UnregistrationRequest::e_endpointIdentifier))
    {
        Tools::SString endpointId = urq.m_endpointIdentifier;
        if (!endpointId.empty())
        {
            pClient = FindClientByEndpointId(urq.m_endpointIdentifier);
        }
    }
    // 2. Search for client of URQ alias and NonStandard data if present
    if (SBAD_READPTR(pClient))
    {
        SwyxH225NonStandardData nsData(g_pH323Module->m_RasTraceId);
        if (urq.HasOptionalField(H225_UnregistrationRequest::e_nonStandardData))
        {
		    SPerStream ExtStrm;
            // get NonStandardData out of RCF
            GetURQNonStandardData(urq, ExtStrm);
            // Handle NonStandardData
            nsData.HandleSwH225NonStandardData(ExtStrm);

            SStringVector H323Aliases;
            H323GetAliasAddressStringVector(urq.m_endpointAlias,
                H225_AliasAddress::e_h323_ID,
                H323Aliases);
            if (!H323Aliases.empty())
            {
			    for(SStringVector::size_type ii = 0; ii < H323Aliases.size(); ii++)
                {
                    // in case if SwyxGate concatinate alias and line id:
                    SString szH323Alias;
                    if ( nsData.m_NSDataChoice == SwyxH225NonStandardData::CallInfo )
                        pClient = FindClientByAliasAndLineId(H323Aliases[ii], nsData.m_uLineId);
                    else
                        pClient = FindClientByAlias(H323Aliases[ii]);
                }
            }
        }
        if (SBAD_READPTR(pClient))
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() No client found for URQ!");
            return false;
        }
    }
    
    STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
        "() URQ received -> Endpoint '" << pClient->m_ClientData.szH323Alias << "' is now unregistered!");
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS UnregistrationConfirm to " << transport->GetRemoteAddressString());
    
    H323RasPDU response;
    response.BuildUnregistrationConfirm(urq.m_requestSeqNum);
    if ( response.Write(*transport) == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() Write RAS UnregistrationConfirm to " << transport->GetRemoteAddressString() << " failed with error= " << transport->GetErrorNumber());
    }
    
    // cleanup client data
    pClient->m_uLastReqSequenceNumber = 0;
    pClient->m_GKRegStatus = gk_RegistrationStatusRegistrationLost;

    // perform all necessary actions to reestablish GK connection in endpoint...
    endpoint.OnGatekeeperRegistrationLost( reinterpret_cast<int>(pClient) );

    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::LocationRequest(const SString & alias,
                                     H323TransportAddress & address)
{
    SStringList aliases;
    aliases.push_back(alias);
    return LocationRequest(aliases, address);
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::LocationRequest(const SStringList & aliases,
                                     H323TransportAddress & address)
{
    SDEFFUNC("H323Gatekeeper::LocationRequest");
    
    H323RasPDU request;
    H225_LocationRequest & lrq = request.BuildLocationRequest(GetNextSequenceNumber());
    
    lrq.m_destinationInfo.SetSize(aliases.size());
    int i;
    SStringList::const_iterator pos;
    for (pos  = aliases.begin(), i = 0; pos != aliases.end(); ++pos, ++i)
    {
        H323SetAliasAddress((*pos), lrq.m_destinationInfo[i]);
    }

    OnSendLocationRequest(lrq);
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, lrq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                address = m_LocatedAddress;

                bResult = true;
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending LRQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveLocationConfirm(const H225_LocationConfirm & lcf)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_locationRequest, lcf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    m_LocatedAddress = lcf.m_callSignalAddress;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::AdmissionRequest(H323Connection & connection)
{
    H323TransportAddress address;
    return AdmissionRequest(connection, address);
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::AdmissionRequest(H323Connection & connection,
                                      H323TransportAddress & address)
{
    SDEFFUNC("H323Gatekeeper::AdmissionRequest");
    
    Tools::SString szH323Alias = connection.GetH323Alias();
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No H.323 alias found for connection!");
        return false;
    }
    // Check for special SwyxGate line alias
    long iLineId = 0;
    SStringVector AliasParts = szH323Alias.Tokenize("|");
    if (AliasParts.size() == 2)
    {
        // Found line alias
        szH323Alias = AliasParts[0];
        iLineId = AliasParts[1].AsInteger();
    }
    
    Client * pClient = FindClientByAliasAndLineId(szH323Alias, iLineId);
    if (pClient == 0)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "'!");
        return false;
    }
    
    if (!pClient->IsRegistered())
    {
        STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending AdmissionRequest failed because endpoint is not registered!");
        return false;
    }
    
    H323RasPDU request;
    H225_AdmissionRequest & arq = request.BuildAdmissionRequest(GetNextSequenceNumber());
    
    
    arq.m_callType.SetTag(H225_CallType::e_pointToPoint);
    arq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    
    bool answeringCall = connection.HadAnsweredCall();
    arq.m_answerCall = answeringCall;
    
    arq.m_srcInfo.SetSize(1);
    if (answeringCall)
        H323SetAliasAddress(connection.GetRemotePartyName(), arq.m_srcInfo[0]);
    else {
        H323SetAliasAddress(pClient->m_ClientData.szH323Alias, arq.m_srcInfo[0]);
        if (!connection.GetRemotePartyName().empty())
        {
            arq.IncludeOptionalField(H225_AdmissionRequest::e_destinationInfo);
            arq.m_destinationInfo.SetSize(1);
            H323SetAliasAddress(connection.GetRemotePartyName(), arq.m_destinationInfo[0]);
        }
    }
    
    const H323Transport * signallingChannel = connection.GetSignallingChannel();
    arq.IncludeOptionalField(H225_AdmissionRequest::e_srcCallSignalAddress);
    signallingChannel->SetUpTransportPDU(arq.m_srcCallSignalAddress, !answeringCall);
    arq.IncludeOptionalField(H225_AdmissionRequest::e_destCallSignalAddress);
    signallingChannel->SetUpTransportPDU(arq.m_destCallSignalAddress, answeringCall);
    
    arq.m_bandWidth = connection.GetBandwidthAvailable();
    arq.m_callReferenceValue = connection.GetCallReference();
    arq.m_conferenceID = connection.GetConferenceIdentifier();
    arq.m_callIdentifier.m_guid = connection.GetCallIdentifier();
    
    OnSendAdmissionRequest(arq);
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, arq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                connection.SetBandwidthAvailable(m_uAllocatedBandwidth);
                address = m_LocatedAddress;

                bResult = true;
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending ARQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveAdmissionConfirm(const H225_AdmissionConfirm & acf)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_admissionRequest, acf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    if (acf.m_callModel.GetTag() == H225_CallModel::e_direct)
        m_LocatedAddress = "";
    else
        m_LocatedAddress = acf.m_destCallSignalAddress;
    
    m_uAllocatedBandwidth = acf.m_bandWidth;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::DisengageRequest(const H323Connection & connection, unsigned reason)
{
    SDEFFUNC("H323Gatekeeper::DisengageRequest");
    
    Tools::SString szH323Alias = connection.GetH323Alias();
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() No H.323 alias found for connection!");
        return false;
    }
    // Check for special SwyxGate line alias
    long iLineId = 0;
    SStringVector AliasParts = szH323Alias.Tokenize("|");
    if (AliasParts.size() == 2)
    {
        // Found line alias
        szH323Alias = AliasParts[0];
        iLineId = AliasParts[1].AsInteger();
    }

    Client * pClient = FindClientByAliasAndLineId(szH323Alias, iLineId);
    if (pClient == 0)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "'!");
        return false;
    }
    if (!pClient->IsRegistered())
    {
        STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending DisengageRequest failed because endpoint is not registered!");
        return false;
    }
    
    H323RasPDU request;
    H225_DisengageRequest & drq = request.BuildDisengageRequest(GetNextSequenceNumber());
    
    drq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    drq.m_conferenceID = connection.GetConferenceIdentifier();
    drq.m_callReferenceValue = connection.GetCallReference();
    drq.m_callIdentifier.m_guid = connection.GetCallIdentifier();
    drq.m_disengageReason.SetTag(reason);
    drq.m_answeredCall = connection.HadAnsweredCall();
    
    if (!m_szGatekeeperIdentifier.empty()) {
        drq.IncludeOptionalField(H225_DisengageRequest::e_gatekeeperIdentifier);
        drq.m_gatekeeperIdentifier = m_szGatekeeperIdentifier;
    }
    
    OnSendDisengageRequest(drq);
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, drq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                bResult = true;

            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending DRQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveDisengageConfirm(const H225_DisengageConfirm & dcf)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_disengageRequest, dcf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveDisengageRequest(const H225_DisengageRequest & drq)
{
    SDEFFUNC("H323Gatekeeper::OnReceiveDRQ");
    
    H323GloballyUniqueID id = drq.m_callIdentifier.m_guid;
    endpoint.ClearCall(id.AsString(), H323Connection::GatekeeperForcedTermination);
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS DisengageConfirm to " << transport->GetRemoteAddressString());
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;
    response.BuildDisengageConfirm(drq.m_requestSeqNum);
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::BandwidthRequest(H323Connection & connection,
                                      unsigned requestedBandwidth)
{
    SDEFFUNC("H323Gatekeeper::BandwidthRequest");
    
    Tools::SString szH323Alias = connection.GetH323Alias();
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() No H.323 alias found for connection!");
        return false;
    }
    Client * pClient = FindClientByAlias(szH323Alias);
    if (pClient == 0)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "'!");
        return false;
    }
    if (!pClient->IsRegistered())
    {
        STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending BandwidthRequest failed because endpoint is not registered!");
        return false;
    }
    
    H323RasPDU request;
    H225_BandwidthRequest & brq = request.BuildBandwidthRequest(GetNextSequenceNumber());
    
    brq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    brq.m_conferenceID = connection.GetConferenceIdentifier();
    brq.m_callReferenceValue = connection.GetCallReference();
    brq.m_callIdentifier.m_guid = connection.GetCallIdentifier();
    brq.m_bandWidth = requestedBandwidth;
    
    OnSendBandwidthRequest(brq);
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, brq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                connection.SetBandwidthAvailable(m_uAllocatedBandwidth);

                bResult = true;
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending BRQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveBandwidthConfirm(const H225_BandwidthConfirm & bcf)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_bandwidthRequest, bcf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    m_uAllocatedBandwidth = bcf.m_bandWidth;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveBandwidthRequest(const H225_BandwidthRequest & brq)
{
    SDEFFUNC("H323Gatekeeper::OnReceiveBRQ");
    
    H323GloballyUniqueID id = brq.m_callIdentifier.m_guid;
    H323Connection * connection = endpoint.FindConnectionWithLock(id.AsString());
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;

    if (connection == NULL)
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS BandwidthReject to " << transport->GetRemoteAddressString());
        response.BuildBandwidthReject(brq.m_requestSeqNum,
            H225_BandRejectReason::e_invalidConferenceID);
    }
    else
    {
        if (connection->SetBandwidthAvailable(brq.m_bandWidth))
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() SEND RAS BandwidthConfirm to " << transport->GetRemoteAddressString());
            response.BuildBandwidthConfirm(brq.m_requestSeqNum, brq.m_bandWidth);
        }
        else
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() SEND RAS BandwidthReject to " << transport->GetRemoteAddressString());
            response.BuildBandwidthReject(brq.m_requestSeqNum,
                H225_BandRejectReason::e_insufficientResources);
        }
        connection->Unlock();
    }
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::InfoRequestResponse(const H323Connection * connection,
                                         unsigned seqNum)
{
    SDEFFUNC("H323Gatekeeper::InfoRequestResp");
    
    Tools::SString szH323Alias = connection->GetH323Alias();
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() No H.323 alias found for connection!");
        return false;
    }
    Client * pClient = FindClientByAlias(szH323Alias);
    if (pClient == 0)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "'!");
        return false;
    }
    if (!pClient->IsRegistered())
    {
        STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending InfoRequestResponse failed because endpoint is not registered!");
        return false;
    }
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;
    
    H225_InfoRequestResponse & irr = response.BuildInfoRequestResponse(seqNum);
    
    endpoint.SetEndpointTypeInfo(irr.m_endpointType);
    irr.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    transport->SetUpTransportPDU(irr.m_rasAddress, true);
    const H323ListenerList & listeners = endpoint.GetListeners();
	for (H323ListenerList::size_type i = 0; i < listeners.size(); i++)
        listeners[i]->SetUpTransportPDU(irr.m_callSignalAddress, irr.m_rasAddress);
    
    if (endpoint.IsTerminal()) {
        irr.IncludeOptionalField(H225_InfoRequestResponse::e_endpointAlias);
        irr.m_endpointAlias.SetSize(1);
        H323SetAliasAddress(endpoint.GetLocalUserName(), irr.m_endpointAlias[0]);
    }
    
    if (connection != NULL) {
        irr.m_perCallInfo.SetSize(1);
        H225_InfoRequestResponse_perCallInfo_subtype & info = irr.m_perCallInfo[0];
        info.m_callReferenceValue = connection->GetCallReference();
        info.m_callIdentifier.m_guid = connection->GetCallIdentifier();
        info.m_conferenceID = connection->GetConferenceIdentifier();
        RTP_Session * session = connection->GetSession(RTP_Session::DefaultAudioSessionID);
        if (session != NULL) {
            info.IncludeOptionalField(H225_InfoRequestResponse_perCallInfo_subtype::e_audio);
            info.m_audio.SetSize(1);
            session->OnSendRasInfo(info.m_audio[0]);
        }
        session = connection->GetSession(RTP_Session::DefaultVideoSessionID);
        if (session != NULL) {
            info.IncludeOptionalField(H225_InfoRequestResponse_perCallInfo_subtype::e_video);
            info.m_video.SetSize(1);
            session->OnSendRasInfo(info.m_video[0]);
        }
        const H323Transport & controlChannel = connection->GetControlChannel();
        controlChannel.SetUpTransportPDU(info.m_h245.m_recvAddress, true);
        controlChannel.SetUpTransportPDU(info.m_h245.m_sendAddress, false);
        info.m_callType.SetTag(H225_CallType::e_pointToPoint);
        info.m_bandWidth = connection->GetBandwidthUsed();
        info.m_callModel.SetTag(H225_CallModel::e_direct);
    }
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS InfoRequestResponse to " << transport->GetRemoteAddressString());
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveInfoRequest(const H225_InfoRequest & irq)
{
    H323GloballyUniqueID id = irq.m_callIdentifier.m_guid;
    H323Connection * connection = endpoint.FindConnectionWithLock(id.AsString());
    
    InfoRequestResponse(connection, irq.m_requestSeqNum);
    if (connection != NULL)
        connection->Unlock();
    
    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::AttachRRQNonStandardData(const Tools::SByteVector & data,
                                              H225_RegistrationRequest & rrq)
{
    SDEFFUNC("H323Gatekeeper::AttachRRQNSData");

    rrq.IncludeOptionalField(H225_RegistrationRequest::e_nonStandardData );

    if (!m_szObjectIdentifier.empty())
    {
        // Connection is using object identifier
        rrq.m_nonStandardData.m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_object);
        ((SAsnObjectId &)rrq.m_nonStandardData.m_nonStandardIdentifier) = m_szObjectIdentifier;
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        rrq.m_nonStandardData.m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_h221NonStandard);
        H225_H221NonStandard & h221 = rrq.m_nonStandardData.m_nonStandardIdentifier;
        endpoint.SetH221NonStandardInfo(h221);
    }

    // Adding nonStandard data
    rrq.m_nonStandardData.m_data.SetValue(data);

    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::AttachURQNonStandardData(const Tools::SByteVector   & data,
                                              H225_UnregistrationRequest & urq)
{
    SDEFFUNC("H323Gatekeeper::AttachURQNSData");

    urq.IncludeOptionalField(H225_UnregistrationRequest::e_nonStandardData );

    if (!m_szObjectIdentifier.empty())
    {
        // Connection is using object identifier
        urq.m_nonStandardData.m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_object);
        ((SAsnObjectId &)urq.m_nonStandardData.m_nonStandardIdentifier) = m_szObjectIdentifier;
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        urq.m_nonStandardData.m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_h221NonStandard);
        H225_H221NonStandard & h221 = urq.m_nonStandardData.m_nonStandardIdentifier;
        endpoint.SetH221NonStandardInfo(h221);
    }

    // Adding nonStandard data
    urq.m_nonStandardData.m_data.SetValue(data);

    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::GetRCFNonStandardData(const H225_RegistrationConfirm & rcf,
                                           Tools::SByteVector             & data)
{
    SDEFFUNC("H323Gatekeeper::GetRCFNSData");

    // Get NonStandardData part of RCF
    if (!rcf.HasOptionalField(H225_RegistrationConfirm::e_nonStandardData))
        return false;

    const H225_NonStandardIdentifier & nonStdId
        = rcf.m_nonStandardData.m_nonStandardIdentifier;

    if (!m_szObjectIdentifier.empty())
    {
        // Connection is using object identifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_object)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using object identifier "
                         << m_szObjectIdentifier);
            return false;
        }
        const SAsnObjectId & oid = nonStdId;
        if (oid != m_szObjectIdentifier.c_str())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received object identifier " << oid
                         << " != " << m_szObjectIdentifier);
            return false;
        }
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_h221NonStandard)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using h221NonStandard identifier!");
            return false;
        }
        const H225_H221NonStandard & h221Msg = nonStdId;

        // Get endpoints H.221 NonStandardIdentifier
        H225_H221NonStandard h221Endpoint;
        endpoint.SetH221NonStandardInfo(h221Endpoint);

        // Compare H.221 NonStandardIdentifier
        if (h221Msg.m_t35CountryCode.GetValue() != h221Endpoint.m_t35CountryCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35CountryCode="
                         << h221Msg.m_t35CountryCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35CountryCode.GetValue());
            return false;
        }
        if (h221Msg.m_t35Extension.GetValue() != h221Endpoint.m_t35Extension.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35Extension="
                         << h221Msg.m_t35Extension.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35Extension.GetValue());
            return false;
        }
        if (h221Msg.m_manufacturerCode.GetValue() != h221Endpoint.m_manufacturerCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received manufacturerCode="
                         << h221Msg.m_manufacturerCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_manufacturerCode.GetValue());
            return false;
        }
    }

    // Get nonStandard data
    data.clear();
    data = rcf.m_nonStandardData.m_data.GetValue();

    return true;
}


//-----------------------------------------------------------------------------

bool H323Gatekeeper::GetRRJNonStandardData(const H225_RegistrationReject & rrj,
                                           Tools::SByteVector            & data)
{
    SDEFFUNC("H323Gatekeeper::GetRRJNSData");

    // Get NonStandardData part of RRJ
    if (!rrj.HasOptionalField(H225_RegistrationReject::e_nonStandardData))
        return false;

    const H225_NonStandardIdentifier & nonStdId
        = rrj.m_nonStandardData.m_nonStandardIdentifier;

    if (!m_szObjectIdentifier.empty())
    {
        // Connection is using object identifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_object)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using object identifier "
                         << m_szObjectIdentifier);
            return false;
        }
        const SAsnObjectId & oid = nonStdId;
        if (oid != m_szObjectIdentifier.c_str())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received object identifier " << oid
                         << " != " << m_szObjectIdentifier);
            return false;
        }
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_h221NonStandard)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using h221NonStandard identifier!");
            return false;
        }
        const H225_H221NonStandard & h221Msg = nonStdId;

        // Get endpoints H.221 NonStandardIdentifier
        H225_H221NonStandard h221Endpoint;
        endpoint.SetH221NonStandardInfo(h221Endpoint);

        // Compare H.221 NonStandardIdentifier
        if (h221Msg.m_t35CountryCode.GetValue() != h221Endpoint.m_t35CountryCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35CountryCode="
                         << h221Msg.m_t35CountryCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35CountryCode.GetValue());
            return false;
        }
        if (h221Msg.m_t35Extension.GetValue() != h221Endpoint.m_t35Extension.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35Extension="
                         << h221Msg.m_t35Extension.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35Extension.GetValue());
            return false;
        }
        if (h221Msg.m_manufacturerCode.GetValue() != h221Endpoint.m_manufacturerCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received manufacturerCode="
                         << h221Msg.m_manufacturerCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_manufacturerCode.GetValue());
            return false;
        }
    }

    // Get nonStandard data
    data.clear();
    data = rrj.m_nonStandardData.m_data.GetValue();

    return true;
}

//-----------------------------------------------------------------------------

bool H323Gatekeeper::GetURQNonStandardData(const H225_UnregistrationRequest & urq,
                                           Tools::SByteVector               & data)
{
    SDEFFUNC("H323Gatekeeper::GetURQNSData");

    // Get NonStandardData part of RCF
    if (!urq.HasOptionalField(H225_UnregistrationRequest::e_nonStandardData))
        return false;

    const H225_NonStandardIdentifier & nonStdId
        = urq.m_nonStandardData.m_nonStandardIdentifier;

    if (!m_szObjectIdentifier.empty())
    {
        // Connection is using object identifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_object)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using object identifier "
                         << m_szObjectIdentifier);
            return false;
        }
        const SAsnObjectId & oid = nonStdId;
        if (oid != m_szObjectIdentifier.c_str())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received object identifier " << oid
                         << " != " << m_szObjectIdentifier);
            return false;
        }
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_h221NonStandard)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using h221NonStandard identifier!");
            return false;
        }
        const H225_H221NonStandard & h221Msg = nonStdId;

        // Get endpoints H.221 NonStandardIdentifier
        H225_H221NonStandard h221Endpoint;
        endpoint.SetH221NonStandardInfo(h221Endpoint);

        // Compare H.221 NonStandardIdentifier
        if (h221Msg.m_t35CountryCode.GetValue() != h221Endpoint.m_t35CountryCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35CountryCode="
                         << h221Msg.m_t35CountryCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35CountryCode.GetValue());
            return false;
        }
        if (h221Msg.m_t35Extension.GetValue() != h221Endpoint.m_t35Extension.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35Extension="
                         << h221Msg.m_t35Extension.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35Extension.GetValue());
            return false;
        }
        if (h221Msg.m_manufacturerCode.GetValue() != h221Endpoint.m_manufacturerCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received manufacturerCode="
                         << h221Msg.m_manufacturerCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_manufacturerCode.GetValue());
            return false;
        }
    }

    // Get nonStandard data
    data.clear();
    data = urq.m_nonStandardData.m_data.GetValue();

    return true;
}


//-----------------------------------------------------------------------------

#define DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
void H323Gatekeeper::OnSend##name(H225_##name &) { }

#define DEFINE_GATEKEEPER_PROTOCOL_REQUEST(name) \
    DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
bool H323Gatekeeper::OnReceive##name(const H225_##name &) { return false; }

#define DEFINE_GATEKEEPER_PROTOCOL_REJECT(name, req) \
    DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
    bool H323Gatekeeper::OnReceive##name(const H225_##name & pdu) { \
    SDEFFUNC("H323Gatekeeper::OnReceive"#name); \
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_##req, pdu.m_requestSeqNum); \
    if (pRASReqInProgress == 0) return false; \
    STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0, "() Reveived " << #name ", reason: " << pdu.m_rejectReason.GetTagName()); \
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived; \
    pRASReqInProgress->m_PendingReqSync.Signal(); \
    return true; \
}

#define DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(name, req) \
    DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
    bool H323Gatekeeper::OnReceive##name(const H225_##name & pdu) { \
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_##req, pdu.m_requestSeqNum); \
    if (pRASReqInProgress == 0) return false; \
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived; \
    pRASReqInProgress->m_PendingReqSync.Signal(); \
    return true; \
}


DEFINE_GATEKEEPER_PROTOCOL_REQUEST(GatekeeperRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(GatekeeperConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(GatekeeperReject, gatekeeperRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(RegistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(RegistrationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_SEND(RegistrationReject);
//DEFINE_GATEKEEPER_PROTOCOL_REJECT(RegistrationReject, registrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(UnregistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(UnregistrationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(UnregistrationReject, unregistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(AdmissionRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(AdmissionConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(AdmissionReject, admissionRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(BandwidthRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(BandwidthConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(BandwidthReject, bandwidthRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(DisengageRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(DisengageConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(DisengageReject, disengageRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(LocationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(LocationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(LocationReject, locationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(InfoRequest);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestResponse, infoRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(NonStandardMessage);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(UnknownMessageResponse);
DEFINE_GATEKEEPER_PROTOCOL_SEND(RequestInProgress);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(ResourcesAvailableIndicate);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(ResourcesAvailableConfirm, resourcesAvailableIndicate);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestAck, infoRequest);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestNak, infoRequest);

//-----------------------------------------------------------------------------

bool H323Gatekeeper::OnReceiveUnkown(const H323RasPDU &)
{
    SDEFFUNC("H323Gatekeeper::OnReceiveUnkown");
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;
    
    response.BuildUnknownMessageResponse(0);
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS UnknownMessageResponse to " << transport->GetRemoteAddressString());
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}


/////////////////////////////////////////////////////////////////////////////
