/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Gatekeeper functionality for links between server
 
    Comments: Gatekeeper/RAS protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/LinkMgr v5.0.x.x.root/H323/source/gklink.cpp         $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 9.09.05 12:49                                                $
   $Modtime:: 9.09.05 11:48                                                $
  $Revision:: 5                                                            $
  $Workfile:: gklink.cpp                                                   $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "gklink.h"
#include "h323pdu.h"

using namespace std;
using namespace Tools;

#define USE_DISCOVER


class H225RasLinkThread : public SThread
{
public:
    H225RasLinkThread( H323EndPoint & rEndpoint,
                       H323GkLink   & rGkLink );
    
protected:
    void Main();
    
    H323GkLink & m_rGkLink;
};


#define new SNEW


/////////////////////////////////////////////////////////////////////////////

H225RasLinkThread::H225RasLinkThread( H323EndPoint & rEndpoint,
                                      H323GkLink   & rGkLink )
 : SThread(rEndpoint.GetRasThreadStackSize(), NoAutoDeleteThread),
   m_rGkLink(rGkLink)
{
    Resume();
}


void H225RasLinkThread::Main()
{
    SDEFFUNC("H225RasLinkThread::Main");
    
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Started RAS link thread");
    
    m_rGkLink.HandleRasChannel();
    m_rGkLink.SetRASChannelInvalid();
    
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Stopped RAS link thread");
}


/////////////////////////////////////////////////////////////////////////////
// GK client (H323 endpoint) specific data.

H323GkLink::Client::Client(H323GkLink     & gk,
                           H323EndPoint   & ep,
                           H323Transport  * tp,
                           Tools::SString   szH323Alias)
 : m_rGkLink(gk),
   endpoint(ep),
   transport(tp),
   m_szH323Alias(szH323Alias),
   m_bIsRegistered(false),
   m_bLightweightRRQInProgress(false),
   m_uTTLRegRequestRetries(ep.GetRasRequestRetries())
{
    m_RegTimer.SetNotifier( SCREATE_NOTIFIER( OnRegistrationExpiry ) );
}

//-----------------------------------------------------------------------------

H323GkLink::Client::~Client()
{
    m_RegTimer.Stop();
}

//-----------------------------------------------------------------------------

void H323GkLink::Client::OnRegistrationExpiry( STimer & Timer, int iContinous )
{
    SDEFFUNC("H323GkLink::Client::OnRegExp");
    
    // Check if RAS thread is ok
    if (m_rGkLink.IsRASChannelInvalid())
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() RAS channel invalid, starting GK register thread (alias: '"
            << m_szH323Alias << "' ...");
        
        m_RegTimer.Stop();
        // If thread is already running it will return!
        m_rGkLink.StartGkRegisterThread();
        return;
    }

    if ( !m_bLightweightRRQInProgress )
    {
        m_bLightweightRRQInProgress = true;
        
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Registration TTL expired -> try a lightweight RRQ..." );
        
        if (m_rGkLink.RegisterClientAgain(m_szH323Alias, true) == false)
        {
            m_RegTimer.Stop();

            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Lightweight registration failed!" );
            
            m_bIsRegistered             = false;
            m_bLightweightRRQInProgress = false;
            m_szEndpointIdentifier.clear();

            m_rGkLink.OnRegisterFailure();
        }
    }
}

//-----------------------------------------------------------------------------

void H323GkLink::Client::OnTTLRegRequestTimeout( STimer & Timer, int iContinous )
{
    SDEFFUNC("H323GkLink::Client::OnTTLRegReqTO");
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() No response on lightweight RRQ!" );
    
    if (m_uTTLRegRequestRetries-- > 0)
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Lightweight registration retry (remaining " << m_uTTLRegRequestRetries << ")");
        
        if (m_rGkLink.RegisterClientAgain(m_szH323Alias, true) == false)
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Lightweight registration failed!" );
            m_rGkLink.OnRegisterFailure();
            m_bIsRegistered             = false;
            m_bLightweightRRQInProgress = false;
            m_szEndpointIdentifier.clear();
        }
    }
    else
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Lightweight registration failed after retries!");
        m_rGkLink.OnRegisterFailure();
        m_bIsRegistered             = false;
        m_bLightweightRRQInProgress = false;
        m_szEndpointIdentifier.clear();
        m_uTTLRegRequestRetries = endpoint.GetRasRequestRetries(); // reset variable!
    }
    return;
}


/////////////////////////////////////////////////////////////////////////////
// GK register thread.
#pragma warning(disable:4355)
H323GkLink::GkLinkRegisterThread::GkLinkRegisterThread( H323GkLink                 & rGkLink,
                                                        H323EndPoint               & rEndpoint,
                                                        H323Transport              * rTransport,
                                                        const Tools::SStringVector & rH323Aliases )
 : SThread(rEndpoint.GetRasThreadStackSize(), NoAutoDeleteThread),
   m_rGkLink(rGkLink),
   m_rEndpoint(rEndpoint),
   m_pTransport(rTransport),
   m_H323Aliases(rH323Aliases),
   m_bStopReRegistration(false)
#ifdef _DEBUG
 , m_ReRegistrationSyncPoint(Tools::SString(Tools::SString::Printf, "H323GkLink::GkLinkRegisterThread::ReRegistrationSyncPoint[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#endif
#pragma warning(default:4355)
{
    Resume();
}

//-----------------------------------------------------------------------------

H323GkLink::GkLinkRegisterThread::~GkLinkRegisterThread()
{
    if ( !m_bStopReRegistration )
    {
        m_bStopReRegistration = true;
        // stop reregistration attemps
        m_ReRegistrationSyncPoint.Signal();
        WaitForTermination();
    }
}

//-----------------------------------------------------------------------------

void H323GkLink::GkLinkRegisterThread::Main()
{
    SDEFFUNC("H323GkLink::GkLinkRegThread::Main");
    
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Started GK link registration thread");
    
    SIPSocket::Address binding = m_rGkLink.GetTransportAddress().m_Address;
    unsigned uStartPort = m_rGkLink.GetTransportAddress().m_Port;
    
    m_pTransport->CleanUpOnTermination();
    delete m_pTransport;
    m_pTransport = new H323TransportUDP(m_rEndpoint, binding, uStartPort);
    m_rGkLink.SetH323Transport(m_pTransport);
    
    if ( m_rGkLink.Discover(m_rGkLink.GetGatekeeperName()) )
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Stop link registration thread because GK discovery successful");

        m_rGkLink.SetRASChannelValid();
        return;
    }
    
    m_pTransport->CleanUpOnTermination();
    delete m_pTransport;
    m_pTransport = new H323TransportUDP(m_rEndpoint, binding, uStartPort);
    m_rGkLink.SetH323Transport(m_pTransport);

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() link registration thread failed to discover GK");
}


/////////////////////////////////////////////////////////////////////////////

#pragma warning(disable:4355)
H323GkLink::H323GkLink( H323EndPoint            & ep,
                        const Tools::SIpAddress & transportAddress )
 : endpoint(ep),
   m_LocalInterfaceAddress(transportAddress)
#ifdef _DEBUG
 , m_GkMutex(Tools::SString(Tools::SString::Printf, "H323GkLink::GkMutex[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#endif
#pragma warning(default:4355)
{
    SDEFFUNC("H323GkLink::H323GkLink");
    
    m_uNextRASReqSequenceNumber    = 0;
    m_uLastRASReqSequenceNumber    = 0;
    m_uLastRASReqTag               = 0;
    m_uAllocatedBandwidth          = 0;
    
    m_bGatekeeperDiscoveryComplete = false;
    m_bRASChannelInvalid           = false;
    
    // calculate MAXIMUM RAS request timeout as: timeout in seconds for RAS request * number of retries + 5 second spare time
    m_MaxMakeReqestTimeout   = STimeInterval( 0, (endpoint.GetGatekeeperRequestRetries() * endpoint.GetGatekeeperRequestTimeout().GetSeconds()) + 5);
    
    // create transport object from transport address given...
    transport = new H323TransportUDP( ep,
        transportAddress.m_Address,
        transportAddress.m_Port );
    
    m_TransportAddress      = transportAddress;
    m_LastTransportError    = 0;
    
    m_pGKLinkRegisterThread = 0;

    m_RASReqExpiryCheckTimer.SetNotifier( SCREATE_NOTIFIER( OnRASReqExpiryCheck ) );
    m_RASReqExpiryCheckTimer = 1000;

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "() using transport address: "
        << transportAddress << ", master RAS request timeout "
        << m_MaxMakeReqestTimeout.GetSeconds() << "sec" );
}

//-----------------------------------------------------------------------------

H323GkLink::~H323GkLink()
{
    SDEFFUNC("H323GkLink::~H323GkLink");

    m_RASReqExpiryCheckTimer.Stop(true);
    
    // Check for remaining entries in the client list and delete them.
    ClientsMap::iterator iter;
    while ((iter = m_Clients.begin()) != m_Clients.end())
    {
        if (iter->second)
        {
            delete iter->second;
            iter->second = 0;
        }
        m_Clients.erase(iter);
    }
    
    if ( m_pGKLinkRegisterThread )
    {
        delete m_pGKLinkRegisterThread;
        m_pGKLinkRegisterThread = 0;
    }
    
    if (transport != NULL)
    {
        transport->CleanUpOnTermination();
        delete transport;
        transport = 0;
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "()");
}

//-----------------------------------------------------------------------------

void H323GkLink::PrintOn(ostream & strm) const
{
    if (!m_szGatekeeperIdentifier.empty())
        strm << m_szGatekeeperIdentifier << "@";
    strm << transport->GetRemoteAddressString();
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnRegisterFailure()
{
    return true;
}

//-----------------------------------------------------------------------------

void H323GkLink::StartGkRegisterThread()
{
    SDEFFUNC("H323GkLink::StartGkRegThread");
    
    SWaitAndSignal wait(m_GkMutex);
    
    if (m_pGKLinkRegisterThread)
    {
        STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
            "() Gk reregister thread already running.");
        // stop and destroy current thread to restart with new thread
        delete m_pGKLinkRegisterThread;
        m_pGKLinkRegisterThread = 0;
    }
    
    // Get alias of all clients in map.
    SStringVector Aliases;
    ClientsMap::iterator iter;
    for (iter = m_Clients.begin(); iter != m_Clients.end(); ++iter)
    {
        if ((*iter).second)
        {
            Aliases.push_back((*iter).first);
        }
    }
    m_pGKLinkRegisterThread = new H323GkLink::GkLinkRegisterThread( *this, endpoint, transport, Aliases );
}

//-----------------------------------------------------------------------------

bool H323GkLink::AreAllClientsRegistered()
{
    SDEFFUNC("H323GkLink::AreAllClientsReg");
    
    // Check if all clients are registerd again
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Check if all clients are registered...");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator iter;
    for (iter = m_Clients.begin(); iter != m_Clients.end(); ++iter)
    {
        if ((*iter).second)
        {
            if (!(*iter).second->IsRegistered())
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Client with alias '" << (*iter).first << "' NOT registered!");
                return false;
            }
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Client with alias '" << (*iter).first << "' is registered!");
        }
        else
        {
            m_Clients.erase(iter);
        }
    }
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::UnregisterAllClients(int reason)
{
    // Send URQ for all clients in map.
    ClientsMap::iterator iter;
    for (iter = m_Clients.begin(); iter != m_Clients.end(); ++iter)
    {
        if ((*iter).second)
        {
            UnregistrationRequest((*iter).first, reason);
        }
    }
    return true;
}

//-----------------------------------------------------------------------------

H323GkLink::Client * H323GkLink::FindClientByAlias(const Tools::SString & rAlias)
{
    SDEFFUNC("H323GkLink::FindClientByAlias");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator it;
    for (it = m_Clients.begin(); it != m_Clients.end(); ++it)
    {
        if (rAlias.ToLower() == (*it).first.ToLower())
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() Found client by alias '" << rAlias << "'");
            return it->second;
        }
    }
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Client not found by alias '" << rAlias << "'");
    return 0;
}

//-----------------------------------------------------------------------------

H323GkLink::Client * H323GkLink::FindClientBySeqNum(unsigned seqNum)
{
    SDEFFUNC("H323GkLink::FindClientBySeqNu");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator it;
    for (it = m_Clients.begin(); it != m_Clients.end(); ++it)
    {
        if (   (*it).second
            && ((*it).second->m_uLastReqSequenceNumber == seqNum) )
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() Found client by seqNum " << seqNum);
            return it->second;
        }
    }
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Client not found by seqNum " << seqNum);
    return 0;
}

//-----------------------------------------------------------------------------

H323GkLink::Client * H323GkLink::FindClientByEndpointId(const Tools::SString & rEndpointId)
{
    SDEFFUNC("H323GkLink::FindClientByEpId");
    
    SWaitAndSignal wait(m_GkMutex);
    
    ClientsMap::iterator it;
    for (it = m_Clients.begin(); it != m_Clients.end(); ++it)
    {
        if (   (*it).second
            && ((*it).second->m_szEndpointIdentifier == rEndpointId) )
        {
            STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                "() Found client by endpointIdentifier '" << rEndpointId << "'");
            return it->second;
        }
    }
    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() Client not found by endpointIdentifier '" << rEndpointId << "'");
    return 0;
}

//-----------------------------------------------------------------------------

bool H323GkLink::RegisterClientAgain(const Tools::SString & szH323Alias,
                                     bool                   bIsLightWeightRRQ)
{
    SDEFFUNC("H323GkLink::RegClientAgain");
    
    if ( RegistrationRequest(szH323Alias, bIsLightWeightRRQ) == false ) // perform RRQ
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() RRQ failed for client with alias '" << szH323Alias << "'");
        return false;
    }
    return true;
}

//-----------------------------------------------------------------------------

void H323GkLink::HandleRasChannel()
{
    SDEFFUNC("H323GkLink::HandleRasChannel");
    
    transport->SetReadTimeout(SMaxTimeInterval);
    H323RasPDU response;
    
    for(;;)
    {
        switch(response.Read(*transport))
        {
        case H323RasPDU::Success:
            {
                dynamic_cast<H323TransportUDP *>(transport)->
                    GetLastReceiveAddress(m_LastReceiveIpAddress.m_Address,
                    m_LastReceiveIpAddress.m_Port);

                if ( HandleRasPDU(response) == false )
                {
                    STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                        "() failed to handle RAS mesage!");
                }
                else if (    response.GetTag() == H225_RasMessage::e_registrationConfirm
                          && m_pGKLinkRegisterThread )
                {
                    // Check if all clients are registerd again
                    if (AreAllClientsRegistered() == true)
                    {
                        // Stop and delete the GK reregistration thread
                        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                            "() Deleting reregister thread...");
                        delete m_pGKLinkRegisterThread;
                        m_pGKLinkRegisterThread = 0;
                    }
                }
            }
            break;
            
        case H323RasPDU::TransportFailure:
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() RAS Read(): TransportFailure, No: "
                    << transport->GetErrorNumber() << ", " << transport->GetErrorText());
                
                if (transport->GetErrorNumber() == ECONNRESET)
                {
                    // This can occure and is not handled as an error, because
                    // when gatekeeper is down and we send e.g. a RRQ we get
                    // an icmp-message from the machine running gatekeeper.
                    // This will result in a failed read-operation on the socket,
                    // but here we will ignore it and everything works fine!
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                        "() Ignoring transport error 'connection reset'!");
                    continue;
                }
                return;
            }
            break;
            
        case H323RasPDU::DecodeFailure:
            {
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() Decode error - discarding RAS message!");
                continue;
            }
            break;
            
        default:
            {
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() Unknown error reading RAS message - ignoring it!");
                continue;
            }
            break;
        }
    }
}

//-----------------------------------------------------------------------------

bool H323GkLink::HandleRasPDU(const H323RasPDU & response)
{
    SDEFFUNC("H323GkLink::HandleRasPDU");
    
    // Tracing
    if (STrace::CanTrace(g_pH323Module->m_GkTraceId, STrace::Info))
    {
        SString szMessage = response.GetTagName();
        szMessage[0] = ::toupper(szMessage[0]);
        
        if (m_LastReceiveIpAddress == SIpAddress())
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS " << szMessage);
        }
        else
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS " << szMessage << " from " << m_LastReceiveIpAddress);
        }
    }
    
    switch (response.GetTag()) {
    case H225_RasMessage::e_gatekeeperRequest :
        return OnReceiveGatekeeperRequest(response);
        
    case H225_RasMessage::e_gatekeeperConfirm :
        return OnReceiveGatekeeperConfirm(response);
        
    case H225_RasMessage::e_gatekeeperReject :
        return OnReceiveGatekeeperReject(response);
        
    case H225_RasMessage::e_registrationRequest :
        return OnReceiveRegistrationRequest(response);
        
    case H225_RasMessage::e_registrationConfirm :
        return OnReceiveRegistrationConfirm(response);
        
    case H225_RasMessage::e_registrationReject :
        return OnReceiveRegistrationReject(response);
        
    case H225_RasMessage::e_unregistrationRequest :
        return OnReceiveUnregistrationRequest(response);
        
    case H225_RasMessage::e_unregistrationConfirm :
        return OnReceiveUnregistrationConfirm(response);
        
    case H225_RasMessage::e_unregistrationReject :
        return OnReceiveUnregistrationReject(response);
        
    case H225_RasMessage::e_admissionRequest :
        return OnReceiveAdmissionRequest(response);
        
    case H225_RasMessage::e_admissionConfirm :
        return OnReceiveAdmissionConfirm(response);
        
    case H225_RasMessage::e_admissionReject :
        return OnReceiveAdmissionReject(response);
        
    case H225_RasMessage::e_bandwidthRequest :
        return OnReceiveBandwidthRequest(response);
        
    case H225_RasMessage::e_bandwidthConfirm :
        return OnReceiveBandwidthConfirm(response);
        
    case H225_RasMessage::e_bandwidthReject :
        return OnReceiveBandwidthReject(response);
        
    case H225_RasMessage::e_disengageRequest :
        return OnReceiveDisengageRequest(response);
        
    case H225_RasMessage::e_disengageConfirm :
        return OnReceiveDisengageConfirm(response);
        
    case H225_RasMessage::e_disengageReject :
        return OnReceiveDisengageReject(response);
        
    case H225_RasMessage::e_locationRequest :
        return OnReceiveLocationRequest(response);
        
    case H225_RasMessage::e_locationConfirm :
        return OnReceiveLocationConfirm(response);
        
    case H225_RasMessage::e_locationReject :
        return OnReceiveLocationReject(response);
        
    case H225_RasMessage::e_infoRequest :
        return OnReceiveInfoRequest(response);
        
    case H225_RasMessage::e_infoRequestResponse :
        return OnReceiveInfoRequestResponse(response);
        
    case H225_RasMessage::e_nonStandardMessage :
        return OnReceiveNonStandardMessage(response);
        
    case H225_RasMessage::e_unknownMessageResponse :
        return OnReceiveUnknownMessageResponse(response);
        
    case H225_RasMessage::e_requestInProgress :
        return OnReceiveRequestInProgress(response);
        
    case H225_RasMessage::e_resourcesAvailableIndicate :
        return OnReceiveResourcesAvailableIndicate(response);
        
    case H225_RasMessage::e_resourcesAvailableConfirm :
        return OnReceiveResourcesAvailableConfirm(response);
        
    case H225_RasMessage::e_infoRequestAck :
        return OnReceiveInfoRequestAck(response);
        
    case H225_RasMessage::e_infoRequestNak :
        return OnReceiveInfoRequestNak(response);
        
    default :
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() RECV RAS unknown message from " << m_LastReceiveIpAddress); 
        return OnReceiveUnkown(response);
    }
    
    return false;
}

//-----------------------------------------------------------------------------

unsigned H323GkLink::GetNextSequenceNumber()
{
    SWaitAndSignal lock (m_GkMutex);

    ++m_uNextRASReqSequenceNumber;
    if (m_uNextRASReqSequenceNumber > 65535)
        m_uNextRASReqSequenceNumber = 1;
    return m_uNextRASReqSequenceNumber;
}

//-----------------------------------------------------------------------------

H323GkLink::RASReqInProgress *  H323GkLink::MakeRequest(const H323RasPDU & request, const unsigned uSeqNumber)
{
    SDEFFUNC("H323GkLink::MakeRequest");
    
    if ( m_bRASChannelInvalid )
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() RAS channel invalid -> discard request!" );
        return 0;
    }
    
    RASReqInProgress *pRASReqInProgress = new RASReqInProgress(request);
    if (SBAD_READPTR(pRASReqInProgress))
        return 0;

    // lock all other RAS request related activity...
    m_GkMutex.Wait();
    
    // Tracing
    if (STrace::CanTrace(g_pH323Module->m_GkTraceId, STrace::Info))
    {
        Tools::SIPSocket::Address ip = dynamic_cast<H323TransportUDP *>(transport)->GetRemoteIpAddress();
        WORD port                    = dynamic_cast<H323TransportUDP *>(transport)->GetRemotePort();
        SString szMessage = request.GetTagName();
        szMessage[0] = ::toupper(szMessage[0]);
        
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS " << szMessage << ", seqNum= " << uSeqNumber << " to " << ip << ":" <<  port);
    }
    
    // get retry and timeout values from endpoint
    pRASReqInProgress->m_ReqTimeout = endpoint.GetRasRequestTimeout();
    pRASReqInProgress->m_MaxNumberOfReqRetries = endpoint.GetRasRequestRetries();
    
    // try to send the RAS request message and wait for reply...
    m_RASReqInProgressMap.insert(pair< const unsigned, H323GkLink::RASReqInProgress *>(uSeqNumber, pRASReqInProgress));
    if (pRASReqInProgress->m_rPendingReq.Write(*transport) != H323RasPDU::Success)
    {
        m_LastTransportError = transport->GetErrorNumber();
        
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "write to socket failed, error=" << m_LastTransportError );
        
        m_RASReqInProgressMap.erase(uSeqNumber);
        delete pRASReqInProgress;
        
        m_GkMutex.Signal();
        return 0;
    }
    pRASReqInProgress->m_LastReqAttempt = STimer::Tick();
    m_GkMutex.Signal();
    
    return pRASReqInProgress;
}

//-----------------------------------------------------------------------------

void H323GkLink::CleanupReqInProgress(RASReqInProgress * pRASReqInProgress)
{
    SWaitAndSignal lock(m_GkMutex);
    
    RASReqInProgressMap::iterator pos = m_RASReqInProgressMap.begin();
    while ( pos != m_RASReqInProgressMap.end() )
    {
        if (pRASReqInProgress == (*pos).second)
        {
            m_RASReqInProgressMap.erase((*pos).first);
            break;
        }
        ++pos;
    }
    delete pRASReqInProgress;
}

//-----------------------------------------------------------------------------

bool H323GkLink::CheckForResponse(unsigned reqTag, unsigned seqNum)
{
    SDEFFUNC("H323GkLink::CheckForResponse");
    
    SWaitAndSignal lock(m_GkMutex);
    
    if (m_uLastRASReqTag != reqTag)
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "Received message with wrong tag (reqTag=" << reqTag << ") " <<
            " (requestTag=" << m_uLastRASReqTag << ") seqNum=" << seqNum << 
            " lastReqSequenceNumber=" << m_uLastRASReqSequenceNumber);
        
        return false;
    }
    
    if (seqNum == m_uLastRASReqSequenceNumber)
        return true;
    
    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
        "() Received incorrect sequence number, want "
        << m_uLastRASReqSequenceNumber << ", got " << seqNum <<
        " reqTag=" << reqTag);
    
    return false;
}

H323GkLink::RASReqInProgress * H323GkLink::CheckForReqInProgress(unsigned reqTag, unsigned seqNum)
{
    SDEFFUNC("H323GkLink::CheckForReqInProgres");
    
    SWaitAndSignal lock (m_GkMutex);
    
    // try to find coresponding request in map
    RASReqInProgress *pRASReqInProgress = 0;
    RASReqInProgressMap::iterator pos = m_RASReqInProgressMap.find(seqNum);
    if ( pos == m_RASReqInProgressMap.end() )
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Received unknown sequence number " << seqNum <<
            " with reqTag=" << reqTag);
        
        return 0;
    }
    
    pRASReqInProgress = (*pos).second;
    if (!SBAD_READPTR(pRASReqInProgress))
    {
        if (pRASReqInProgress->m_rPendingReq.GetTag() != reqTag)
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Received message with wrong tag= " << reqTag
                << " (requestTag=" << pRASReqInProgress->m_rPendingReq.GetTag() << ") seqNum=" << seqNum);
            return 0;
        }
        
        // okay we found the req in queue AND it's the correct one...
        STRACE2_INFO3(g_pH323Module->m_RasTraceId, this, 0,
            "() Found req tag= " << reqTag
            << " for seqNum=" << seqNum);
        
        m_RASReqInProgressMap.erase(seqNum);
        return pRASReqInProgress;
    }
    
    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
        "() Queued request '" << reqTag << "' invalid! sequence number " << seqNum );
    
    return 0;
}

void H323GkLink::OnRASReqExpiryCheck( STimer & Timer, int iContinous )
{
    SDEFFUNC("H323GkLink::OnRASReqExpiryCheck");
    
    SWaitAndSignal lock(m_GkMutex);
    
    RASReqInProgress *pRASReqInProgress = 0;
    unsigned uRASReqSequenceNumber = 0;

    RASReqInProgressMap::iterator pos = m_RASReqInProgressMap.begin();
    while ( pos != m_RASReqInProgressMap.end() )
    {
        // get values from map element...
        uRASReqSequenceNumber = (*pos).first;
        pRASReqInProgress = (*pos).second;
        // ...and increment iterator (may be reset after erase() later on!)
        ++pos;

        if (!SBAD_READPTR(pRASReqInProgress))
        {
            if ( pRASReqInProgress->m_LastReqAttempt + endpoint.GetRasRequestTimeout() < STimer::Tick() )
            {
                ++pRASReqInProgress->m_NumberOfReqRetry;

                if ( pRASReqInProgress->m_NumberOfReqRetry > endpoint.GetRasRequestRetries() )
                {
                    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0, "() RAS request '"
                        << pRASReqInProgress->m_rPendingReq.GetTagName() << "' , seqNum="
                        << uRASReqSequenceNumber << " failed with timeout!");
                    
                    // max number of retries exceeded -> remove request from list!
                    m_RASReqInProgressMap.erase(uRASReqSequenceNumber);
                    // reset iterator position
                    pos = m_RASReqInProgressMap.begin();
                    // unblock pending request with timeout
                    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RequestTimeout;
                    pRASReqInProgress->m_PendingReqSync.Signal();
                }
                else
                {
                    STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0, "() Timeout: will resent '"
                        << pRASReqInProgress->m_rPendingReq.GetTagName() << "', seq num= "
                        << uRASReqSequenceNumber << ", retry number "
                        << pRASReqInProgress->m_NumberOfReqRetry);
                
                    // resent stored RAS request
                    if (pRASReqInProgress->m_rPendingReq.Write(*transport) == H323RasPDU::Success)
                    {
                        // save new time stamp
                        pRASReqInProgress->m_LastReqAttempt = STimer::Tick();
                    }
                    else // write() failed -> discard RAS request with timeout
                    {
                        m_LastTransportError = transport->GetErrorNumber();
                        
                        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                            "() write to socket failed, error=" << hex << m_LastTransportError );
                        
                        // failed to repeat request due socket error -> remove request from list!
                        m_RASReqInProgressMap.erase(uRASReqSequenceNumber);
                        // reset iterator position
                        pos = m_RASReqInProgressMap.begin();
                        // unblock pending request with timeout
                        pRASReqInProgress->m_RASReqResult = RASReqInProgress::RequestTimeout;
                        pRASReqInProgress->m_PendingReqSync.Signal();
                    }
                }
            }
        }
    }
    
    // start over with next check...
    m_RASReqExpiryCheckTimer = 1000;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveRequestInProgress(const H225_RequestInProgress & rip)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_requestInProgress, rip.m_requestSeqNum);
    
    if (pRASReqInProgress == 0)
        return false;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RequestInProgress;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::Discover(const H323TransportAddress & address)
{
    SDEFFUNC("H323GkLink::Discover");
    
#ifdef USE_DISCOVER
    H323RasPDU request;
    H225_GatekeeperRequest & grq = request.BuildGatekeeperRequest(GetNextSequenceNumber());
    endpoint.SetEndpointTypeInfo(grq.m_endpointType);
    transport->SetUpTransportPDU(grq.m_rasAddress, true);
    
    OnSendGatekeeperRequest(grq);
    
    m_bGatekeeperDiscoveryComplete = false;
    
    m_uLastRASReqTag            = request.GetTag();
    m_uLastRASReqSequenceNumber = grq.m_requestSeqNum;
    
    m_szGatekeeperName = address;

    unsigned retries = endpoint.GetGatekeeperRequestRetries();
    while (!transport->DiscoverGatekeeper(*this, request, address)) {
        if (--retries == 0)
            return false;
    }
    
    if (!m_bGatekeeperDiscoveryComplete)
        return false;
    
#else
    if (transport->SetRemoteAddress(address) == false)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() Could not set remote address!");
        return false;
    }
#endif // USE_DISCOVER
    
    if (transport->Connect())
        transport->AttachThread(new H225RasLinkThread(endpoint, *this));
    
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveGatekeeperConfirm(const H225_GatekeeperConfirm & gcf)
{
    SDEFFUNC("H323GkLink::OnReceiveGCF");
    
    if (!CheckForResponse(H225_RasMessage::e_gatekeeperRequest, gcf.m_requestSeqNum))
        return false;
    
    m_LocatedAddress = gcf.m_rasAddress;
    m_LastGKAddress  = gcf.m_rasAddress;

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Gatekeeper discovery found " << m_LocatedAddress);
    
    if (!transport->SetRemoteAddress(m_LastGKAddress)) {
        STRACE2_WARN(g_pH323Module->m_H225TraceId, this, 0,
            "() Invalid gatekeeper discovery address: \"" << m_LastGKAddress << '"');
    }
    
    if (gcf.HasOptionalField(H225_GatekeeperConfirm::e_gatekeeperIdentifier))
    {
        m_szGatekeeperIdentifier = (SString)gcf.m_gatekeeperIdentifier;
    }

    m_bGatekeeperDiscoveryComplete = true;

    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::RegistrationRequest(const Tools::SString & szH323Alias,
                                     bool                   bIsLightWeightRRQ)
{
    SDEFFUNC("H323GkLink::RegistrationReq");
    
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Cannot register with empty H.323 alias!");
        return false;
    }
    
    // Check if RAS thread is ok
    if (IsRASChannelInvalid())
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() RAS channel invalid, starting GK discovery thread for link with alias: '"
            << szH323Alias << "' ...");
        
        //m_RegTimer.Stop();
        // If thread is already running it will return!
        StartGkRegisterThread();
        return false;
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Register client with alias '" << szH323Alias
        << (bIsLightWeightRRQ ? "' (lightweight RRQ) ..." : "' (full RRQ) ..."));
    
    Client * pClient = FindClientByAlias(szH323Alias);
    if (pClient == 0)
    {
        // No client found, add new one.
        m_GkMutex.Wait();
        
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Insertion of new client with alias '" << szH323Alias << "'");
        pClient = new Client(*this, endpoint, transport, szH323Alias);
        
        pair<ClientsMap::iterator, bool> Answer;
        Answer = m_Clients.insert(pair<const SString, Client *>(szH323Alias, pClient));
        if (Answer.second == false)
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Insertion of client with alias '" << szH323Alias
                << "' in map failed!");
            m_GkMutex.Signal();
            return false;
        }
        m_GkMutex.Signal();
    }
    
    H323RasPDU request;
    H225_RegistrationRequest & rrq = request.BuildRegistrationRequest(GetNextSequenceNumber());
    rrq.m_discoveryComplete = m_bGatekeeperDiscoveryComplete;
    m_bGatekeeperDiscoveryComplete = false;
    
    rrq.m_rasAddress.SetSize(1);
    transport->SetUpTransportPDU(rrq.m_rasAddress[0], true);
    
    const H323ListenerList & listeners = endpoint.GetListeners();
    if (listeners.empty()) {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Cannot register with Gatekeeper without a H323Listener!");
        return false;
    }
    
    for (size_t i = 0; i < listeners.size(); i++)
    {
        H323ListenerTCP * pListener = dynamic_cast<H323ListenerTCP *>(listeners[i]);
        if (SBAD_READPTR(pListener))
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Invalid listener pointer.");
            continue;
        }
        else
        {
            SIPSocket::Address addr = GetLocalInterfaceAddress();
            if (pListener->IsListeningOnAddress(addr))
            {
                STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                    "() Is listening on " << addr);
                pListener->SetUpTransportPDU(rrq.m_callSignalAddress, addr);
            }
            else
            {
                STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                    "() Is not listening on " << addr);
            }
        }
    }
    
    endpoint.SetEndpointTypeInfo(rrq.m_terminalType,szH323Alias);
    endpoint.SetVendorIdentifierInfo(rrq.m_endpointVendor);
    
    if ( !bIsLightWeightRRQ )
    {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_terminalAlias);
        rrq.m_terminalAlias.SetSize(1);
        H323SetAliasAddress(pClient->m_szH323Alias, rrq.m_terminalAlias[0]);
    }
    
    if (!m_szGatekeeperIdentifier.empty()) {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_gatekeeperIdentifier);
        rrq.m_gatekeeperIdentifier = m_szGatekeeperIdentifier;
    }
    
    if (!pClient->m_szEndpointIdentifier.empty()) {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_endpointIdentifier);
        rrq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    }
    
    if ( endpoint.GetRegistrationTTL() > 0 )
    {
        rrq.IncludeOptionalField(H225_RegistrationRequest::e_timeToLive);
        rrq.m_timeToLive = endpoint.GetRegistrationTTL(); // set RRQ time to live
    }
    
    rrq.m_keepAlive = bIsLightWeightRRQ; // true = keep alive with light weight RRQ
    
    OnSendRegistrationRequest(rrq);

    // while RRQ we have to match the RCF to client by the sequence number
    pClient->m_uLastReqSequenceNumber = rrq.m_requestSeqNum;
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, rrq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if ( pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived )
            {
                bResult = true;
            }
            else
            {
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() RRQ NOT confirmed!");
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout in request!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveRegistrationConfirm(const H225_RegistrationConfirm & rcf)
{
    SDEFFUNC("H323GkLink::OnReceiveRCF");
    
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_registrationRequest, rcf.m_requestSeqNum);
    
    if (pRASReqInProgress == 0)
        return false;

    SwyxH225NonStandardData nsData(g_pH323Module->m_RasTraceId);
    if (rcf.HasOptionalField(H225_RegistrationConfirm::e_nonStandardData))
    {
		SPerStream ExtStrm;
        // get NonStandardData out of RCF
        GetRCFNonStandardData(rcf, ExtStrm);
        // Handle NonStandardData
        nsData.HandleSwH225NonStandardData(ExtStrm);

        // Check server version if available
        // (may be used e.g. for a codec compatibility check -> TODO)
        if (!nsData.m_H225VendorIdentifier.m_versionId.AsString().empty())
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Server version: " << nsData.m_H225VendorIdentifier.m_versionId.AsString());
        }
    }

    m_GkMutex.Wait();
    // Search for client of this RCF
    Client * pClient = 0;
    // 1. Search for client of RCF terminalAlias if present
    if (rcf.HasOptionalField(H225_RegistrationConfirm::e_terminalAlias))
    {
        SStringVector H323Aliases;
        H323GetAliasAddressStringVector(rcf.m_terminalAlias,
            H225_AliasAddress::e_h323_ID,
            H323Aliases);
        if (!H323Aliases.empty())
        {
            for(size_t ii = 0; ii < H323Aliases.size(); ii++)
            {
                pClient = FindClientByAlias(H323Aliases[ii]);
                if (pClient != 0)
                    goto EndpointFound;
            }
        }
    }
    // 2. Search for client of RCF endpointIdentifier
    pClient = FindClientByEndpointId(rcf.m_endpointIdentifier);
    if (!SBAD_READPTR(pClient))
        goto EndpointFound;
    
    // 3. Search for client of RCF sequence number
    pClient = FindClientBySeqNum(rcf.m_requestSeqNum);
    if (pClient == 0)
    {
        m_GkMutex.Signal();

        // No client found
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client for this RCF found!");

        pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
        pRASReqInProgress->m_PendingReqSync.Signal();
        return false;
    }
    
EndpointFound:

    m_GkMutex.Signal();
    
    pClient->m_bIsRegistered = true;
    pClient->m_bLightweightRRQInProgress = false;
    
    if (rcf.HasOptionalField(H225_RegistrationConfirm::e_gatekeeperIdentifier))
    {
        m_szGatekeeperIdentifier = (SString)rcf.m_gatekeeperIdentifier;
    }
    pClient->m_szEndpointIdentifier = (SString)rcf.m_endpointIdentifier;
    
    pClient->m_RegTimer.Stop();
    if (rcf.HasOptionalField(H225_RegistrationConfirm::e_timeToLive))
    {
        pClient->m_RegTimer = STimeInterval(0, (long)(rcf.m_timeToLive * 0.75)); // sent keep alive RRQ after 75% expiration time
    }
    else
    {
        pClient->m_RegTimer = SMaxTimeInterval;
    }
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveRegistrationReject(const H225_RegistrationReject & rrj)
{
    SDEFFUNC("H323GkLink::OnReceiveRRJ");
    
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_registrationRequest, rrj.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;

    // Search for client of this RRJ
    Client * pClient = 0;
    pClient = FindClientBySeqNum(rrj.m_requestSeqNum);
    if (pClient == 0)
    {
        // No client found
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client for this RRJ found!");

        pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived;
        pRASReqInProgress->m_PendingReqSync.Signal();
        return false;
    }
    pClient->m_RegTimer.Stop();
    pClient->m_bLightweightRRQInProgress = false;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();

    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::UnregistrationRequest(const Tools::SString & szH323Alias,
                                       int                    reason)
{
    SDEFFUNC("H323GkLink::UnregistrationReq");
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Unregister client with alias '" << szH323Alias << "' ...");
    
    Client * pClient = 0;
    Tools::SString szEndpointIdentifier;
    
    if (!szH323Alias.empty())
    {
        pClient = FindClientByAlias(szH323Alias);
        if (pClient == 0)
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() No client found for alias '" << szH323Alias << "' (URQ)");
            return false;
        }
    }
    
    if (pClient)
    {
        szEndpointIdentifier = pClient->m_szEndpointIdentifier;
        pClient->m_RegTimer.Stop();
    }
    
    H323RasPDU request;
    H225_UnregistrationRequest & urq = request.BuildUnregistrationRequest(GetNextSequenceNumber());
    
    H225_TransportAddress rasAddress;
    transport->SetUpTransportPDU(rasAddress, true);
    
    const H323ListenerList & listeners = endpoint.GetListeners();
    for (size_t i = 0; i < listeners.size(); i++)
    {
        H323ListenerTCP * pListener = dynamic_cast<H323ListenerTCP *>(listeners[i]);
        if (SBAD_READPTR(pListener))
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Invalid listener pointer.");
            continue;
        }
        else
        {
            SIPSocket::Address addr = GetLocalInterfaceAddress();
            if (pListener->IsListeningOnAddress(addr))
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Is listening on " << addr);
                pListener->SetUpTransportPDU(urq.m_callSignalAddress, addr);
            }
            else
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Is not listening on " << addr);
            }
        }
    }
    
    if (endpoint.IsTerminal() || endpoint.IsGateway() || endpoint.IsGatekeeper())
    {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_endpointAlias);
        if (!szH323Alias.empty())
        {
            urq.m_endpointAlias.SetSize(1);
            H323SetAliasAddress(szH323Alias, urq.m_endpointAlias[0]);
        }
    }
    
    if (!m_szGatekeeperIdentifier.empty()) {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_gatekeeperIdentifier);
        urq.m_gatekeeperIdentifier = m_szGatekeeperIdentifier;
    }
    
    if (!szEndpointIdentifier.empty()) {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_endpointIdentifier);
        urq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    }
    
    if (reason >= 0) {
        urq.IncludeOptionalField(H225_UnregistrationRequest::e_reason);
        urq.m_reason = reason;
    }
    
    OnSendUnregistrationRequest(urq);
    
    // while URQ we have to match the UCF to client by the sequence number
    pClient->m_uLastReqSequenceNumber = urq.m_requestSeqNum;
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, urq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                bResult = true;
            }
            else
            {
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() RRQ NOT confirmed!");
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending URQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveUnregistrationConfirm(const H225_UnregistrationConfirm & ucf)
{
    SDEFFUNC("H323GkLink::OnReceiveUCF");
    
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_unregistrationRequest, ucf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    // Search for client of this UCF
    Client * pClient = 0;
    pClient = FindClientBySeqNum(ucf.m_requestSeqNum);
    if (pClient == 0)
    {
        // No client found
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() No client for this UCF found!");

        pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
        pRASReqInProgress->m_PendingReqSync.Signal();
        return false;
    }
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() URQ successful -> EP is now unregistered!");
    
    pClient->m_bIsRegistered = false;
    pClient->m_szEndpointIdentifier.clear();
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();

    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveUnregistrationRequest(const H225_UnregistrationRequest & urq)
{
    SDEFFUNC("H323GkLink::OnReceiveURQ");
    
    Client * pClient = 0;
    
    // 1. Search for client of URQ endpointIdentifier if present
    if (urq.HasOptionalField(H225_UnregistrationRequest::e_endpointIdentifier))
    {
        Tools::SString endpointId = urq.m_endpointIdentifier;
        if (!endpointId.empty())
        {
            pClient = FindClientByEndpointId(urq.m_endpointIdentifier);
        }
    }
    if (pClient == 0)
    {
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for URQ!");
        return true;
    }
    
    STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
        "() URQ received -> Endpoint '" << pClient->m_szH323Alias << "' is now unregistered!");
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS UnregistrationConfirm to " << transport->GetRemoteAddressString());
    
    H323RasPDU response;
    response.BuildUnregistrationConfirm(urq.m_requestSeqNum);
    if ( response.Write(*transport) == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() Write RAS UnregistrationConfirm to " << transport->GetRemoteAddressString() << " failed with error= " << transport->GetErrorNumber());
    }
    
    // Now client is not registered -> set and start timer to try reregister
    // Hack: We add a random value to prevent the same timeout (seq. no. problems etc.)
    //    int iAddOnTime = SRandom::GetInt()%10000;
    //    int iTimeout   = (endpoint.GetReRegisterTimeout() * 1000) + iAddOnTime;
    //    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
    //                 "() Endpoint '" << pClient->m_szH323Alias << "' will try to reregister in "
    //                << iTimeout << " milliseconds.");
    //    pClient->m_RegTimer = iTimeout;
    pClient->m_RegTimer.Stop();
    pClient->m_bIsRegistered = false;
    pClient->m_szEndpointIdentifier.clear();
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveUnregistrationReject(const H225_UnregistrationReject & urj)
{
    SDEFFUNC("H323GkLink::OnReceiveURJ");
    
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_unregistrationRequest, urj.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    // Search for client of this URJ
    Client * pClient = 0;
    pClient = FindClientBySeqNum(urj.m_requestSeqNum);
    if (pClient == 0)
    {
        // No client found
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client for this URJ found!");

        pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived;
        pRASReqInProgress->m_PendingReqSync.Signal();

        return false;
    }

    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();

    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::LocationRequest(const SString & alias,
                                 H323TransportAddress & address)
{
    SStringList aliases;
    aliases.push_back(alias);
    return LocationRequest(aliases, address);
}

//-----------------------------------------------------------------------------

bool H323GkLink::LocationRequest(const SStringList & aliases,
                                 H323TransportAddress & address)
{
    SDEFFUNC("H323GkLink::LocationRequest");

    H323RasPDU request;
    H225_LocationRequest & lrq = request.BuildLocationRequest(GetNextSequenceNumber());
    
    lrq.m_destinationInfo.SetSize(aliases.size());
    int i;
    SStringList::const_iterator pos;
    for (pos  = aliases.begin(), i = 0; pos != aliases.end(); ++pos, ++i)
    {
        H323SetAliasAddress((*pos), lrq.m_destinationInfo[i]);
    }

    OnSendLocationRequest(lrq);
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, lrq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                address = m_LocatedAddress;

                bResult = true;
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending LRQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveLocationConfirm(const H225_LocationConfirm & lcf)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_locationRequest, lcf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    m_LocatedAddress = lcf.m_callSignalAddress;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::AdmissionRequest(H323Connection & connection)
{
    H323TransportAddress address;
    return AdmissionRequest(connection, address);
}

//-----------------------------------------------------------------------------

bool H323GkLink::AdmissionRequest(H323Connection       & connection,
                                  H323TransportAddress & address)
{
    SDEFFUNC("H323GkLink::AdmissionRequest");
    
    Tools::SString szH323Alias = connection.GetH323Alias();
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No H.323 alias found for connection!");
        return false;
    }
    Client * pClient = FindClientByAlias(szH323Alias);
    if (pClient == 0)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "'!");
        return false;
    }
    
    if (!pClient->m_bIsRegistered)
    {
        STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending AdmissionRequest failed because endpoint is not registered!");
        return false;
    }
    
    H323RasPDU request;
    H225_AdmissionRequest & arq = request.BuildAdmissionRequest(GetNextSequenceNumber());
    
    
    arq.m_callType.SetTag(H225_CallType::e_pointToPoint);
    arq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    
    bool answeringCall = connection.HadAnsweredCall();
    arq.m_answerCall = answeringCall;
    
    arq.m_srcInfo.SetSize(1);
    if (answeringCall)
        H323SetAliasAddress(connection.GetRemotePartyName(), arq.m_srcInfo[0]);
    else {
        H323SetAliasAddress(pClient->m_szH323Alias, arq.m_srcInfo[0]);
        if (!connection.GetRemotePartyName().empty())
        {
            arq.IncludeOptionalField(H225_AdmissionRequest::e_destinationInfo);
            arq.m_destinationInfo.SetSize(1);
            H323SetAliasAddress(connection.GetRemotePartyName(), arq.m_destinationInfo[0]);
        }
    }
    
    const H323Transport * signallingChannel = connection.GetSignallingChannel();
    
    if (answeringCall)
    {
        arq.IncludeOptionalField(H225_AdmissionRequest::e_srcCallSignalAddress);
        signallingChannel->SetUpTransportPDU(arq.m_srcCallSignalAddress, false);
        arq.IncludeOptionalField(H225_AdmissionRequest::e_destCallSignalAddress);
        signallingChannel->SetUpTransportPDU(arq.m_destCallSignalAddress, true);
    }
    else
    {
        if (signallingChannel != NULL && signallingChannel->IsOpen())
        {
            arq.IncludeOptionalField(H225_AdmissionRequest::e_srcCallSignalAddress);
            signallingChannel->SetUpTransportPDU(arq.m_srcCallSignalAddress, true);
        }
        // Commented out the following because IpPbxServer does not need it and
        // third party gatekeeper have problems with it.
        //arq.IncludeOptionalField(H225_AdmissionRequest::e_destCallSignalAddress);
        //signallingChannel->SetUpTransportPDU(arq.m_destCallSignalAddress, false);
    }
    
    arq.m_bandWidth = connection.GetBandwidthAvailable();
    arq.m_callReferenceValue = connection.GetCallReference();
    arq.m_conferenceID = connection.GetConferenceIdentifier();
    arq.m_callIdentifier.m_guid = connection.GetCallIdentifier();
    
    if (!m_szGatekeeperIdentifier.empty())
    {
        arq.IncludeOptionalField(H225_AdmissionRequest::e_gatekeeperIdentifier);
        arq.m_gatekeeperIdentifier = m_szGatekeeperIdentifier;
    }
    
    OnSendAdmissionRequest(arq);
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, arq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                connection.SetBandwidthAvailable(m_uAllocatedBandwidth);
                address = m_LocatedAddress;

                bResult = true;
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending ARQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveAdmissionConfirm(const H225_AdmissionConfirm & acf)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_admissionRequest, acf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    if (acf.m_callModel.GetTag() == H225_CallModel::e_direct)
        m_LocatedAddress = "";
    else
        m_LocatedAddress = acf.m_destCallSignalAddress;
    
    m_uAllocatedBandwidth = acf.m_bandWidth;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::DisengageRequest(const H323Connection & connection, unsigned reason)
{
    SDEFFUNC("H323GkLink::DisengageRequest");
    
    Tools::SString szH323Alias = connection.GetH323Alias();
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() No H.323 alias found for connection!");
        return false;
    }
    Client * pClient = FindClientByAlias(szH323Alias);
    if (pClient == 0)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "'!");
        return false;
    }
    if (!pClient->m_bIsRegistered)
    {
        STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending DisengageRequest failed because endpoint is not registered!");
        return false;
    }
    
    H323RasPDU request;
    H225_DisengageRequest & drq = request.BuildDisengageRequest(GetNextSequenceNumber());
    
    drq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    drq.m_conferenceID = connection.GetConferenceIdentifier();
    drq.m_callReferenceValue = connection.GetCallReference();
    drq.m_callIdentifier.m_guid = connection.GetCallIdentifier();
    drq.m_disengageReason.SetTag(reason);
    drq.m_answeredCall = connection.HadAnsweredCall();
    
    if (!m_szGatekeeperIdentifier.empty()) {
        drq.IncludeOptionalField(H225_DisengageRequest::e_gatekeeperIdentifier);
        drq.m_gatekeeperIdentifier = m_szGatekeeperIdentifier;
    }
    
    OnSendDisengageRequest(drq);
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, drq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                bResult = true;
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending DRQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveDisengageConfirm(const H225_DisengageConfirm & dcf)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_disengageRequest, dcf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveDisengageRequest(const H225_DisengageRequest & drq)
{
    SDEFFUNC("H323GkLink::OnReceiveDRQ");
    
    H323GloballyUniqueID id = drq.m_callIdentifier.m_guid;
    endpoint.ClearCall(id.AsString(), H323Connection::GatekeeperForcedTermination);
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS DisengageConfirm to " << transport->GetRemoteAddressString());
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;
    response.BuildDisengageConfirm(drq.m_requestSeqNum);
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}

//-----------------------------------------------------------------------------

bool H323GkLink::BandwidthRequest(H323Connection & connection,
                                  unsigned requestedBandwidth)
{
    SDEFFUNC("H323GkLink::BandwidthRequest");
    
    Tools::SString szH323Alias = connection.GetH323Alias();
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() No H.323 alias found for connection!");
        return false;
    }
    Client * pClient = FindClientByAlias(szH323Alias);
    if (pClient == 0)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "'!");
        return false;
    }
    if (!pClient->m_bIsRegistered)
    {
        STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending BandwidthRequest failed because endpoint is not registered!");
        return false;
    }
    
    H323RasPDU request;
    H225_BandwidthRequest & brq = request.BuildBandwidthRequest(GetNextSequenceNumber());
    
    brq.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    brq.m_conferenceID = connection.GetConferenceIdentifier();
    brq.m_callReferenceValue = connection.GetCallReference();
    brq.m_callIdentifier.m_guid = connection.GetCallIdentifier();
    brq.m_bandWidth = requestedBandwidth;
    
    OnSendBandwidthRequest(brq);
    
    bool bResult = false;    
    RASReqInProgress *pRASReqInProgress = MakeRequest(request, brq.m_requestSeqNum);
    if ( pRASReqInProgress )
    {
        // now wait for competion of request
        if ( pRASReqInProgress->m_PendingReqSync.Wait(m_MaxMakeReqestTimeout) )
        {
            if (   pRASReqInProgress->m_RASReqResult == RASReqInProgress::ConfirmReceived
                || pRASReqInProgress->m_RASReqResult == RASReqInProgress::RejectReceived )
            {
                connection.SetBandwidthAvailable(m_uAllocatedBandwidth);

                bResult = true;
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Timeout while sending BRQ!");
        }
        CleanupReqInProgress( pRASReqInProgress );
    }
    
    return bResult;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveBandwidthConfirm(const H225_BandwidthConfirm & bcf)
{
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_bandwidthRequest, bcf.m_requestSeqNum);
    if (pRASReqInProgress == 0)
        return false;
    
    m_uAllocatedBandwidth = bcf.m_bandWidth;
    
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived;
    pRASReqInProgress->m_PendingReqSync.Signal();
    return true;
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveBandwidthRequest(const H225_BandwidthRequest & brq)
{
    SDEFFUNC("H323GkLink::OnReceiveBRQ");
    
    H323GloballyUniqueID id = brq.m_callIdentifier.m_guid;
    H323Connection * connection = endpoint.FindConnectionWithLock(id.AsString());
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;
    
    if (connection == NULL)
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS BandwidthReject to " << transport->GetRemoteAddressString());
        response.BuildBandwidthReject(brq.m_requestSeqNum,
            H225_BandRejectReason::e_invalidConferenceID);
    }
    else
    {
        if (connection->SetBandwidthAvailable(brq.m_bandWidth))
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() SEND RAS BandwidthConfirm to " << transport->GetRemoteAddressString());
            response.BuildBandwidthConfirm(brq.m_requestSeqNum, brq.m_bandWidth);
        }
        else
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() SEND RAS BandwidthReject to " << transport->GetRemoteAddressString());
            response.BuildBandwidthReject(brq.m_requestSeqNum,
                H225_BandRejectReason::e_insufficientResources);
        }
        connection->Unlock();
    }
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}

//-----------------------------------------------------------------------------

bool H323GkLink::InfoRequestResponse(const H323Connection * connection,
                                     unsigned seqNum)
{
    SDEFFUNC("H323GkLink::InfoRequestResp");
    
    Tools::SString szH323Alias = connection->GetH323Alias();
    if (szH323Alias.empty())
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() No H.323 alias found for connection!");
        return false;
    }
    Client * pClient = FindClientByAlias(szH323Alias);
    if (pClient == 0)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No client found for alias '" << szH323Alias << "'!");
        return false;
    }
    if (!pClient->m_bIsRegistered)
    {
        STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending InfoRequestResponse failed because endpoint is not registered!");
        return false;
    }
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;
    
    H225_InfoRequestResponse & irr = response.BuildInfoRequestResponse(seqNum);
    
    endpoint.SetEndpointTypeInfo(irr.m_endpointType,pClient!=NULL?pClient->m_szH323Alias:"");
    irr.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    transport->SetUpTransportPDU(irr.m_rasAddress, true);
    const H323ListenerList & listeners = endpoint.GetListeners();
    for (size_t i = 0; i < listeners.size(); i++)
        listeners[i]->SetUpTransportPDU(irr.m_callSignalAddress, irr.m_rasAddress);
    
    if (endpoint.IsTerminal()) {
        irr.IncludeOptionalField(H225_InfoRequestResponse::e_endpointAlias);
        irr.m_endpointAlias.SetSize(1);
        H323SetAliasAddress(pClient->m_szH323Alias, irr.m_endpointAlias[0]);
    }
    
    if (connection != NULL) {
        irr.m_perCallInfo.SetSize(1);
        H225_InfoRequestResponse_perCallInfo_subtype & info = irr.m_perCallInfo[0];
        info.m_callReferenceValue = connection->GetCallReference();
        info.m_callIdentifier.m_guid = connection->GetCallIdentifier();
        info.m_conferenceID = connection->GetConferenceIdentifier();
        RTP_Session * session = connection->GetSession(RTP_Session::DefaultAudioSessionID);
        if (session != NULL) {
            info.IncludeOptionalField(H225_InfoRequestResponse_perCallInfo_subtype::e_audio);
            info.m_audio.SetSize(1);
            session->OnSendRasInfo(info.m_audio[0]);
        }
        session = connection->GetSession(RTP_Session::DefaultVideoSessionID);
        if (session != NULL) {
            info.IncludeOptionalField(H225_InfoRequestResponse_perCallInfo_subtype::e_video);
            info.m_video.SetSize(1);
            session->OnSendRasInfo(info.m_video[0]);
        }
        const H323Transport & controlChannel = connection->GetControlChannel();
        controlChannel.SetUpTransportPDU(info.m_h245.m_recvAddress, true);
        controlChannel.SetUpTransportPDU(info.m_h245.m_sendAddress, false);
        info.m_callType.SetTag(H225_CallType::e_pointToPoint);
        info.m_bandWidth = connection->GetBandwidthUsed();
        info.m_callModel.SetTag(H225_CallModel::e_direct);
    }
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS InfoRequestResponse to " << transport->GetRemoteAddressString());
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveInfoRequest(const H225_InfoRequest & irq)
{
    SDEFFUNC("H323GkLink::OnRecvInfoRequest");
    
    // This is a default implementation which sends an IRR
    // "with all appropriate fields provided" if a client could be found.
    // A correct implementation using the connections and sending correct
    // IRRs for all connections (see OpenH323) has to be done!
    
    // Search for client of this IRQ
    Client * pClient = 0;
    pClient = FindClientBySeqNum(irq.m_requestSeqNum);
    if (SBAD_READPTR(pClient))
    {
        // No client found
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() No client for this IRQ found!");
        return false;
    }
    if (SBAD_READPTR(transport))
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() Invalid transport object!");
        return false;
    }
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;
    
    H225_InfoRequestResponse & irr = response.BuildInfoRequestResponse(irq.m_requestSeqNum);
    
    endpoint.SetEndpointTypeInfo(irr.m_endpointType,pClient!=NULL?pClient->m_szH323Alias:"");
    irr.m_endpointIdentifier = pClient->m_szEndpointIdentifier;
    transport->SetUpTransportPDU(irr.m_rasAddress, true);
    const H323ListenerList & listeners = endpoint.GetListeners();
    for (size_t i = 0; i < listeners.size(); i++)
    {
        if (SBAD_READPTR(listeners[i]))
            continue;
        listeners[i]->SetUpTransportPDU(irr.m_callSignalAddress, irr.m_rasAddress);
    }
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS InfoRequestResponse to " << transport->GetRemoteAddressString());
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}

//-----------------------------------------------------------------------------

bool H323GkLink::GetRCFNonStandardData(const H225_RegistrationConfirm & rcf,
                                       Tools::SByteVector             & data)
{
    SDEFFUNC("H323GkLink::GetRCFNSData");

    // Get NonStandardData part of RCF
    if (!rcf.HasOptionalField(H225_RegistrationConfirm::e_nonStandardData))
        return false;

    const H225_NonStandardIdentifier & nonStdId
        = rcf.m_nonStandardData.m_nonStandardIdentifier;

    if (!m_szObjectIdentifier.empty())
    {
        // Connection is using object identifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_object)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using object identifier "
                         << m_szObjectIdentifier);
            return false;
        }
        const SAsnObjectId & oid = nonStdId;
        if (oid != m_szObjectIdentifier.c_str())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received object identifier " << oid
                         << " != " << m_szObjectIdentifier);
            return false;
        }
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_h221NonStandard)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using h221NonStandard identifier!");
            return false;
        }
        const H225_H221NonStandard & h221Msg = nonStdId;

        // Get endpoints H.221 NonStandardIdentifier
        H225_H221NonStandard h221Endpoint;
        endpoint.SetH221NonStandardInfo(h221Endpoint);

        // Compare H.221 NonStandardIdentifier
        if (h221Msg.m_t35CountryCode.GetValue() != h221Endpoint.m_t35CountryCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35CountryCode="
                         << h221Msg.m_t35CountryCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35CountryCode.GetValue());
            return false;
        }
        if (h221Msg.m_t35Extension.GetValue() != h221Endpoint.m_t35Extension.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35Extension="
                         << h221Msg.m_t35Extension.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35Extension.GetValue());
            return false;
        }
        if (h221Msg.m_manufacturerCode.GetValue() != h221Endpoint.m_manufacturerCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received manufacturerCode="
                         << h221Msg.m_manufacturerCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_manufacturerCode.GetValue());
            return false;
        }
    }

    // Get nonStandard data
    data.clear();
    data = rcf.m_nonStandardData.m_data.GetValue();

    return true;
}

//-----------------------------------------------------------------------------

#define DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
void H323GkLink::OnSend##name(H225_##name &) { }

#define DEFINE_GATEKEEPER_PROTOCOL_REQUEST(name) \
    DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
bool H323GkLink::OnReceive##name(const H225_##name &) { return false; }

#define DEFINE_GATEKEEPER_PROTOCOL_REJECT(name, req) \
    DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
    bool H323GkLink::OnReceive##name(const H225_##name & pdu) { \
    SDEFFUNC("H323Gatekeeper::OnReceive"#name); \
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_##req, pdu.m_requestSeqNum); \
    if (pRASReqInProgress == 0) return false; \
    STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0, "() Reveived " << #name ", reason: " << pdu.m_rejectReason.GetTagName()); \
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::RejectReceived; \
    pRASReqInProgress->m_PendingReqSync.Signal(); \
    return true; \
}

#define DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(name, req) \
    DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
    bool H323GkLink::OnReceive##name(const H225_##name & pdu) { \
    RASReqInProgress *pRASReqInProgress = CheckForReqInProgress(H225_RasMessage::e_##req, pdu.m_requestSeqNum); \
    if (pRASReqInProgress == 0) return false; \
    pRASReqInProgress->m_RASReqResult = RASReqInProgress::ConfirmReceived; \
    pRASReqInProgress->m_PendingReqSync.Signal(); \
    return true; \
}

DEFINE_GATEKEEPER_PROTOCOL_REQUEST(GatekeeperRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(GatekeeperConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(GatekeeperReject, gatekeeperRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(RegistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(RegistrationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_SEND(RegistrationReject);
//DEFINE_GATEKEEPER_PROTOCOL_REJECT(RegistrationReject, registrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(UnregistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(UnregistrationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_SEND(UnregistrationReject);
//DEFINE_GATEKEEPER_PROTOCOL_REJECT(UnregistrationReject, unregistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(AdmissionRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(AdmissionConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(AdmissionReject, admissionRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(BandwidthRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(BandwidthConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(BandwidthReject, bandwidthRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(DisengageRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(DisengageConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(DisengageReject, disengageRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(LocationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(LocationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(LocationReject, locationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(InfoRequest);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestResponse, infoRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(NonStandardMessage);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(UnknownMessageResponse);
DEFINE_GATEKEEPER_PROTOCOL_SEND(RequestInProgress);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(ResourcesAvailableIndicate);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(ResourcesAvailableConfirm, resourcesAvailableIndicate);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestAck, infoRequest);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestNak, infoRequest);

//-----------------------------------------------------------------------------

bool H323GkLink::OnReceiveUnkown(const H323RasPDU &)
{
    SDEFFUNC("H323GkLink::OnReceiveUnkown");
    
    H323RasPDU response;
    H323RasPDU::H323RasPDUError Result;
    
    response.BuildUnknownMessageResponse(0);
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS UnknownMessageResponse to " << transport->GetRemoteAddressString());
    
    Result = response.Write(*transport);
    if ( Result == H323RasPDU::TransportFailure )
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = transport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}


/////////////////////////////////////////////////////////////////////////////
