/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: GSM Codec
 
    Comments: H.323 protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/gsmcodec.cpp               $
    $Author:: Martin.hueser                                                $
      $Date:: 31.10.05 10:18                                               $
   $Modtime:: 26.10.05 15:40                                               $
  $Revision:: 2                                                            $
  $Workfile:: gsmcodec.cpp                                                 $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "gsmcodec.h"
#include "h245.h"
#include "../include/rtp.h" // prevent conflicts with platform SDK
#include "h323lib.h"

using namespace std;
using namespace Tools;

// Module including Swyx H.323 specific stuff, e.g. tracing
extern H323Module *g_pH323Module;

#define new SNEW


H323_GSM0610Capability::H323_GSM0610Capability()
  : H323AudioCapability(7*H323_GSM0610Codec::BytesPerFrame,
                        4*H323_GSM0610Codec::BytesPerFrame)
{
  maxFrameSize = H323_GSM0610Codec::BytesPerFrame;
}


SObject * H323_GSM0610Capability::Clone() const
{
  return new H323_GSM0610Capability(*this);
}


void H323_GSM0610Capability::PrintOn(ostream & strm) const
{
  strm << "GSM Full Rate (06.10) ";
  H323AudioCapability::PrintOn(strm);
}


unsigned H323_GSM0610Capability::GetSubType() const
{
  return H245_AudioCapability::e_gsmFullRate;
}


bool H323_GSM0610Capability::OnSendingPDU(H245_AudioCapability & cap,
                                          unsigned packetSize) const
{
  cap.SetTag(H245_AudioCapability::e_gsmFullRate);

  H245_GSMAudioCapability & gsm = cap;
  gsm.m_audioUnitSize = packetSize;
  return true;
}


bool H323_GSM0610Capability::OnReceivedPDU(const H245_AudioCapability & cap,
                                           unsigned & packetSize)
{
  if (cap.GetTag() != H245_AudioCapability::e_gsmFullRate)
    return false;

  const H245_GSMAudioCapability & gsm = cap;
  packetSize = gsm.m_audioUnitSize;
  if (packetSize == 0)
    packetSize = 1;  
  return true;
}


H323Codec * H323_GSM0610Capability::CreateCodec(H323Codec::Direction direction) const
{
  return new H323_GSM0610Codec(direction);
}


/////////////////////////////////////////////////////////////////////////////

extern "C" gsm_state * gsm_create();
//gsm_state * gsm_create() { return NULL; }

H323_GSM0610Codec::H323_GSM0610Codec(Direction dir)
  : H323FramedAudioCodec(dir, SamplesPerFrame, BytesPerFrame)
{
    SDEFFUNC("H323_GSM0610Codec::H323_GSM0610Co");

    rtpPayloadType = RTP_DataFrame::GSM;
    gsm = gsm_create();
    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "() GSM " << (dir == Encoder ? "en" : "de") << "coder created");
}


extern "C" void gsm_destroy(gsm_state *);
//void gsm_destroy(gsm_state *) { }

H323_GSM0610Codec::~H323_GSM0610Codec()
{
  gsm_destroy(gsm);
}


extern "C" void gsm_encode(gsm_state *, const short *, BYTE *);
//void gsm_encode(gsm_state *, const short *, BYTE *) { }

bool H323_GSM0610Codec::EncodeFrame(BYTE * buffer)
{
  gsm_encode(gsm, &(*(sampleBuffer.begin())), buffer);
  return true;
}


extern "C" void gsm_decode(gsm_state *, const BYTE *, short *);
//void gsm_decode(gsm_state *, const BYTE *, short *) { }

bool H323_GSM0610Codec::DecodeFrame(const BYTE * buffer)
{
  gsm_decode(gsm, buffer, &(*(sampleBuffer.begin())));
  return true;
}


/////////////////////////////////////////////////////////////////////////////
