/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323
 
    Comments: Swyx H.323 specific stuff 

  --------------------------------------------------------------------------

   $Archive:: /Server/SwyxSrv.root/H323/source/h323lib.cpp                 $
    $Author:: Martin.hueser                                                $
      $Date:: 6.08.04 16:29                                                $
   $Modtime:: 3.08.04 9:45                                                 $
  $Revision:: 12                                                           $
  $Workfile:: h323lib.cpp                                                  $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "h323lib.h"

using namespace std;
using namespace Tools;


H323Module *g_pH323Module = 0;


void H323LibInit(bool bUsingSwtcl)
{
    if (g_pH323Module != 0)
    {
        SAssertAlways("H323Module already created!");
    }
    g_pH323Module = new H323Module(bUsingSwtcl);

    g_pH323Module->InitTracing();
}


void H323LibExit()
{
    if (g_pH323Module != 0)
    {
        g_pH323Module->DestroyTracing();
        delete g_pH323Module;
        g_pH323Module = 0;
    }
}

//-----------------------------------------------------------------------------

H323Module::H323Module(bool bUsingSwtcl)
{
    // Init the tracing stuff
    m_CodecTraceId      = 0;
    m_H225TraceId       = 0;
    m_GkTraceId         = 0;
    m_H245TraceId       = 0;
    m_H323TraceId       = 0;
    m_RasTraceId        = 0;
    m_RtpTraceId        = 0;

    /// Default: Not using Swyx Telephony Class Library
    m_bUsingSwtcl = bUsingSwtcl;
}


H323Module::~H323Module()
{
}


void H323Module::InitTracing()
{
    g_pH323Module->m_CodecTraceId       = STrace::AppendModule("H323Codec ");
    g_pH323Module->m_H225TraceId        = STrace::AppendModule("H323H225  ");
    g_pH323Module->m_GkTraceId          = STrace::AppendModule("H323Gk    ");
    g_pH323Module->m_H245TraceId        = STrace::AppendModule("H323H245  ");
    g_pH323Module->m_H323TraceId        = STrace::AppendModule("H323      ");
    g_pH323Module->m_RasTraceId         = STrace::AppendModule("H323RAS   ");
    g_pH323Module->m_RtpTraceId         = STrace::AppendModule("H323RTP   ");
}


void H323Module::DestroyTracing()
{
    STrace::DisableModule(g_pH323Module->m_CodecTraceId);
    STrace::RemoveModule(g_pH323Module->m_CodecTraceId);

    STrace::DisableModule(g_pH323Module->m_H225TraceId);
    STrace::RemoveModule(g_pH323Module->m_H225TraceId);

    STrace::DisableModule(g_pH323Module->m_GkTraceId);
    STrace::RemoveModule(g_pH323Module->m_GkTraceId);

    STrace::DisableModule(g_pH323Module->m_H245TraceId);
    STrace::RemoveModule(g_pH323Module->m_H245TraceId);

    STrace::DisableModule(g_pH323Module->m_H323TraceId);
    STrace::RemoveModule(g_pH323Module->m_H323TraceId);

    STrace::DisableModule(g_pH323Module->m_RasTraceId);
    STrace::RemoveModule(g_pH323Module->m_RasTraceId);

    STrace::DisableModule(g_pH323Module->m_RtpTraceId);
    STrace::RemoveModule(g_pH323Module->m_RtpTraceId);
}
