/*---------------------------------------------------------------------------

   Swyx Solutions GmbH
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Solutions GmbH are
  Copyright (C) 2000-2005 Swyx Solutions GmbH.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323 PDU
 
    Comments: H.323 PDU definitions

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/h323pdu.cpp                $
    $Author:: Martin.hueser                                                $
      $Date:: 31.10.05 10:18                                               $
   $Modtime:: 26.10.05 15:42                                               $
  $Revision:: 5                                                            $
  $Workfile:: h323pdu.cpp                                                  $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "h323pdu.h"

using namespace std;
using namespace Tools;

#define new SNEW

const char H225_ProtocolID[] = "0.0.8.2250.0.5";
const char H245_ProtocolID[] = "0.0.8.245.0.12";


///////////////////////////////////////////////////////////////////////////////

H323GloballyUniqueID::H323GloballyUniqueID()
{
  bool bFoundMAC = false;

  resize(16);

  //Time of UTC in 0.1 microseconds since 15 Oct 1582.
  static const SInt64 DaysToJan1970 = (365*(1970-1583) // Days in years
                                     + (1970-1583)/4 // Leap days
                                     - 3  // Allow for 1700, 1800, 1900 not leap years
                                     + 31  // Days in December 1583
                                     + 30  // Days in November 1583
                                     + 16); // Days from 15th October
  SInt64 timestamp = (time(NULL) + DaysToJan1970*60*60*24)*10000000;

  vector<BYTE>::operator[](0) = (BYTE)(timestamp&0xff);
  vector<BYTE>::operator[](1) = (BYTE)((timestamp>>8)&0xff);
  vector<BYTE>::operator[](2) = (BYTE)((timestamp>>16)&0xff);
  vector<BYTE>::operator[](3) = (BYTE)((timestamp>>24)&0xff);
  vector<BYTE>::operator[](4) = (BYTE)((timestamp>>32)&0xff);
  vector<BYTE>::operator[](5) = (BYTE)((timestamp>>40)&0xff);
  vector<BYTE>::operator[](6) = (BYTE)((timestamp>>48)&0xff);
  vector<BYTE>::operator[](7) = (BYTE)(((timestamp>>56)&0x0f) + 0x10);  // Version number is 1

  static WORD clockSequence = (WORD)(rand()&0x2ff);
  static SInt64 lastTimestamp = 0;
  if (lastTimestamp < timestamp)
    lastTimestamp = timestamp;
  else
    clockSequence = (WORD)((clockSequence+1)&0x2ff);

  vector<BYTE>::operator[](8) = (BYTE)((clockSequence>>8)&0x1f);
  vector<BYTE>::operator[](9) = (BYTE)(clockSequence&0x1f);

  SIPSocket::InterfaceTable           interfaces;
  SIPSocket::InterfaceTable::iterator interfacesIter;
  if (SIPSocket::GetInterfaceTable(interfaces)) {
    for (SByteVector::size_type i = 0; i < interfaces.size(); i++) {
      SString macAddrStr = interfaces[i]->GetMACAddress();
      if (!macAddrStr.empty() && macAddrStr != "44-45-53-54-00-00") {  /* not Win32 PPP device */
        SEthSocket::Address macAddress = macAddrStr;
        if (macAddress != SEthSocket::Address(""))
        {
          Tools::copy_n(macAddress.b, 6, begin()+10);
          bFoundMAC = true;
        }
      }
    }
  }

  if (!bFoundMAC)
  {
    vector<BYTE>::operator[](10) = (BYTE)(rand()|0x80);
    vector<BYTE>::operator[](11) = (BYTE)rand();
    vector<BYTE>::operator[](12) = (BYTE)rand();
    vector<BYTE>::operator[](13) = (BYTE)rand();
    vector<BYTE>::operator[](14) = (BYTE)rand();
    vector<BYTE>::operator[](15) = (BYTE)rand();
  }

  for (interfacesIter = interfaces.begin(); interfacesIter != interfaces.end(); ++interfacesIter)
  {
    delete (*interfacesIter);
  }
}


H323GloballyUniqueID::H323GloballyUniqueID(const SAsnOctetString & newId)
  : vector<BYTE>(newId.GetValue())
{
  SAssert(size() == 16, SInvalidParameter);
}


SString H323GloballyUniqueID::AsString() const
{
  SStringStream strm;
  PrintOn(strm);
  return strm;
}


void H323GloballyUniqueID::PrintOn(ostream & strm) const
{
  char fillchar = strm.fill();
  strm << hex << setfill('0')
       << setw(2) << (int)vector<BYTE>::operator[](0)
       << setw(2) << (int)vector<BYTE>::operator[](1)
       << setw(2) << (int)vector<BYTE>::operator[](2)
       << setw(2) << (int)vector<BYTE>::operator[](3) << '-'
       << setw(2) << (int)vector<BYTE>::operator[](4)
       << setw(2) << (int)vector<BYTE>::operator[](5) << '-'
       << setw(2) << (int)vector<BYTE>::operator[](6)
       << setw(2) << (int)vector<BYTE>::operator[](7) << '-'
       << setw(2) << (int)vector<BYTE>::operator[](8)
       << setw(2) << (int)vector<BYTE>::operator[](9) << '-'
       << setw(2) << (int)vector<BYTE>::operator[](10)
       << setw(2) << (int)vector<BYTE>::operator[](11)
       << setw(2) << (int)vector<BYTE>::operator[](12)
       << setw(2) << (int)vector<BYTE>::operator[](13)
       << setw(2) << (int)vector<BYTE>::operator[](14)
       << setw(2) << (int)vector<BYTE>::operator[](15)
       << dec << setfill(fillchar);
}


///////////////////////////////////////////////////////////////////////////////

H323SignalPDU::H323SignalPDU()
{
}


H225_Facility_UUIE * H323SignalPDU::BuildFacility(const H323Connection & connection,
                                                  bool empty)
{
  q931pdu.BuildFacility(connection.GetCallReference(), connection.HadAnsweredCall());
  if (empty) {
    m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_empty);
    return NULL;
  }

  m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_facility);
  H225_Facility_UUIE & fac = m_h323_uu_pdu.m_h323_message_body;

  fac.m_protocolIdentifier.SetValue(H225_ProtocolID);
  fac.IncludeOptionalField(H225_Facility_UUIE::e_callIdentifier);
  fac.m_callIdentifier.m_guid = connection.GetCallIdentifier();

  fac.m_reason.SetTag(H225_FacilityReason::e_undefinedReason);
  fac.m_multipleCalls = true;
  fac.m_maintainConnection = true;

  return &fac;
}


H225_Facility_UUIE * H323SignalPDU::BuildFacility(const H323Connection & connection,
                                                  bool empty,
                                                  const Tools::SByteVector & facilityIE)
{
    H225_Facility_UUIE * fac = BuildFacility(connection, empty);

    q931pdu.SetFacilityIE(facilityIE);

    return fac;
}


H225_CallProceeding_UUIE &
        H323SignalPDU::BuildCallProceeding(const H323Connection & connection,
                                           int callRef,
                                           const H323GloballyUniqueID & callIdentifier)
{
  q931pdu.BuildCallProceeding(callRef);

  m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_callProceeding);
  H225_CallProceeding_UUIE & proceeding = m_h323_uu_pdu.m_h323_message_body;

  proceeding.m_multipleCalls=true;//SAsnBoolean(true);  
  proceeding.m_maintainConnection=true;//SAsnBoolean(true);  

  proceeding.m_protocolIdentifier.SetValue(H225_ProtocolID);
  proceeding.m_callIdentifier.m_guid = callIdentifier;
  connection.GetEndPoint().SetEndpointTypeInfo(proceeding.m_destinationInfo,connection.GetH323Alias());

  return proceeding;
}


H225_CallProceeding_UUIE &
        H323SignalPDU::BuildCallProceeding(const H323Connection & connection,
                                           int callRef,
                                           const H323GloballyUniqueID & callIdentifier,
                                           const SIPSocket::Address & h245Address,
                                           WORD port)
{
    H225_CallProceeding_UUIE & proceeding = BuildCallProceeding(connection,
                                                                callRef,
                                                                callIdentifier);

    // indicate we are including the optional H245 address in the PDU
    proceeding.IncludeOptionalField(H225_CallProceeding_UUIE::e_h245Address);

    // convert IP address into the correct H245 type
    proceeding.m_h245Address.SetTag(H225_TransportAddress::e_ipAddress);
    H225_TransportAddress_ipAddress & ipAddress = proceeding.m_h245Address;
    ipAddress.m_ip.SetValue((DWORD)h245Address);
    ipAddress.m_port  = port;

    return proceeding;
}


H225_Alerting_UUIE & H323SignalPDU::BuildAlerting(const H323Connection & connection,
                                                  int callRef,
                                                  const H323GloballyUniqueID & callIdentifier,
                                                  const Q931::ProgressIndicator & ProgressInd)
{
  q931pdu.BuildAlerting(callRef, ProgressInd);

  m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_alerting);
  H225_Alerting_UUIE & alerting = m_h323_uu_pdu.m_h323_message_body;

  alerting.m_protocolIdentifier.SetValue(H225_ProtocolID);
  alerting.m_callIdentifier.m_guid = callIdentifier;
  connection.GetEndPoint().SetEndpointTypeInfo(alerting.m_destinationInfo,connection.GetH323Alias());

  return alerting;
}


H225_Alerting_UUIE & H323SignalPDU::BuildAlerting(const H323Connection & connection,
                                                  int callRef,
                                                  const H323GloballyUniqueID & callIdentifier,
                                                  const Q931::ProgressIndicator & ProgressInd,
                                                  const SIPSocket::Address & h245Address,
                                                  WORD port)
{
    H225_Alerting_UUIE & alerting = BuildAlerting(connection,
                                                  callRef,
                                                  callIdentifier,
                                                  ProgressInd);

    // indicate we are including the optional H245 address in the PDU
    alerting.IncludeOptionalField(H225_Alerting_UUIE::e_h245Address);

    // convert IP address into the correct H245 type
    alerting.m_h245Address.SetTag(H225_TransportAddress::e_ipAddress);
    H225_TransportAddress_ipAddress & ipAddress = alerting.m_h245Address;
    ipAddress.m_ip.SetValue((DWORD)h245Address);
    ipAddress.m_port  = port;

    return alerting;
}


H225_ReleaseComplete_UUIE &
        H323SignalPDU::BuildReleaseComplete(const H323Connection & connection,
                                            const H225_ReleaseCompleteReason & reason,
                                            const Q931::Cause & cause)
{
  q931pdu.BuildReleaseComplete(connection.GetCallReference(), connection.HadAnsweredCall(), cause);

  m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_releaseComplete);
  H225_ReleaseComplete_UUIE & release = m_h323_uu_pdu.m_h323_message_body;

  release.m_protocolIdentifier.SetValue(H225_ProtocolID);
  release.m_callIdentifier.m_guid = connection.GetCallIdentifier();

  if (reason.IsValid())
  {
    release.IncludeOptionalField(H225_ReleaseComplete_UUIE::e_reason);
    release.m_reason.SetTag(reason.GetTag());
  }
   
  return release;
}


H225_Connect_UUIE & H323SignalPDU::BuildConnect(const H323Connection & connection,
                                                int callRef,
                                                const H323GloballyUniqueID & callIdentifier,
                                                const H323GloballyUniqueID & conferenceId)
{
  q931pdu.BuildConnect(callRef, m_DisplayName);

  m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_connect);
  H225_Connect_UUIE & connect = m_h323_uu_pdu.m_h323_message_body;

  connect.m_protocolIdentifier.SetValue(H225_ProtocolID);
  connect.m_callIdentifier.m_guid = callIdentifier;
  connect.m_conferenceID = conferenceId;

  connect.m_multipleCalls = true;
  connect.m_maintainConnection = true;

  connection.GetEndPoint().SetEndpointTypeInfo(connect.m_destinationInfo,connection.GetH323Alias());

  return connect;
}


H225_Connect_UUIE & H323SignalPDU::BuildConnect(const H323Connection & connection,
                                                int callRef,
                                                const H323GloballyUniqueID & callIdentifier,
                                                const H323GloballyUniqueID & conferenceId,
                                                const SIPSocket::Address & h245Address,
                                                WORD port)
{
  H225_Connect_UUIE & connect = BuildConnect(connection,
                                             callRef,
                                             callIdentifier,
                                             conferenceId);

    // Check if h245Address is initialised.
    if (   h245Address == SIPSocket::Address::Address()
        || port == 0)
    {
        return connect;
    }

  // indicate we are including the optional H245 address in the PDU
  connect.IncludeOptionalField(H225_Connect_UUIE::e_h245Address);

  // convert IP address into the correct H245 type
  connect.m_h245Address.SetTag(H225_TransportAddress::e_ipAddress);
  H225_TransportAddress_ipAddress & ipAddress = connect.m_h245Address;
  ipAddress.m_ip.SetValue((DWORD)h245Address);
  ipAddress.m_port  = port;

  return connect;
}


H225_Setup_UUIE & H323SignalPDU::BuildSetup(const H323Connection & connection,
                                            int callRef,
                                            const SString & displayInfo,
                                            const SString & callingPartyNumber,
                                            const SString & calledPartyNumber,
                                            const SString & destAlias, // destination address
                                            const SString & srcAlias,  // source address
                                            bool canOverlapSend,
                                            bool bSuppressNumber,
                                            bool bSendingComplete)
{
  q931pdu.BuildSetup(callRef, displayInfo, bSendingComplete);

  m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_setup);
  H225_Setup_UUIE & setup = m_h323_uu_pdu.m_h323_message_body;
  setup.m_protocolIdentifier.SetValue(H225_ProtocolID);

  if (!srcAlias.empty()) {
    setup.IncludeOptionalField(H225_Setup_UUIE::e_sourceAddress);
    setup.m_sourceAddress.SetSize(1);
    setup.m_sourceAddress[0].SetTag(H225_AliasAddress::e_h323_ID);
    (SAsnBmpString &)setup.m_sourceAddress[0] = srcAlias;
  }

    if (!callingPartyNumber.empty())
    {
        unsigned plan = 1;             // 1 = ISDN/Telephony numbering system
        unsigned type = 0;             // 0 = Unknown number type
        int presentation = -1;         // 0 = presentation allowed, 1 = presentation restricted, -1 = no octet3a
        int screening = 0;             // 0 = user provided, not screened
        if (bSuppressNumber == true)
        {
            presentation = 1; // presentation restricted
        }
        q931pdu.SetCallingPartyNumber(callingPartyNumber, plan, type, presentation, screening);
    }
    if (!calledPartyNumber.empty())
    {
        q931pdu.SetCalledPartyNumber(calledPartyNumber);
    }


  setup.m_conferenceID = H323GloballyUniqueID();
  setup.m_conferenceGoal.SetTag(H225_Setup_UUIE_conferenceGoal::e_create);
  setup.m_callType.SetTag(H225_CallType::e_pointToPoint);

  setup.m_callIdentifier.m_guid = H323GloballyUniqueID();
  setup.m_mediaWaitForConnect = false;
  setup.m_canOverlapSend = canOverlapSend;

    if (!destAlias.empty())
    {
        if (!calledPartyNumber.empty() && calledPartyNumber[0] != '+')
        {
            setup.IncludeOptionalField(H225_Setup_UUIE::e_destinationAddress);
            setup.m_destinationAddress.SetSize(2);
            H323SetAliasAddress(destAlias, setup.m_destinationAddress[0]);
            H323SetAliasAddress(calledPartyNumber, setup.m_destinationAddress[1]);
        }
        else
        {
            setup.IncludeOptionalField(H225_Setup_UUIE::e_destinationAddress);
            setup.m_destinationAddress.SetSize(1);
            H323SetAliasAddress(destAlias, setup.m_destinationAddress[0]);
        }
    }
    else
    {
        if (!calledPartyNumber.empty())
        {
            setup.IncludeOptionalField(H225_Setup_UUIE::e_destinationAddress);
            setup.m_destinationAddress.SetSize(1);
            H323SetAliasAddress(calledPartyNumber, setup.m_destinationAddress[0]);
        }        
    }

    connection.GetEndPoint().SetEndpointTypeInfo(setup.m_sourceInfo,connection.GetH323Alias());

  return setup;
}


void H323SetAliasAddress(const SString & name, H225_AliasAddress & alias)
{
  alias.SetTag(H225_AliasAddress::e_dialedDigits);
  SAsnIA5String & ia5 = (SAsnIA5String &)alias;
  ia5 = name;
  if (name == ia5.GetValue()) // (SString)ia5 confuses gcc
    return;

  // Could not encode it as a phone number, so do it as a full string.
  alias.SetTag(H225_AliasAddress::e_h323_ID);
  (SAsnBmpString &)alias = name;
}


H225_Setup_UUIE & H323SignalPDU::BuildSetup(const H323Connection & connection,
                                            int callRef,
                                            const SString & origAlias,
                                            const SString & destAlias,
                                            const SIPSocket::Address & destAddr,
                                            WORD port)
{
  H225_Setup_UUIE & setup = BuildSetup(connection, callRef, origAlias, SString(), SString(), SString(), destAlias);

  setup.IncludeOptionalField(H225_Setup_UUIE::e_destCallSignalAddress);
  setup.m_destCallSignalAddress.SetTag(H225_TransportAddress::e_ipAddress);
  H225_TransportAddress_ipAddress & addr = setup.m_destCallSignalAddress;
  addr.m_ip.SetValue((DWORD)destAddr);
  addr.m_port = port;

  return setup;
}


void H323SignalPDU::BuildSetupAck(int callRef)
{
    q931pdu.BuildSetupAck(callRef);
    // Setup Ack has no UUIE but member variables are existing!
    m_h323_uu_pdu.m_h323_message_body.SetTag(UINT_MAX);
}


H225_Progress_UUIE & H323SignalPDU::BuildProgress(const H323Connection          & connection,
                                                  const Q931::ProgressIndicator & ProgressInd,
                                                  const Q931::Cause             & cause)
{
    q931pdu.BuildProgress(connection.GetCallReference(), connection.HadAnsweredCall(), ProgressInd, cause);

    m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_progress);
    H225_Progress_UUIE & progress = m_h323_uu_pdu.m_h323_message_body;

    progress.m_protocolIdentifier.SetValue(H225_ProtocolID);
    connection.GetEndPoint().SetEndpointTypeInfo(progress.m_destinationInfo,connection.GetH323Alias());
    progress.m_callIdentifier.m_guid = connection.GetCallIdentifier();

    return progress;
}


void H323SignalPDU::BuildNotify(const H323Connection              & connection,
                                const Q931::NotificationIndicator & NotificationInd)
{
    q931pdu.BuildNotify(connection.GetCallReference(), connection.HadAnsweredCall(), NotificationInd);
    // Notify has no UUIE but member variables are existing!
    m_h323_uu_pdu.m_h323_message_body.SetTag(UINT_MAX);
}


H225_Information_UUIE & H323SignalPDU::BuildInformation(int callRef,
                                                        bool fromDest,
                                                        const SString & calledPartyNumber,
                                                        const H323GloballyUniqueID & callIdentifier)
{
    q931pdu.BuildInformation(callRef, fromDest);
    q931pdu.SetCalledPartyNumber(calledPartyNumber);

    m_h323_uu_pdu.m_h323_message_body.SetTag(H225_H323_UU_PDU_h323_message_body::e_information);
    H225_Information_UUIE & info = m_h323_uu_pdu.m_h323_message_body;

    info.m_protocolIdentifier.SetValue(H225_ProtocolID);
    info.IncludeOptionalField(H225_Information_UUIE::e_callIdentifier);
    info.m_callIdentifier.m_guid = callIdentifier;

    return info;
}


void H323SignalPDU::BuildStatus(int callRef, bool fromDest)
{
  q931pdu.BuildStatus(callRef, fromDest);
  m_h323_uu_pdu.m_h323_message_body.SetTag(UINT_MAX);
}


void H323SignalPDU::BuildQ931()
{
  // Encode the H225 PDU into the Q931 PDU as User-User data
  SPerStream strm;
  Encode(strm);
  strm.CompleteEncoding();
  q931pdu.SetIE(Q931::UserUserIE, strm);
}


void H323SignalPDU::PrintOn(ostream & strm) const
{
    long long indent = strm.precision() + 2;

    if (q931pdu.HasIE(Q931::UserUserIE))
    {
        strm << "{\n" << STrace::MultiLinePrefix
             << setw(indent+10) << "q931pdu = " << setprecision(indent) << q931pdu << '\n' << STrace::MultiLinePrefix
             << setw(indent+10) << "h225pdu = " << setprecision(indent);
        H225_H323_UserInformation::PrintOn(strm);
        strm << '\n' << STrace::MultiLinePrefix
             << setw(indent-1) << "}";
    }
    else
    {
        strm << "{\n" << STrace::MultiLinePrefix
             << setw(indent+10) << "q931pdu = " << setprecision(indent) << q931pdu << '\n' << STrace::MultiLinePrefix
             << setw(indent-1) << "}";
    }
}

bool H323SignalPDU::Read(H323Transport & transport)
{
    SDEFFUNC("H323SignalPDU::Read");

    SByteVector rawData;

    if (!transport.ReadPDU(rawData)) {
#if STRACING
    if (STrace::CanTrace(g_pH323Module->m_H225TraceId, STrace::Error))
    {
        if (transport.GetErrorCode() != SChannel::Timeout ||
            transport.GetReadTimeout() != 0)
        {
            if (   transport.GetErrorText().empty()
                || transport.GetErrorText() == "Connection aborted"
                || transport.GetErrorText() == "Socket closed or invalid")
            {
                // Some errors are irrelevant
                STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
                    "() Read error: " << transport.GetErrorText());
            }
            else
            {
                // Other errors may be important
                STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
                    "() Read error: " << transport.GetErrorText());
            }
        }
    }
#endif
      return false;
    }

    return Read( transport, rawData );
}


bool H323SignalPDU::Read(H323Transport & transport, SByteVector& rawData )
{
    SDEFFUNC("H323SignalPDU::Read");


  STRACE2_INFO2(g_pH323Module->m_H225TraceId, this, 0,
      "() Received raw data:\n" << STrace::MultiLinePrefix << hex << setfill('0')
      << setprecision(2) << rawData << dec << setfill(' '));

  if (!q931pdu.Decode(rawData)) {
    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
        "() Parse error of Q931 PDU");
    STRACE2_DATA(STrace::Error, g_pH323Module->m_H225TraceId, this, 0,
        (const unsigned char*) &(*(rawData.begin())),rawData.size(), "Hexdump of invalid message: ");
    return false;
  }

    if (!q931pdu.HasIE(Q931::UserUserIE))
    {
        switch (q931pdu.GetMessageType())
        {
            // Some messages have no Q.931 UUIE
            case Q931::SetupAckMsg:
            case Q931::NotifyMsg:    
            case Q931::InformationMsg:
            case Q931::StatusMsg:
            case Q931::StatusEnquiryMsg:
                STRACE2_INFO2(g_pH323Module->m_H225TraceId, this, 0,
                    "() Received PDU:\n" << STrace::MultiLinePrefix << "  "
                    << setprecision(2) << setfill(' ') << *this);
                return true;

            // Otherwise it is an error
            default:
                STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
                    "() No Q931 User-User IE");
                STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
                    "() Received Q.931 PDU:\n" << STrace::MultiLinePrefix << "  "
                    << setprecision(2) << q931pdu);
                return false;
        }
    }

    SPerStream strm = q931pdu.GetIE(Q931::UserUserIE);
    strm.ResetDecoder();

    try
    {
        if (!Decode(strm))
        {
            STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
                "() Read error: PER decode failure if Q.931 User-User IE");
            STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
                "() Partial PDU:\n" << STrace::MultiLinePrefix << "  " << setprecision(2)
                << setfill(' ') << *this);
            return false;
        }
    }

    catch(...)
    {
        STRACE2_FATAL(g_pH323Module->m_H225TraceId, this, 0,
            "() ### Exception in ASN decode thrown! ###");
        return false;
    }

    STRACE2_INFO2(g_pH323Module->m_H225TraceId, this, 0,
        "() Received PDU:\n" << STrace::MultiLinePrefix << "  " << setprecision(2)
        << setfill(' ') << *this);
  return true;
}


bool H323SignalPDU::Write(H323Transport & transport)
{
    SDEFFUNC("H323SignalPDU::Write");

  if (!q931pdu.HasIE(Q931::UserUserIE) && m_h323_uu_pdu.m_h323_message_body.IsValid())
    BuildQ931();

  SByteVector rawData;
  if (!q931pdu.Encode(rawData))
    return false;
  
  STRACE2_INFO2(g_pH323Module->m_H225TraceId, this, 0,
        "() Sending PDU:\n" << STrace::MultiLinePrefix << "  "
        << setprecision(2) << setfill(' ') << *this
        << '\n' << STrace::MultiLinePrefix << hex << setfill('0')
        << setprecision(2) << rawData << dec << setfill(' '));

  return transport.WritePDU(rawData);
}


/////////////////////////////////////////////////////////////////////////////

H245_RequestMessage & H323ControlPDU::Build(H245_RequestMessage::Choices request)
{
  SetTag(e_request);
  H245_RequestMessage & msg = *this;
  msg.SetTag(request);
  return msg;
}


H245_ResponseMessage & H323ControlPDU::Build(H245_ResponseMessage::Choices response)
{
  SetTag(e_response);
  H245_ResponseMessage & resp = *this;
  resp.SetTag(response);
  return resp;
}


H245_CommandMessage & H323ControlPDU::Build(H245_CommandMessage::Choices command)
{
  SetTag(e_command);
  H245_CommandMessage & cmd = *this;
  cmd.SetTag(command);
  return cmd;
}


H245_IndicationMessage & H323ControlPDU::Build(H245_IndicationMessage::Choices indication)
{
  SetTag(e_indication);
  H245_IndicationMessage & ind = *this;
  ind.SetTag(indication);
  return ind;
}


H245_MasterSlaveDetermination & 
      H323ControlPDU::BuildMasterSlaveDetermination(unsigned terminalType,
                                                    unsigned statusDeterminationNumber)
{
  H245_MasterSlaveDetermination & msd = Build(H245_RequestMessage::e_masterSlaveDetermination);
  msd.m_terminalType = terminalType;
  msd.m_statusDeterminationNumber = statusDeterminationNumber;
  return msd;
}


H245_MasterSlaveDeterminationAck &
      H323ControlPDU::BuildMasterSlaveDeterminationAck(bool isMaster)
{
  H245_MasterSlaveDeterminationAck & msda = Build(H245_ResponseMessage::e_masterSlaveDeterminationAck);
  msda.m_decision.SetTag(isMaster
                            ? H245_MasterSlaveDeterminationAck_decision::e_slave
                            : H245_MasterSlaveDeterminationAck_decision::e_master);
  return msda;
}


H245_MasterSlaveDeterminationReject &
      H323ControlPDU::BuildMasterSlaveDeterminationReject(unsigned cause)
{
  H245_MasterSlaveDeterminationReject & msdr = Build(H245_ResponseMessage::e_masterSlaveDeterminationReject);
  msdr.m_cause.SetTag(cause);
  return msdr;
}


H245_TerminalCapabilitySet &
      H323ControlPDU::BuildTerminalCapabilitySet(const H323Connection & connection,
                                                 unsigned sequenceNumber)
{
  H245_TerminalCapabilitySet & cap = Build(H245_RequestMessage::e_terminalCapabilitySet);

  cap.m_sequenceNumber = sequenceNumber;
  cap.m_protocolIdentifier = H245_ProtocolID;

  cap.IncludeOptionalField(H245_TerminalCapabilitySet::e_multiplexCapability);
  cap.m_multiplexCapability.SetTag(H245_MultiplexCapability::e_h2250Capability);
  H245_H2250Capability & h225_0 = cap.m_multiplexCapability;
  h225_0.m_maximumAudioDelayJitter = connection.GetEndPoint().GetMaxAudioDelayJitter();
  h225_0.m_receiveMultipointCapability.m_mediaDistributionCapability.SetSize(1);
  h225_0.m_transmitMultipointCapability.m_mediaDistributionCapability.SetSize(1);
  h225_0.m_receiveAndTransmitMultipointCapability.m_mediaDistributionCapability.SetSize(1);

  // Set the table of capabilities
  connection.GetLocalCapabilities().BuildPDU(cap);

  return cap;
}


H245_TerminalCapabilitySet & H323ControlPDU::BuildEmptyTerminalCapabilitySet(unsigned sequenceNumber)
{
    H245_TerminalCapabilitySet & cap = Build(H245_RequestMessage::e_terminalCapabilitySet);

    cap.m_sequenceNumber = sequenceNumber;
    cap.m_protocolIdentifier = H245_ProtocolID;

    return cap;
}


H245_TerminalCapabilitySetAck &
      H323ControlPDU::BuildTerminalCapabilitySetAck(unsigned sequenceNumber)
{
  H245_TerminalCapabilitySetAck & cap = Build(H245_ResponseMessage::e_terminalCapabilitySetAck);
  cap.m_sequenceNumber = sequenceNumber;
  return cap;
}


H245_TerminalCapabilitySetReject &
      H323ControlPDU::BuildTerminalCapabilitySetReject(unsigned sequenceNumber,
                                                       unsigned cause)
{
  H245_TerminalCapabilitySetReject & cap = Build(H245_ResponseMessage::e_terminalCapabilitySetReject);
  cap.m_sequenceNumber = sequenceNumber;
  cap.m_cause.SetTag(cause);

  return cap;
}


H245_OpenLogicalChannel &
      H323ControlPDU::BuildOpenLogicalChannel(unsigned forwardLogicalChannelNumber)
{
  H245_OpenLogicalChannel & open = Build(H245_RequestMessage::e_openLogicalChannel);
  open.m_forwardLogicalChannelNumber = forwardLogicalChannelNumber;
  return open;
}


H245_RequestChannelClose &
      H323ControlPDU::BuildRequestChannelClose(unsigned channelNumber,
                                               unsigned reason)
{
  H245_RequestChannelClose & rcc = Build(H245_RequestMessage::e_requestChannelClose);
  rcc.m_forwardLogicalChannelNumber = channelNumber;
  rcc.IncludeOptionalField(H245_RequestChannelClose::e_reason);
  rcc.m_reason.SetTag(reason);
  return rcc;
}


H245_CloseLogicalChannel &
      H323ControlPDU::BuildCloseLogicalChannel(unsigned channelNumber)
{
  H245_CloseLogicalChannel & clc = Build(H245_RequestMessage::e_closeLogicalChannel);
  clc.m_forwardLogicalChannelNumber = channelNumber;
  clc.m_source.SetTag(H245_CloseLogicalChannel_source::e_lcse);
  return clc;
}


H245_OpenLogicalChannelAck &
      H323ControlPDU::BuildOpenLogicalChannelAck(unsigned channelNumber)
{
  H245_OpenLogicalChannelAck & ack = Build(H245_ResponseMessage::e_openLogicalChannelAck);
  ack.m_forwardLogicalChannelNumber = channelNumber;
  return ack;
}


H245_OpenLogicalChannelReject &
      H323ControlPDU::BuildOpenLogicalChannelReject(unsigned channelNumber,
                                                    unsigned cause)
{
  H245_OpenLogicalChannelReject & reject = Build(H245_ResponseMessage::e_openLogicalChannelReject);
  reject.m_forwardLogicalChannelNumber = channelNumber;
  reject.m_cause.SetTag(cause);
  return reject;
}


H245_OpenLogicalChannelConfirm &
      H323ControlPDU::BuildOpenLogicalChannelConfirm(unsigned channelNumber)
{
  H245_OpenLogicalChannelConfirm & chan = Build(H245_IndicationMessage::e_openLogicalChannelConfirm);
  chan.m_forwardLogicalChannelNumber = channelNumber;
  return chan;
}


H245_CloseLogicalChannelAck &
      H323ControlPDU::BuildCloseLogicalChannelAck(unsigned channelNumber)
{
  H245_CloseLogicalChannelAck & chan = Build(H245_ResponseMessage::e_closeLogicalChannelAck);
  chan.m_forwardLogicalChannelNumber = channelNumber;
  return chan;
}


H245_RequestChannelCloseAck &
      H323ControlPDU::BuildRequestChannelCloseAck(unsigned channelNumber)
{
  H245_RequestChannelCloseAck & rcca = Build(H245_ResponseMessage::e_requestChannelCloseAck);
  rcca.m_forwardLogicalChannelNumber = channelNumber;
  return rcca;
}


H245_RequestChannelCloseReject &
      H323ControlPDU::BuildRequestChannelCloseReject(unsigned channelNumber)
{
  H245_RequestChannelCloseReject & rccr = Build(H245_ResponseMessage::e_requestChannelCloseReject);
  rccr.m_forwardLogicalChannelNumber = channelNumber;
  return rccr;
}


H245_RequestChannelCloseRelease &
      H323ControlPDU::BuildRequestChannelCloseRelease(unsigned channelNumber)
{
  H245_RequestChannelCloseRelease & rccr = Build(H245_IndicationMessage::e_requestChannelCloseRelease);
  rccr.m_forwardLogicalChannelNumber = channelNumber;
  return rccr;
}


H245_RequestMode & H323ControlPDU::BuildRequestMode(unsigned sequenceNumber)
{
  H245_RequestMode & rm = Build(H245_RequestMessage::e_requestMode);
  rm.m_sequenceNumber = sequenceNumber;

  return rm;
}


H245_RequestModeAck & H323ControlPDU::BuildRequestModeAck(unsigned sequenceNumber,
                                                          unsigned response)
{
  H245_RequestModeAck & ack = Build(H245_ResponseMessage::e_requestModeAck);
  ack.m_sequenceNumber = sequenceNumber;
  ack.m_response.SetTag(response);
  return ack;
}


H245_RequestModeReject & H323ControlPDU::BuildRequestModeReject(unsigned sequenceNumber,
                                                                unsigned cause)
{
  H245_RequestModeReject & reject = Build(H245_ResponseMessage::e_requestModeReject);
  reject.m_sequenceNumber = sequenceNumber;
  reject.m_cause.SetTag(cause);
  return reject;
}


H245_RoundTripDelayRequest &
      H323ControlPDU::BuildRoundTripDelayRequest(unsigned sequenceNumber)
{
  H245_RoundTripDelayRequest & req = Build(H245_RequestMessage::e_roundTripDelayRequest);
  req.m_sequenceNumber = sequenceNumber;
  return req;
}


H245_RoundTripDelayResponse &
      H323ControlPDU::BuildRoundTripDelayResponse(unsigned sequenceNumber)
{
  H245_RoundTripDelayResponse & resp = Build(H245_ResponseMessage::e_roundTripDelayResponse);
  resp.m_sequenceNumber = sequenceNumber;
  return resp;
}


H245_UserInputIndication &
      H323ControlPDU::BuildUserInputIndication(const SString & value)
{
  H245_UserInputIndication & ind = Build(H245_IndicationMessage::e_userInput);
  ind.SetTag(H245_UserInputIndication::e_alphanumeric);
  (SAsnGeneralString &)ind = value;
  return ind;
}


H245_UserInputIndication & H323ControlPDU::BuildUserInputIndication(const SString & tone,
                                                                    unsigned duration,
                                                                    unsigned logicalChannel,
                                                                    unsigned rtpTimestamp)
{
  H245_UserInputIndication & ind = Build(H245_IndicationMessage::e_userInput);

  ind.SetTag(H245_UserInputIndication::e_signal);
  H245_UserInputIndication_signal & sig = ind;

  sig.m_signalType.SetValue(tone);

  if (duration > 0) {
    sig.IncludeOptionalField(H245_UserInputIndication_signal::e_duration);
    sig.m_duration = duration;
  }

  if (logicalChannel > 0) {
    sig.IncludeOptionalField(H245_UserInputIndication_signal::e_rtp);
    sig.m_rtp.m_logicalChannelNumber = logicalChannel;
    sig.m_rtp.m_timestamp = rtpTimestamp;
  }

  return ind;
}


H245_FunctionNotUnderstood &
      H323ControlPDU::BuildFunctionNotUnderstood(const H323ControlPDU & pdu)
{
  H245_FunctionNotUnderstood & fnu = Build(H245_IndicationMessage::e_functionNotUnderstood);

  switch (pdu.GetTag()) {
    case H245_MultimediaSystemControlMessage::e_request :
      fnu.SetTag(H245_FunctionNotUnderstood::e_request);
      (H245_RequestMessage &)fnu = (const H245_RequestMessage &)pdu;
      break;

    case H245_MultimediaSystemControlMessage::e_response :
      fnu.SetTag(H245_FunctionNotUnderstood::e_response);
      (H245_ResponseMessage &)fnu = (const H245_ResponseMessage &)pdu;
      break;

    case H245_MultimediaSystemControlMessage::e_command :
      fnu.SetTag(H245_FunctionNotUnderstood::e_command);
      (H245_CommandMessage &)fnu = (const H245_CommandMessage &)pdu;
      break;
  }

  return fnu;
}


H245_EndSessionCommand & H323ControlPDU::BuildEndSessionCommand(unsigned reason)
{
  H245_EndSessionCommand & end = Build(H245_CommandMessage::e_endSessionCommand);
  end.SetTag(reason);
  return end;
}


bool H323ControlPDU::GetMikey(const H245_TerminalCapabilitySet & termCapSet, SByteVector & rMikey)
{
    SDEFFUNC("H323ControlPDU::GetMikey");

    STRACE2_INFO3(g_pH323Module->m_H225TraceId, 0, 0, "() try to extract Mikey from TermCapSet");

    try
    {
        if (termCapSet.HasOptionalField(H245_TerminalCapabilitySet::e_capabilityTable))
        {
            for (int j = 0; j < termCapSet.m_capabilityTable.GetSize(); j++)
            {
                if (termCapSet.m_capabilityTable[j].HasOptionalField(H245_CapabilityTableEntry::e_capability))
                {
                    const H245_Capability & cap = termCapSet.m_capabilityTable[j].m_capability;
                    if (cap.GetTag() == H245_Capability::e_h235SecurityCapability)
                    {
                        const H245_H235SecurityCapability & secCap = cap;
                        const H245_EncryptionAuthenticationAndIntegrity & encr = secCap.m_encryptionAuthenticationAndIntegrity;
                        if (encr.HasOptionalField(H245_EncryptionAuthenticationAndIntegrity::e_genericH235SecurityCapability))
                        {
                            const H245_GenericCapability & genCap = encr.m_genericH235SecurityCapability;
                            if (genCap.HasOptionalField(H245_GenericCapability::e_nonCollapsing))
                            {
                                for (int k = 0; k < genCap.m_nonCollapsing.GetSize(); k++)
                                {
                                    const H245_ParameterValue & paramValue = genCap.m_nonCollapsing[k].m_parameterValue;
                                    if (paramValue.GetTag() == H245_ParameterValue::e_octetString)
                                    {
                                        // get MIKEY
                                        const SAsnOctetString & mikeyStrm = paramValue;
                                        STRACE2_INFO3(g_pH323Module->m_H225TraceId, 0, 0, "() found mikey!");
                                        rMikey = mikeyStrm.GetValue();

                                        //for (CorNetMediaInfoList::size_type ii = 0; ii < mediaInfoList.size(); ii++)
                                        //{
                                        //    // put MIKEY into each mediainfolist element
                                        //    mediaInfoList[ii].m_bIsOffer = true;
                                        //    mediaInfoList[ii].m_mikey0 = mikeyStrm.GetValue();

                                        //    STRACE2_INFO(m_ulTraceModuleId, this, GetConnId(),
                                        //        "() received mikey, store it: ii=" << ii << " size=" << mediaInfoList[ii].m_mikey0.size());
                                        //}
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    catch(...)
    {
    }

    return true;
}


/////////////////////////////////////////////////////////////////////////////

SString H323GetAliasAddressString(const H225_AliasAddress & alias)
{
  switch (alias.GetTag()) {
    case H225_AliasAddress::e_dialedDigits :
    case H225_AliasAddress::e_url_ID :
    case H225_AliasAddress::e_email_ID :
      return ((const SAsnIA5String &)alias).GetValue();

    case H225_AliasAddress::e_h323_ID :
      return (const SAsnBmpString &)alias;

    case H225_AliasAddress::e_transportID :
      return H323TransportAddress(alias);

    case H225_AliasAddress::e_partyNumber :
    {
      const H225_PartyNumber & party = alias;
      switch (party.GetTag()) {
        case H225_PartyNumber::e_e164Number :
        {
          const H225_PublicPartyNumber & number = party;
          return "PublicParty:" + number.m_publicNumberDigits.GetValue();
        }

        case H225_PartyNumber::e_privateNumber :
        {
          const H225_PrivatePartyNumber & number = party;
          return "PrivateParty:" + number.m_privateNumberDigits.GetValue();
        }

        case H225_PartyNumber::e_dataPartyNumber :
        case H225_PartyNumber::e_telexPartyNumber :
        case H225_PartyNumber::e_nationalStandardPartyNumber :
          return ((const H225_NumberDigits &)party).GetValue();
      }
      break;
    }

    default :
      break;
  }

  return SString();
}


SString H323GetAliasAddressString(const H225_AliasAddress & alias,
                                  H225_AliasAddress::Choices tag)
{
    if (alias.GetTag() == tag)
        return H323GetAliasAddressString(alias);

    return SString();
}
    

void H323GetAliasAddressStringVector(
    const H225_ArrayOf_AliasAddress & rAliasArray,
    H225_AliasAddress::Choices tag,
    SStringVector & rAliases)
{
    for (int i = 0; i < rAliasArray.GetSize(); i++)
    {
        SString Alias = H323GetAliasAddressString(rAliasArray[i], tag);
        if (!Alias.empty())
            rAliases.push_back(Alias);
    }
}


H225_TransportAddress H323GetTransportAddress(
    const H225_ArrayOf_TransportAddress & AddressArray,
    H225_TransportAddress::Choices tag)
{
    for (int i = 0; i < AddressArray.GetSize(); i++)
    {
        if (AddressArray[i].GetTag() == tag)
            return AddressArray[i];
    }

    return H225_TransportAddress();
}


/////////////////////////////////////////////////////////////////////////////

H225_GatekeeperRequest & H323RasPDU::BuildGatekeeperRequest(unsigned seqNum)
{
  SetTag(e_gatekeeperRequest);
  H225_GatekeeperRequest & grq = *this;
  grq.m_requestSeqNum = seqNum;
  grq.m_protocolIdentifier.SetValue(H225_ProtocolID);
  return grq;
}


H225_GatekeeperConfirm & H323RasPDU::BuildGatekeeperConfirm(unsigned seqNum)
{
    SetTag(e_gatekeeperConfirm);
    H225_GatekeeperConfirm & gcf = *this;
    gcf.m_requestSeqNum = seqNum;
    gcf.m_protocolIdentifier.SetValue(H225_ProtocolID);
    return gcf;
}


H225_GatekeeperReject & H323RasPDU::BuildGatekeeperReject(unsigned seqNum,
                                                          unsigned reason)
{
    SetTag(e_gatekeeperReject);
    H225_GatekeeperReject & grj = *this;
    grj.m_requestSeqNum = seqNum;
    grj.m_protocolIdentifier.SetValue(H225_ProtocolID);
    grj.m_rejectReason.SetTag(reason);
    return grj;
}


H225_RegistrationRequest & H323RasPDU::BuildRegistrationRequest(unsigned seqNum)
{
  SetTag(e_registrationRequest);
  H225_RegistrationRequest & rrq = *this;
  rrq.m_requestSeqNum = seqNum;
  rrq.m_protocolIdentifier.SetValue(H225_ProtocolID);
  return rrq;
}


H225_RegistrationConfirm & H323RasPDU::BuildRegistrationConfirm(unsigned seqNum)
{
    SetTag(e_registrationConfirm);
    H225_RegistrationConfirm & rcf = *this;
    rcf.m_requestSeqNum = seqNum;
    rcf.m_protocolIdentifier.SetValue(H225_ProtocolID);
    return rcf;
}


H225_RegistrationReject & H323RasPDU::BuildRegistrationReject(unsigned seqNum,
                                                              unsigned reason)
{
    SetTag(e_registrationReject);
    H225_RegistrationReject & rrj = *this;
    rrj.m_requestSeqNum = seqNum;
    rrj.m_protocolIdentifier.SetValue(H225_ProtocolID);
    rrj.m_rejectReason.SetTag(reason);
    return rrj;
}

        
H225_UnregistrationRequest & H323RasPDU::BuildUnregistrationRequest(unsigned seqNum)
{
  SetTag(e_unregistrationRequest);
  H225_UnregistrationRequest & urq = *this;
  urq.m_requestSeqNum = seqNum;
  return urq;
}


H225_UnregistrationConfirm & H323RasPDU::BuildUnregistrationConfirm(unsigned seqNum)
{
  SetTag(e_unregistrationConfirm);
  H225_UnregistrationConfirm & ucf = *this;
  ucf.m_requestSeqNum = seqNum;
  return ucf;
}


H225_UnregistrationReject & H323RasPDU::BuildUnregistrationReject(unsigned seqNum,
                                                                  unsigned reason)
{
    SetTag(e_unregistrationReject);
    H225_UnregistrationReject & urj = *this;
    urj.m_requestSeqNum = seqNum;
    urj.m_rejectReason.SetTag(reason);
    return urj;
}


H225_LocationRequest & H323RasPDU::BuildLocationRequest(unsigned seqNum)
{
  SetTag(e_locationRequest);
  H225_LocationRequest & lrq = *this;
  lrq.m_requestSeqNum = seqNum;
  return lrq;
}


H225_LocationConfirm & H323RasPDU::BuildLocationConfirm(unsigned seqNum)
{
    SetTag(e_locationConfirm);
    H225_LocationConfirm & lcf = *this;
    lcf.m_requestSeqNum = seqNum;
    return lcf;
}


H225_LocationReject & H323RasPDU::BuildLocationReject(unsigned seqNum,
                                                      unsigned reason)
{
    SetTag(e_locationReject);
    H225_LocationReject & lrj = *this;
    lrj.m_requestSeqNum = seqNum;
    lrj.m_rejectReason.SetTag(reason);
    return lrj;
}


H225_AdmissionRequest & H323RasPDU::BuildAdmissionRequest(unsigned seqNum)
{
  SetTag(e_admissionRequest);
  H225_AdmissionRequest & arq = *this;
  arq.m_requestSeqNum = seqNum;
  return arq;
}


H225_AdmissionConfirm & H323RasPDU::BuildAdmissionConfirm(unsigned seqNum,
                                                          unsigned bandWidth)
{
    SetTag(e_admissionConfirm);
    H225_AdmissionConfirm & acf = *this;
    acf.m_requestSeqNum = seqNum;
    acf.m_bandWidth = bandWidth;
    return acf;
}


H225_AdmissionReject & H323RasPDU::BuildAdmissionReject(unsigned seqNum,
                                                        unsigned reason)
{
    SetTag(e_admissionReject);
    H225_AdmissionReject & arj = *this;
    arj.m_requestSeqNum = seqNum;
    arj.m_rejectReason.SetTag(reason);
    return arj;
}


H225_DisengageRequest & H323RasPDU::BuildDisengageRequest(unsigned seqNum)
{
  SetTag(e_disengageRequest);
  H225_DisengageRequest & drq = *this;
  drq.m_requestSeqNum = seqNum;
  return drq;
}


H225_DisengageConfirm & H323RasPDU::BuildDisengageConfirm(unsigned seqNum)
{
  SetTag(e_disengageConfirm);
  H225_DisengageConfirm & dcf = *this;
  dcf.m_requestSeqNum = seqNum;
  return dcf;
}


H225_DisengageReject & H323RasPDU::BuildDisengageReject(unsigned seqNum,
                                                        unsigned reason)
{
    SetTag(e_disengageReject);
    H225_DisengageReject & drj = *this;
    drj.m_requestSeqNum = seqNum;
    drj.m_rejectReason.SetTag(reason);
    return drj;
}


H225_BandwidthRequest & H323RasPDU::BuildBandwidthRequest(unsigned seqNum)
{
  SetTag(e_bandwidthRequest);
  H225_BandwidthRequest & brq = *this;
  brq.m_requestSeqNum = seqNum;
  return brq;
}


H225_BandwidthConfirm & H323RasPDU::BuildBandwidthConfirm(unsigned seqNum,
                                                          unsigned bandWidth)
{
  SetTag(e_bandwidthConfirm);
  H225_BandwidthConfirm & bcf = *this;
  bcf.m_requestSeqNum = seqNum;
  bcf.m_bandWidth = bandWidth;
  return bcf;
}


H225_BandwidthReject & H323RasPDU::BuildBandwidthReject(unsigned seqNum,
                                                        unsigned reason)
{
  SetTag(e_bandwidthReject);
  H225_BandwidthReject & brj = *this;
  brj.m_requestSeqNum = seqNum;
  brj.m_rejectReason.SetTag(reason);
  return brj;
}


H225_InfoRequestResponse & H323RasPDU::BuildInfoRequestResponse(unsigned seqNum)
{
  SetTag(e_infoRequestResponse);
  H225_InfoRequestResponse & irr = *this;
  irr.m_requestSeqNum = seqNum;
  return irr;
}


H225_UnknownMessageResponse & H323RasPDU::BuildUnknownMessageResponse(unsigned seqNum)
{
  SetTag(e_unknownMessageResponse);
  H225_UnknownMessageResponse & umr = *this;
  umr.m_requestSeqNum = seqNum;
  return umr;
}


H323RasPDU::H323RasPDUError H323RasPDU::Read(H323Transport & transport)
{
    SDEFFUNC("H323RasPDU::Read");
    SPerStream strm;
    if (!transport.ReadPDU(strm))
    {
        if (STrace::CanTrace(g_pH323Module->m_RasTraceId, STrace::Error))
        {
            if (   transport.GetErrorText().empty()
                || transport.GetErrorText() == "Interrupted function call"
                || transport.GetErrorText() == "Connection reset")
            {
                // Some errors are irrelevant
                return DiscardLastAction;
            }
            else
            {
                // Other errors may be important
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() Read error: " << transport.GetErrorText());
            }
        }
        return TransportFailure;
    }

    strm.ResetDecoder();

    // Hexdump of RAS message to find reason for possible exceptions
    STRACE2_DATA(STrace::Info3, g_pH323Module->m_RasTraceId, this, 0,
        &(*(strm.begin())), strm.size(), "RAS message:");
    
    try
    {
        if (!Decode(strm))
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Read error: PER decode failure");
            STRACE2_INFO(g_pH323Module->m_RasTraceId, this, 0,
                "() Received PDU:\n" << STrace::MultiLinePrefix << "  "
                << dec << setfill(' ') << setprecision(2) << *this
                << "\n" << STrace::MultiLinePrefix << hex << setfill('0')
                << setprecision(2) << strm << setfill(' '));
            return DecodeFailure;
        }
        if (STrace::CanTrace(g_pH323Module->m_RasTraceId, STrace::Debug))
        {
            STRACE2_INFO(g_pH323Module->m_RasTraceId, this, 0,
                "() Received PDU:\n" << STrace::MultiLinePrefix << "  "
                << dec << setfill(' ') << setprecision(2) << *this
                << "\n" << STrace::MultiLinePrefix << hex << setfill('0')
                << setprecision(2) << strm << setfill(' '));
        }
        return Success;
    }

    catch(...)
    {
        STRACE2_FATAL(g_pH323Module->m_RasTraceId, this, 0,
            "() ### Exception in ASN decode thrown! ###");
        return DecodeFailure;
    }
}


H323RasPDU::H323RasPDUError H323RasPDU::Write(H323Transport & transport) const
{
    SDEFFUNC("H323RasPDU::Write");

    SPerStream strm;

    try
    {
        strm.BeginEncoding();
        Encode(strm);
        strm.CompleteEncoding();
    }

    catch(...)
    {
        STRACE2_FATAL(g_pH323Module->m_RasTraceId, this, 0,
            "() ### Exception in ASN encode thrown! ###");
        return EncodeFailure;
    }

    // Hexdump of RAS message to find reason for possible exceptions
    STRACE2_DATA(STrace::Info3, g_pH323Module->m_RasTraceId, this, 0,
        &(*(strm.begin())), strm.size(), "RAS message:");

    if (STrace::CanTrace(g_pH323Module->m_RasTraceId, STrace::Debug))
    {
        STRACE2_INFO(g_pH323Module->m_RasTraceId, this, 0,
            "() Sending PDU:\n" << STrace::MultiLinePrefix
            << "  " << setprecision(2) << setfill(' ') << *this
            << "\n" << STrace::MultiLinePrefix << setprecision(2)
            << hex << setfill('0') << strm << setfill(' '));
    }

    return ( transport.WritePDU(strm) == true ? Success : TransportFailure);
}


/////////////////////////////////////////////////////////////////////////////
