/*---------------------------------------------------------------------------

   Swyx Solutions GmbH
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Solutions GmbH are
  Copyright (C) 2002-2005 Swyx Solutions GmbH.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323 T.38
 
    Comments: H.323 T.38 logical channel establishment

  ---------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/h323t38.cpp                $
    $Author:: Ulrich.findeisen                                             $
      $Date:: 14.06.05 17:49                                               $
   $Modtime:: 14.06.05 13:04                                               $
  $Revision:: 2                                                            $
  $Workfile:: h323t38.cpp                                                  $
$NoKeywords: $
----------------------------------------------------------------------------*/

#include <stlib.h>
#include "h245.h"
#include "h323caps.h"
#include "h323t38.h"
#include "h323.h"

using namespace std;
using namespace Tools;

#define new SNEW

#define FAX_BIT_RATE 144  //14.4k

//-----------------------------------------------------------------------------

H323_T38CapParams::H323_T38CapParams()
  : m_uVersion(0),
    m_RateManagement(T38TransferredTCF),
    m_uMaxBitRate(FAX_BIT_RATE),
    m_bFillBitRemoval(false),
    m_bTranscodingJBIG(false),
    m_bTranscodingMMR(false),
    m_uUDPMaxBufferSize(200),
    m_uUDPMaxDatagramSize(72),
    m_UDPErrCorrection(T38UDPRedundancy)
{
}

H323_T38CapParams::~H323_T38CapParams()
{
}

H323_T38CapParams & H323_T38CapParams::operator=(const H323_T38CapParams & params)
{
    m_uVersion              = params.m_uVersion;
    m_RateManagement        = params.m_RateManagement;
    m_uMaxBitRate           = params.m_uMaxBitRate;
    m_bFillBitRemoval       = params.m_bFillBitRemoval;
    m_bTranscodingJBIG      = params.m_bTranscodingJBIG; 
    m_bTranscodingMMR       = params.m_bTranscodingMMR;
    m_uUDPMaxBufferSize     = params.m_uUDPMaxBufferSize;
    m_uUDPMaxDatagramSize   = params.m_uUDPMaxDatagramSize;
    m_UDPErrCorrection      = params.m_UDPErrCorrection;

    return *this;
}

void H323_T38CapParams::PrintOn(std::ostream & strm) const
{
    strm << "v=" << m_uVersion;
    switch (m_RateManagement)
    {
        case T38LocalTCF:
            strm << "|RateMgmt=localTCF";
            break;
        case T38TransferredTCF:
            strm << "|RateMgmt=transferredTCF";
            break;
        default:
            strm << "|RateMgmt=UNKNOWN";
            break;
    }
    strm << "|MaxBitRate="      << m_uMaxBitRate
        << "|FillBitRemoval="   << m_bFillBitRemoval
        << "|TranscodJBIG="     << m_bTranscodingJBIG
        << "|TranscodMMR="      << m_bTranscodingMMR
        << "|UDPMaxBufSize="    << m_uUDPMaxBufferSize
        << "|UDPMaxDatagrSize=" << m_uUDPMaxDatagramSize;
    switch(m_UDPErrCorrection)
    {
        case T38UDPFEC:
            strm << "|UDPErrCorr=FEC";
            break;
        case T38UDPRedundancy:
            strm << "|UDPErrCorr=Redundancy";
            break;
        default:
            strm << "|UDPErrCorr=UNKNOWN";
            break;
    }
}

//-----------------------------------------------------------------------------

H323_T38Capability::H323_T38Capability(TransportMode m)
  : H323DataCapability(FAX_BIT_RATE),
    mode(m)
{
    SDEFFUNC("H323_T38Capability::H323_T38Capab");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, 0,
        "()");
}


SObject::Comparison H323_T38Capability::Compare(const SObject & obj) const
{
  Comparison result = H323DataCapability::Compare(obj);
  if (result != EqualTo)
    return result;

  SAssert(obj.IsDescendant( typeid( *this ) ), SInvalidCast);
  const H323_T38Capability & other = (const H323_T38Capability &)obj;

  if (mode < other.mode)
    return LessThan;

  if (mode > other.mode)
    return GreaterThan;

  return EqualTo;
}


SObject * H323_T38Capability::Clone() const
{
  return new H323_T38Capability(*this);
}


void H323_T38Capability::PrintOn(ostream & strm) const
{
    strm << GetFormatName() << " ";
    H323DataCapability::PrintOn(strm);
}


unsigned H323_T38Capability::GetSubType() const
{
  return H245_DataApplicationCapability_application::e_t38fax;
}


SString H323_T38Capability::GetFormatName() const
{
  static const char * const modes[NumTransportModes] = {
    "UDP", "TCPuni", "TCPbi"
  };
  return SString("T.38-") + modes[mode];
}


H323Channel * H323_T38Capability::CreateChannel(H323Connection & connection,
                                                H323Channel::Directions direction,
                                                unsigned int sessionID,
                             const H245_H2250LogicalChannelParameters * /*params*/) const
{
    SDEFFUNC("H323_T38Capability::CreateChannel");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "() CreateChannel, sessionID=" << sessionID << " direction=" << direction);

    return new H323_T38Channel(connection, *this, direction, sessionID, mode);
}


bool H323_T38Capability::OnSendingPDU(H245_DataApplicationCapability & pdu) const
{
    SDEFFUNC("H323_T38Capability::OnSendingPDU1");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, 0,
        "() OnSendingPDU for capability");

    pdu.m_maxBitRate = m_T38CapParams.m_uMaxBitRate;
    pdu.m_application.SetTag(H245_DataApplicationCapability_application::e_t38fax);
    H245_DataApplicationCapability_application_t38fax & fax = pdu.m_application;
    return OnSendingPDU(fax.m_t38FaxProtocol, fax.m_t38FaxProfile);
}


bool H323_T38Capability::OnSendingPDU(H245_DataMode & pdu) const
{
    SDEFFUNC("H323_T38Capability::OnSendingPDU2");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, 0,
        "()");

    pdu.m_bitRate = m_T38CapParams.m_uMaxBitRate;
    pdu.m_application.SetTag(H245_DataMode_application::e_t38fax);
    H245_DataMode_application_t38fax & fax = pdu.m_application;
    return OnSendingPDU(fax.m_t38FaxProtocol, fax.m_t38FaxProfile);
}


bool H323_T38Capability::OnSendingPDU(H245_DataProtocolCapability & proto,
                                      H245_T38FaxProfile & profile) const
{
    SDEFFUNC("H323_T38Capability::OnSendingPDU3");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, 0,
        "()");

    if (mode == e_UDP)
    {
        proto.SetTag(H245_DataProtocolCapability::e_udp);
        switch(m_T38CapParams.m_RateManagement)
        {
          case H323_T38CapParams::T38LocalTCF:
            profile.m_t38FaxRateManagement.SetTag(H245_T38FaxRateManagement::e_localTCF); // recommended for TCP
            break;
          case H323_T38CapParams::T38TransferredTCF:
          default:
            profile.m_t38FaxRateManagement.SetTag(H245_T38FaxRateManagement::e_transferredTCF); // recommended for UDP
            break;
        }
        profile.IncludeOptionalField(H245_T38FaxProfile::e_t38FaxUdpOptions);
        profile.m_t38FaxUdpOptions.IncludeOptionalField(H245_T38FaxUdpOptions::e_t38FaxMaxBuffer);
        profile.m_t38FaxUdpOptions.m_t38FaxMaxBuffer = m_T38CapParams.m_uUDPMaxBufferSize;
        profile.m_t38FaxUdpOptions.IncludeOptionalField(H245_T38FaxUdpOptions::e_t38FaxMaxDatagram);
        profile.m_t38FaxUdpOptions.m_t38FaxMaxDatagram = m_T38CapParams.m_uUDPMaxDatagramSize;

        switch(m_T38CapParams.m_UDPErrCorrection)
        {
          case H323_T38CapParams::T38UDPFEC:
              profile.m_t38FaxUdpOptions.m_t38FaxUdpEC.SetTag(H245_T38FaxUdpOptions_t38FaxUdpEC::e_t38UDPFEC);
            break;
          case H323_T38CapParams::T38UDPRedundancy:
          default:
            profile.m_t38FaxUdpOptions.m_t38FaxUdpEC.SetTag(H245_T38FaxUdpOptions_t38FaxUdpEC::e_t38UDPRedundancy);
            break;
        }
    }
    else
    {
        proto.SetTag(H245_DataProtocolCapability::e_tcp);
        switch(m_T38CapParams.m_RateManagement)
        {
          case H323_T38CapParams::T38TransferredTCF:
            profile.m_t38FaxRateManagement.SetTag(H245_T38FaxRateManagement::e_transferredTCF); // recommended for UDP
            break;
          case H323_T38CapParams::T38LocalTCF:
          default:
            profile.m_t38FaxRateManagement.SetTag(H245_T38FaxRateManagement::e_localTCF); // recommended for TCP
            break;
        }
        profile.IncludeOptionalField(H245_T38FaxProfile::e_t38FaxTcpOptions);
        profile.m_t38FaxTcpOptions.m_t38TCPBidirectionalMode = mode == e_SingleTCP;
    }
    return true;
}


bool H323_T38Capability::OnReceivedPDU(const H245_DataApplicationCapability & cap)
{
    SDEFFUNC("H323_T38Capability::OnReceivedPDU");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "() capability=" << *this << ", mode=" << mode);

    if (cap.m_application.GetTag() != H245_DataApplicationCapability_application::e_t38fax)
    {
      STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
          "() DataApplicationCapability is not T.38 Fax!");
      return false;
    }

    const H245_DataApplicationCapability_application_t38fax & fax = cap.m_application;
    const H245_DataProtocolCapability & proto = fax.m_t38FaxProtocol;
    const H245_T38FaxProfile & profile = fax.m_t38FaxProfile;

    // Map information to T.38 capability parameter class
    if (profile.HasOptionalField(H245_T38FaxProfile::e_version))
        m_T38CapParams.m_uVersion = profile.m_version;
    if (profile.HasOptionalField(H245_T38FaxProfile::e_t38FaxRateManagement))
    {
        switch(profile.m_t38FaxRateManagement.GetTag())
        {
            case H245_T38FaxRateManagement::e_localTCF:
                m_T38CapParams.m_RateManagement = H323_T38CapParams::T38LocalTCF;
                break;
            case H245_T38FaxRateManagement::e_transferredTCF:
                m_T38CapParams.m_RateManagement = H323_T38CapParams::T38TransferredTCF;
                break;
            default:
                STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
                    "() Unknown T. 38 fax rate management tag!");
                return false;
        }
    }
    m_T38CapParams.m_uMaxBitRate = cap.m_maxBitRate;
    m_T38CapParams.m_bFillBitRemoval = profile.m_fillBitRemoval;
    m_T38CapParams.m_bTranscodingJBIG = profile.m_transcodingJBIG;
    m_T38CapParams.m_bTranscodingMMR = profile.m_transcodingMMR;

    switch(proto.GetTag())
    {
        case H245_DataProtocolCapability::e_udp:
            if (mode != e_UDP)
            {
                STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
                    "() Local capability mode is not T.38 UDP!");
                return false;
            }
            // Map information to T.38 capability parameter class
            if (profile.HasOptionalField(H245_T38FaxProfile::e_t38FaxUdpOptions))
            {
                if (profile.m_t38FaxUdpOptions.HasOptionalField(H245_T38FaxUdpOptions::e_t38FaxMaxBuffer))
                    m_T38CapParams.m_uUDPMaxBufferSize = profile.m_t38FaxUdpOptions.m_t38FaxMaxBuffer;
                if (profile.m_t38FaxUdpOptions.HasOptionalField(H245_T38FaxUdpOptions::e_t38FaxMaxDatagram))
                    m_T38CapParams.m_uUDPMaxDatagramSize = profile.m_t38FaxUdpOptions.m_t38FaxMaxDatagram;
                switch(profile.m_t38FaxUdpOptions.m_t38FaxUdpEC.GetTag())
                {
                    case H245_T38FaxUdpOptions_t38FaxUdpEC::e_t38UDPFEC:
                        m_T38CapParams.m_UDPErrCorrection = H323_T38CapParams::T38UDPFEC;
                        break;
                    case H245_T38FaxUdpOptions_t38FaxUdpEC::e_t38UDPRedundancy:
                        m_T38CapParams.m_UDPErrCorrection = H323_T38CapParams::T38UDPRedundancy;
                        break;
                    default:
                        STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
                            "() Unknown T.38 fax UDP options error correction tag!");
                        return false;
                }
            }
            break;

        case H245_DataProtocolCapability::e_tcp:
        {
            if (   profile.HasOptionalField(H245_T38FaxProfile::e_t38FaxTcpOptions)
                && profile.m_t38FaxTcpOptions.m_t38TCPBidirectionalMode == true)
            {
                if (mode != e_SingleTCP) // one bi-directional channel
                {
                    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
                        "() Local capability mode is not T.38 bi-directional TCP!");
                    return false;
                }
            }
            else
            {
                // If optional field t38FaxTcpOptions is not present "assume that a
                // single bi-directional reliable channel for T.38 is not supported".
                // (H.323v4 Annex D chap. D.3.2)
                //
                if (mode != e_DualTCP) // two unidirectional channels
                {
                    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
                        "() Local capability mode is not T.38 unidirectional TCP!");
                    return false;
                }
            }
        }
        break;

        default:
            STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
                "() T.38 DataProtocolCapability is neither UDP nor TCP!");
            return false;
    }
    return true;
}

//-----------------------------------------------------------------------------

H323_T38Channel::H323_T38Channel(H323Connection & connection,
                                 const H323Capability & capability,
                                 H323Channel::Directions dir,
                                 unsigned sessionID,
                                 H323_T38Capability::TransportMode mode)
  : H323DataChannel(connection, capability, dir, sessionID)
{
    SDEFFUNC("H323_T38Channel::H323_T38Channel");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "(" << capability << ") H323 T.38 channel created.");

    separateReverseChannel = mode != H323_T38Capability::e_SingleTCP;
    usesTCP = mode != H323_T38Capability::e_UDP;
}


H323_T38Channel::~H323_T38Channel()
{
    SDEFFUNC("H323_T38Channel::~H323_T38Channel");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "()");
}


void H323_T38Channel::CleanUpOnTermination()
{
    SDEFFUNC("H323_T38Channel::CleanUpOnTermin");

    if (terminating)
        return;

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "()");

    H323DataChannel::CleanUpOnTermination();
}


bool H323_T38Channel::OnSendingPDU(H245_OpenLogicalChannel & open) const
{
    SDEFFUNC("H323_T38Channel::OnSendingPDU");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "()");

    return H323DataChannel::OnSendingPDU(open);
}


bool H323_T38Channel::OnReceivedPDU(const H245_OpenLogicalChannel & open,
                                    unsigned & errorCode)
{
    SDEFFUNC("H323_T38Channel::OnReceivedPDU");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "()");

    return H323DataChannel::OnReceivedPDU(open, errorCode);
}


void H323_T38Channel::Receive()
{
    SDEFFUNC("H323_T38Channel::Receive");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Not implemented!");

    return;
}


void H323_T38Channel::Transmit()
{
    SDEFFUNC("H323_T38Channel::Transmit");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Not implemented!");

    return;
}


//-----------------------------------------------------------------------------

