/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Transports
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxGate.root/H323/source/transports.cpp            $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 4.01.06 10:51                                                $
   $Modtime:: 3.01.06 17:03                                                $
  $Revision:: 6                                                            $
  $Workfile:: transports.cpp                                               $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "h323.h"
#include "h323pdu.h"
#include "STlsConnection.h"

using namespace std;
using namespace Tools;


class H225TransportThread : public SThread
{
  public:
    H225TransportThread(H323EndPoint & endpoint,
                        H323Transport * pTransport);

  protected:
    void Main();

    H323Transport * m_pTransport;
};


class H245TransportThread : public SThread
{
  public:
    H245TransportThread(H323EndPoint & endpoint,
                        H323Connection & connection,
                        H323Transport & transport);

  protected:
    void Main();

    H323Connection & connection;
    H323Transport  & transport;
};


#define new SNEW


/////////////////////////////////////////////////////////////////////////////

H225TransportThread::H225TransportThread(H323EndPoint & endpoint,
                                         H323Transport * pTransport)
  : SThread(endpoint.GetSignallingThreadStackSize(), AutoDeleteThread),
    m_pTransport(pTransport)
{
    if (!SBAD_READPTR(m_pTransport))
    {
        m_pTransport->AttachThread(this);
    }
    Resume();
}


void H225TransportThread::Main()
{
    SDEFFUNC("H225TransportThread::Main");

    unsigned long dummyConnectionId = 0;
    if (!SBAD_READPTR(m_pTransport))
        dummyConnectionId = (unsigned long)m_pTransport->GetReadChannel();

    STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, dummyConnectionId,
        "() Started accept thread (H225TransportThread)");

    if (SBAD_READPTR(m_pTransport))
    {
        STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, dummyConnectionId,
            "() Invalid transport pointer! Finished accept thread (H225TransportThread)");
        return;
    }
    if (m_pTransport->HandleFirstSignallingChannelPDU() == false)
    {
        m_pTransport->Close();
        delete m_pTransport;
        m_pTransport = 0;
    }
    STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, dummyConnectionId,
        "() Finished accept thread (H225TransportThread)");
}


/////////////////////////////////////////////////////////////////////////////

H245TransportThread::H245TransportThread(H323EndPoint & endpoint,
                                         H323Connection & c,
                                         H323Transport & t)
  : SThread(endpoint.GetSignallingThreadStackSize(), NoAutoDeleteThread),
    connection(c),
    transport(t)
{
  transport.AttachThread(this);
  Resume();
}


void H245TransportThread::Main()
{
    SDEFFUNC("H245TransportThread::Main");

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, 0,
        "() Started thread (H245TransportThread)");

  if (transport.AcceptControlChannel(connection))
    connection.HandleControlChannel();

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, 0,
        "() Finished thread (H245TransportThread)");
}


/////////////////////////////////////////////////////////////////////////////

H323TransportAddress::H323TransportAddress(const char * cstr)
  : SString(cstr)
{
  Validate();
}


H323TransportAddress::H323TransportAddress(const SString & str)
  : SString(str)
{
  Validate();
}


H323TransportAddress::H323TransportAddress(const H225_TransportAddress & transport)
  : SString("ip:")
{
  switch (transport.GetTag()) {
    case H225_TransportAddress::e_ipAddress :
    {
      if (!transport.IsValid())
      {
          *this += "127.0.0.1:0";
      }
      else
      {
          const H225_TransportAddress_ipAddress & ip = transport;
          SIpAddress ipAddr = (DWORD)ip.m_ip;
          ipAddr.m_Port = ip.m_port;
          *this += ipAddr.AsString();
      }
      break;
    }
  }
}


H323TransportAddress::H323TransportAddress(const H245_TransportAddress & transport)
  : SString("ip:")
{
  switch (transport.GetTag()) {
    case H245_TransportAddress::e_unicastAddress :
    {
      const H245_UnicastAddress & unicast = transport;
      switch (transport.GetTag()) {
        case H245_UnicastAddress::e_iPAddress :
        {
          const H245_UnicastAddress_iPAddress & ip = unicast;
          *this += ip.m_network.AsString();
          *this += ":";
          *this += SString(SString::Unsigned, (unsigned)ip.m_tsapIdentifier);
          break;
        }
      }
      break;
    }
  }
}


H323TransportAddress::H323TransportAddress(const SIPSocket::Address & ip, WORD port)
  : SString("ip:")
{
  *this += SIPSocket::GetHostName(ip);
  if (port != 0)
  {
    *this += ":";
    *this += SString(SString::Unsigned, (unsigned)port);
  }
}


void H323TransportAddress::Validate()
{
  if (empty())
    return;

  if (find("ip:") == 0) {
    return;
  }

  *this = SString();
}


bool H323TransportAddress::SetPDU(H225_TransportAddress & pdu) const
{
  SIPSocket::Address ip;
  WORD port;
  if (GetIpAndPort(ip, port)) {
    pdu.SetTag(H225_TransportAddress::e_ipAddress);
    H225_TransportAddress_ipAddress & addr = pdu;
    addr.m_ip.SetValue((DWORD)ip);
    addr.m_port = port;
    return true;
  }


  return false;
}


bool H323TransportAddress::SetPDU(H245_TransportAddress & pdu) const
{
  SIPSocket::Address ip;
  WORD port;
  if (GetIpAndPort(ip, port)) {
    pdu.SetTag(H245_TransportAddress::e_unicastAddress);

    H245_UnicastAddress & unicast = pdu;
    unicast.SetTag(H245_UnicastAddress::e_iPAddress);

    H245_UnicastAddress_iPAddress & addr = unicast;
    addr.m_network.SetValue((DWORD)ip);
    addr.m_tsapIdentifier = port;
    return true;
  }

  return false;
}


bool H323TransportAddress::GetIpAndPort(SIPSocket::Address & ip,
                                        WORD & port,
                                        const char * proto) const
{
    SDEFFUNC("H323TransportAddr::GetIpAndPort");

  if (find("ip:") != 0) {
    STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, 0,
        "() Use of non IP transport address: \"" << *this << '"');
    return false;
  }

  SString host;
  int colon = find(':', 3);
  if (colon == npos)
    host = substr(3);
  else {
    host = substr(3, colon-3);
    SString str = substr(colon+1);
    port = SIPSocket::GetPortByService(proto, substr(colon+1));
  }

  if (host.empty() || port == 0) {
    STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, 0,
        "() Illegal IP transport address: \"" << *this << '"');
    return false;
  }

  if (SIPSocket::GetHostAddress(host, ip))
    return true;

  STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, 0,
      "() Could not find host : \"" << host << '"');
  return false;
}


/////////////////////////////////////////////////////////////////////////////

H323Listener::H323Listener(H323EndPoint & end)
  : SThread(end.GetListenerThreadStackSize(), NoAutoDeleteThread),
    endpoint(end)
{
    SConfig config(SConfig::Machine);
    m_bTraceFullIP = config.GetBoolean("TraceFullIP", false);
}


/////////////////////////////////////////////////////////////////////////////

// we're storing the this pointer only. Therefore the warning can be ignored
#pragma warning(disable:4355)
H323Transport::H323Transport(H323EndPoint & end)
  : endpoint(end)
#ifdef _DEBUG // add name to H323 transport mutex
  , m_Mutex(Tools::SString(Tools::SString::Printf, "H323Transport::Mutex[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#endif
#pragma warning(default:4355)
{
  thread = NULL;
  m_bIsClosed = false;
}


H323Transport::~H323Transport()
{
  SIndirectChannel::Close();
}


bool H323Transport::Close()
{
  /* Do not use SIndirectChannel::Close() as this deletes the sub-channel
     member field crashing the background thread. Just close the base
     sub-channel so breaks the threads I/O block.
   */
  if (IsOpen())
  {    
      if (!SBAD_READPTR(GetBaseReadChannel()))
        GetBaseReadChannel()->Close();
  }

  return true;
}

bool H323Transport::HandleFirstSignallingChannelPDU()
{
    SDEFFUNC("H323Transport::HdlFirstSigChanPDU");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
        "() Awaiting first PDU, socket " << GetReadChannel());

  H323SignalPDU pdu;
  if (!pdu.Read(*this)) {
    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
        "() Failed to get initial Q.931 PDU, connection not started.");
    return false;
  }

  unsigned callReference = pdu.GetQ931().GetCallReference();
  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
      "() First PDU: callReference=" << callReference);

  H323Connection * connection = endpoint.GetConnection(*this, callReference, true);
  if (SBAD_READPTR(connection))
  {
    STRACE2_WARN(g_pH323Module->m_H225TraceId, this, 0,
        "() Could not get H323Connection.");
    return false;
  }

  // This thread is attached to the transport, which is in turn attached to the
  // connection so everything from gets cleaned up by the H323 cleaner thread
  // from now on. So thread must not auto delete and the "transport" variable
  // is not deleted either
  if (!SBAD_READPTR(thread))
    thread->SetNoAutoDelete();

  connection->AttachSignalChannel(this);

    if (connection->HandleSignalPDU(pdu))
    {
        if( connection->HadAnsweredCall()==false )
        {
            // if first pdu isn't an Setup, it is supposed to be a trailing facility with MSDAck
            STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
                "() first pdu isn't an setup --> supposed to be a trailing facility with MSDAck.");
            connection->ClearCall(H323Connection::CallAbortTermination );
        }
        else
        {
            connection->HandleSignallingChannel();
        }
    }
    else
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
            "() Signal channel stopped on first PDU.");
        connection->ClearCall(H323Connection::AbnormalTermination);
    }
    return true;
}


bool H323Transport::ConnectControlChannel(H323Connection & connection)
{
  if (!Connect())
    return false;

  new H245TransportThread(endpoint, connection, *this);
  return true;
}


void H323Transport::AttachThread(SThread * thrd)
{
  SAssert(thread == NULL, SLogicError);
  thread = thrd;
}


void H323Transport::CleanUpOnTermination()
{
  m_bIsClosed = true;
  Close();

  if (thread != NULL) {
    thread->WaitForTermination();
    delete thread;
    thread = NULL;
  }
}


bool H323Transport::IsCompatibleTransport(const H225_TransportAddress & /*pdu*/) const
{
  SAssertAlways(SUnimplementedFunction);
  return false;
}


void H323Transport::SetUpTransportPDU(H225_TransportAddress & /*pdu*/,
                                      bool /*bIsLocalTsap*/) const
{
  SAssertAlways(SUnimplementedFunction);
}


void H323Transport::SetUpTransportPDU(H245_TransportAddress & /*pdu*/,
                                      unsigned /*port*/) const
{
  SAssertAlways(SUnimplementedFunction);
}


void H323Transport::SetUpTransportPDU(H245_TransportAddress & /*pdu*/,
                                      Tools::SIpAddress /*tsap*/) const
{
  SAssertAlways(SUnimplementedFunction);
}


H323Transport * H323Transport::CreateControlChannel(H323Connection & /*connection*/)
{
  SAssertAlways(SUnimplementedFunction);
  return NULL;
}


// Extension for SwTCL
H323Transport * H323Transport::CreateSwTCLControlChannel(H323Connection & /*connection*/)
{
  SAssertAlways(SUnimplementedFunction);
  return NULL;
}


bool H323Transport::AcceptControlChannel(H323Connection & /*connection*/)
{
  SAssertAlways(SUnimplementedFunction);
  return false;
}


bool H323Transport::DiscoverGatekeeper(H323Gatekeeper & /*gk*/,
                                       H323RasPDU & /*pdu*/,
                                       const H323TransportAddress & /*address*/)
{
  SAssertAlways(SUnimplementedFunction);
  return false;
}


bool H323Transport::DiscoverGatekeeper(H323GkLink & /*gk*/,
                                       H323RasPDU & /*pdu*/,
                                       const H323TransportAddress & /*address*/)
{
  SAssertAlways(SUnimplementedFunction);
  return false;
}


/////////////////////////////////////////////////////////////////////////////

H323ListenerTCP::H323ListenerTCP(H323EndPoint & end,
                                 SIPSocket::Address binding,
                                 WORD port)
  : H323Listener(end),
    listener(port),
    localAddress(binding),
	bShutdownListener(false)
{
}


H323ListenerTCP::~H323ListenerTCP()
{
  Close();
}


bool H323ListenerTCP::Open()
{
  return listener.Listen(localAddress);
}


bool H323ListenerTCP::Close()
{
    SDEFFUNC("H323ListenerTCP::Close");

    if (listener.IsOpen())
	{
		bShutdownListener = true;
		
		SIPSocket::Address localAddress;
        listener.GetPreferredLocalAddress( localAddress );
        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0,
            "() sent dummy data to " << localAddress
            << ", port " << listener.GetPort() );

        STCPSocket shutdownHelper(localAddress, listener.GetPort());
		char dummy[] = "The End";
		shutdownHelper.Write(dummy, sizeof(dummy));
		
		WaitForTermination();
		return true;
	}

	return false;
}


static void AppendTransportAddress(H225_ArrayOf_TransportAddress & pdu,
                                   const SIPSocket::Address & addr, WORD port)
{
  H225_TransportAddress_ipAddress pduAddr;
  int i;
  pduAddr.m_ip.SetValue((DWORD)addr);
  pduAddr.m_port = port;

  int lastPos = pdu.GetSize();

  // Check for already have had that IP address.
  for (i = 0; i < lastPos; i++) {
    H225_TransportAddress & taddr = pdu[i];
    if (taddr.GetTag() == H225_TransportAddress::e_ipAddress) {
      H225_TransportAddress_ipAddress & ipAddr = taddr;
      if (ipAddr == pduAddr)
        return;
    }
  }

  // Put new listener into array
  pdu.SetSize(lastPos+1);
  H225_TransportAddress & taddr = pdu[lastPos];

  taddr.SetTag(H225_TransportAddress::e_ipAddress);
  (H225_TransportAddress_ipAddress &)taddr = pduAddr;
}


void H323ListenerTCP::SetUpTransportPDU(H225_ArrayOf_TransportAddress & pdu,
                                        const H225_TransportAddress & first)
{
  WORD port = listener.GetPort();

  if (localAddress != INADDR_ANY) {
    AppendTransportAddress(pdu, localAddress, port);
    return;
  }

  SIPSocket::InterfaceTable interfaces;
  if (!SIPSocket::GetInterfaceTable(interfaces)) {
    SIPSocket::Address ipAddr;
    SIPSocket::GetHostAddress(ipAddr);
    AppendTransportAddress(pdu, ipAddr, port);
    return;
  }

  SIPSocket::Address firstAddress = 0;
  if (first.GetTag() == H225_TransportAddress::e_ipAddress) {
    const H225_TransportAddress_ipAddress & pduAddr = first;
    firstAddress = SIPSocket::Address((DWORD)pduAddr.m_ip);
  }

  SIPSocket::InterfaceTable::size_type i;
  for (i = 0; i < interfaces.size(); i++) {
    SIPSocket::Address ipAddr = interfaces[i]->GetAddress();
    if (ipAddr == firstAddress)
      AppendTransportAddress(pdu, ipAddr, port);
  }

  for (i = 0; i < interfaces.size(); i++) {
    SIPSocket::Address ipAddr = interfaces[i]->GetAddress();
    if (!ipAddr.IsAny() && ipAddr != firstAddress && ipAddr.IsValid() ) 
      AppendTransportAddress(pdu, ipAddr, port);
  }

  for (i = 0; i < interfaces.size(); i++)
      delete interfaces[i];
}


void H323ListenerTCP::SetUpTransportPDU(H225_ArrayOf_TransportAddress   & pdu,
                                        const Tools::SIPSocket::Address & addr)
{
    WORD port = listener.GetPort();
    AppendTransportAddress(pdu, addr, port);
    return;
}


bool H323ListenerTCP::IsListeningOnAddress(
    const SIPSocket::Address & rAddress) const
{
    if (localAddress != INADDR_ANY)
    {
        if (rAddress == localAddress)
            return true;
        else
            return false;
    }

    SIPSocket::InterfaceTable interfaces;
    if (!SIPSocket::GetInterfaceTable(interfaces))
    {
        SIPSocket::Address ipAddr;
        SIPSocket::GetHostAddress(ipAddr);
        if (rAddress == ipAddr)
            return true;
    }

	for (SIPSocket::InterfaceTable::size_type i = 0; i < interfaces.size(); i++)
    {
        SIPSocket::Address ipAddr = interfaces[i]->GetAddress();
        if (ipAddr == rAddress)
        {
            for (i = 0; i < interfaces.size(); i++)
              delete interfaces[i];
            return true;
        }
    }

    for (unsigned int i = 0; i < interfaces.size(); i++)
      delete interfaces[i];
    
    return false;
}


void H323ListenerTCP::Main()
{
    SDEFFUNC("H323ListenerTCP::Main");

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0,
        "() Started thread (H323ListenerTCP)");

  SIPSocket::Address addr;
  listener.GetLocalAddress(addr);
  STRACE2_INFO(g_pH323Module->m_H323TraceId, this, 0,
      "() Awaiting TCP connections on " << addr <<":" << listener.GetPort());

  while (listener.IsOpen()) {
    STCPSocket * signallingChannel = new STCPSocket;
    if (signallingChannel->Accept(listener))
	{
      if ( bShutdownListener )
      {
        delete signallingChannel;
        signallingChannel = 0;
        break;
      }
      H323TransportTCP * transport = new H323TransportTCP(endpoint,true);
      if (transport->Open(signallingChannel))
        new H225TransportThread(endpoint, transport);
      else
        delete transport;
    }
    else {
      if (signallingChannel->GetErrorCode() != SChannel::Interrupted) {
        if (   signallingChannel->GetErrorText().empty()
            || signallingChannel->GetErrorText() == "Socket closed or invalid")
        {
            // Some errors are irrelevant
            STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
                "() Accept error: " << signallingChannel->GetErrorText());
        }
        else
        {
            // Other errors may be important
            STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
                "() Accept error: " << signallingChannel->GetErrorText());
        }
        listener.Close();
      }
      delete signallingChannel;
    }
  }
  STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0,
      "() Finished thread (H323ListenerTCP)");
}


/////////////////////////////////////////////////////////////////////////////

H323TransportIP::H323TransportIP(H323EndPoint & end, WORD remPort)
  : H323Transport(end),
    remoteAddress(0)
{
  SIPSocket::GetHostAddress(localAddress);
  localPort = 0;
  remotePort = remPort;
}


H323TransportAddress H323TransportIP::GetLocalAddress() const
{
  return H323TransportAddress(localAddress, localPort);
}


H323TransportAddress H323TransportIP::GetRemoteAddress() const
{
  return H323TransportAddress(remoteAddress, remotePort);
}


Tools::SString H323TransportIP::GetRemoteAddressString() const
{
    Tools::SString szAddress("ip:");
    szAddress += remoteAddress.AsString();
    szAddress += ":";
    szAddress += SString(SString::Unsigned, (unsigned)remotePort);

    return szAddress;
}


bool H323TransportIP::IsCompatibleTransport(const H225_TransportAddress & pdu) const
{
  return pdu.GetTag() == H225_TransportAddress::e_ipAddress;
}


void H323TransportIP::SetUpTransportPDU(H225_TransportAddress & pdu, bool bIsLocalTsap) const
{
  pdu.SetTag(H225_TransportAddress::e_ipAddress);
  H225_TransportAddress_ipAddress & pduAddr = pdu;

  SIPSocket::Address ipAddr = bIsLocalTsap ? localAddress : remoteAddress;
  pduAddr.m_ip.SetValue((DWORD)ipAddr);
  pduAddr.m_port = bIsLocalTsap ? localPort : remotePort;
}


void H323TransportIP::SetUpTransportPDU(H245_TransportAddress & pdu, unsigned port) const
{
  pdu.SetTag(H245_TransportAddress::e_unicastAddress);

  H245_UnicastAddress & unicast = pdu;
  unicast.SetTag(H245_UnicastAddress::e_iPAddress);

  H245_UnicastAddress_iPAddress & addr = unicast;
  addr.m_network.SetValue((DWORD)localAddress);
  addr.m_tsapIdentifier = port;
}


void H323TransportIP::SetUpTransportPDU(H245_TransportAddress & pdu, Tools::SIpAddress tsap) const
{
  pdu.SetTag(H245_TransportAddress::e_unicastAddress);

  H245_UnicastAddress & unicast = pdu;
  unicast.SetTag(H245_UnicastAddress::e_iPAddress);

  H245_UnicastAddress_iPAddress & addr = unicast;
  addr.m_network.SetValue((DWORD)tsap.m_Address);
  addr.m_tsapIdentifier = tsap.m_Port;
}


/////////////////////////////////////////////////////////////////////////////

H323TransportTCP::H323TransportTCP(H323EndPoint & end, bool bIsIncoming)
  : H323TransportIP(end, /*H323ListenerTls::DefaultSignalTcpPort*/1720),
    h245listener(NULL),
    m_bCancelEarlyH245(false),
    m_connectionState(bIsIncoming?e_HandleFirstSignallingChannelPDU:e_HandleSignallingChannelPDU), 
    m_tpktState(e_ReadingVersion),
    m_readHeaderBytes(0),
    m_bytesLeftToRead(0),
    m_pConnection(NULL),
    m_pRefMutex(new SRefCriticalSection())
{
    SDEFFUNC("H323TransportTCP::H323TransportTCP");
    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "() bIsIncoming=" << bIsIncoming);
}


H323TransportTCP::H323TransportTCP(H323EndPoint & end, WORD port)
  : H323TransportIP(end, port),
    h245listener(NULL),
    m_bCancelEarlyH245(false),
    m_connectionState(e_HandleFirstSignallingChannelPDU), 
    m_tpktState(e_ReadingVersion),
    m_readHeaderBytes(0),
    m_bytesLeftToRead(0),
    m_pConnection(NULL),
    m_pRefMutex(NULL)
{
    SDEFFUNC("H323TransportTCP::H323TransportTCP");
    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "() port"<<port);
}


H323TransportTCP::~H323TransportTCP()
{
    SDEFFUNC("H323TransportTCP::~H323TransportTCP");

    Close();

    delete h245listener;  // Delete and H245 listener that may be present

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "()");
}


bool H323TransportTCP::OnOpen()
{
  SDEFFUNC("H323TransportTCP::OnOpen");

  SIPSocket * socket = (SIPSocket *)GetReadChannel();

  // Get name of the remote computer for information purposes
  if (!socket->GetPeerAddress(remoteAddress, remotePort))
    return false;

  // get local address of incoming socket to ensure that multi-homed machines
  // use a NIC address that is guaranteed to be addressable to destination
  if (!socket->GetLocalAddress(localAddress, localPort))
    return false;
#ifdef SO_HARD_CLOSE
  // make sure that socket will be closed with RESET (hard close)
  const linger ling = { 1, 0 };
  return socket->SetOption(SO_LINGER, &ling, sizeof(ling));
#else

#ifdef __BEOS__
  return true;
#else
  // switch off nagle algorithm as signalling traffic hardly causes congestion issues.
  socket->SetOption(TCP_NODELAY, true, IPPROTO_TCP);

  // switch on TCP keep alive mechanism (Ticket 080702-26598)
  // set TCPKeepAliveTime to 0 via registry means disable mechanism, 
  // otherwise ms for keep alive timeout
  SConfig Config;
  u_long uKeepAliveTime = Config.GetInteger("TCPKeepAliveTime", 60000);
  bool ret = socket->SetKeepAlive((uKeepAliveTime > 0), uKeepAliveTime);

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
      "() Set TCP keep alive to " << uKeepAliveTime << "ms" << (ret ? " succeeded" : " failed"));


  // make sure do not lose outgoing packets on close
  const linger ling = { 1, 10 };
  return socket->SetOption(SO_LINGER, &ling, sizeof(ling));
#endif
#endif
}


bool H323TransportTCP::SetRemoteAddress(const H323TransportAddress & address)
{
  return address.GetIpAndPort(remoteAddress, remotePort, "tcp");
}

bool H323TransportTCP::Close()
{
    SDEFFUNC("H323TransportTCP::Close");

    channelPointerMutex.StartWrite();

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "()");

    SIPSocket::Address peerAddr;
    WORD peerPort = 0;
    SH323TcpConnection * pTcpCon = dynamic_cast<SH323TcpConnection *>(GetReadChannel());

    Errors lastErr = SChannel::NoError;

    if (pTcpCon)
    {
        lastErr = pTcpCon->GetErrorCode(); // store lastErr, because GetPeerAddress will overwrite it!
        pTcpCon->GetPeerAddress(peerAddr, peerPort);

        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "() shutdown STcpConnection " << pTcpCon << " IP " << peerAddr.AsString() << " port " << peerPort);
        
        pTcpCon->RemoveTransport();
        pTcpCon->Shutdown(ShutdownWrite);

        // channel was deleted/destroyed => clear channel member at base 
        // class to avoid multiple deletion
        readChannel = writeChannel = NULL;
    }

    // avoid to close tls socket - they should be reused
    ClearChannel(true, true);

    channelPointerMutex.EndWrite();

    return true;
}



bool H323TransportTCP::AbortH323Connection(bool bIsFromWire)
{
    SDEFFUNC("H323TransportTCP::AbortH323Connection");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "() bIsFromWire=" <<(bIsFromWire?"true":"false"));

    // if connection is not set return false --> there was no connection to abort
    if( m_pConnection==NULL ) 
        return false;

    // remove reference to STcpConnection as it destroyed immediately after this method returns
    if( bIsFromWire )
    {
        // from wire --> this method was already called from destructor of SH323TcpConnection 
        //           --> do not call the destructor again (as formerly done by just calleding ClearChannel(true, true)
        readChannel = NULL;
        writeChannel = NULL;
    }
    else
    {
        ClearChannel(true,true); 
    }

    // trigger deletion of the H323Connection object
    m_pConnection->ClearCall(H323Connection::CallAbortTermination);
    m_pConnection=NULL;

    return true;
}

bool H323TransportTCP::ClosedByWire()
{
    SDEFFUNC("H323TransportTCP::ClosedByWire");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "()");

    if( AbortH323Connection(true/*IsFromWire*/)==false )
    {
        // delete this transport object immediately 
        // because the Tls connection had the only existing reference of this transport object
        delete this;
    }

    return true;
}



bool H323TransportTCP::ProcessRead()
{
    SDEFFUNC("H323TransportTCP::ProcessRead");

    if(!ReadPDU()) 
        return false;

    if( m_tpktState == e_PduComplete )
    {
        H323SignalPDU pdu;
        pdu.Read(*this, m_rawData);

        m_tpktState = e_ReadingVersion;

        if( m_connectionState==e_HandleFirstSignallingChannelPDU )
        {
            unsigned callReference = pdu.GetQ931().GetCallReference();
            STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
              "() First PDU: messageType=" << pdu.GetQ931().GetMessageTypeName() << " callReference=" << callReference);

            if( pdu.GetQ931().GetMessageType() == Q931::SetupMsg)
            {
                // make provisional response actual
                SH323TlsConnection* pTlsCon = dynamic_cast<SH323TlsConnection*>(GetReadChannel());
                // ConfirmTransport() is only needed in the reuse-case with TLS connections
                if(pTlsCon!=NULL) pTlsCon->ConfirmTransport();

                m_pConnection = endpoint.GetConnection(*this, callReference, true);
                if (SBAD_READPTR(m_pConnection))
                {
                    STRACE2_WARN(g_pH323Module->m_H225TraceId, this, 0,
                        "() Could not get H323Connection.");
                    return false;
                }

                m_pConnection->AttachSignalChannel(this);
                
                if (m_pConnection->HandleSignalPDU(pdu))
                {
                    if( m_pConnection->HadAnsweredCall()==false )
                    {
                        // if first pdu isn't an Setup, it is supposed to be a trailing facility with MSDAck
                        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
                            "() call was not answered --> clear call");
                        m_pConnection->ClearCall(H323Connection::CallAbortTermination );
                        m_pConnection=NULL;
                        m_connectionState = e_MaxConnectionState;
                    }
                    else
                    {
                        m_connectionState = e_HandleSignallingChannelPDU;
                    }
                }
                else
                {
                    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
                        "() Signal channel stopped on first PDU.");
                    m_pConnection->ClearCall(H323Connection::AbnormalTermination);
                    m_pConnection=NULL;
                    m_connectionState = e_MaxConnectionState;
                }
            }
            else
            {
                STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
                    "() first pdu isn't an setup --> supposed to be a trailing facility with MSDAck.");
                
                // returning 'false' will force the caller to delete this object
                return false;
            }

        } 
        else if( m_connectionState==e_HandleSignallingChannelPDU )
        {
            if (!m_pConnection->HandleSignalPDU(pdu)) 
            {
                STRACE2_INFO(g_pH323Module->m_H225TraceId, this, (unsigned long)GetReadChannel(),
                    "() Signal channel stopped.");
                m_pConnection->ClearCall(H323Connection::AbnormalTermination);
                m_pConnection=NULL;
                m_connectionState = e_MaxConnectionState;
            }
        }
    }

    return true;
}

bool H323TransportTCP::ReadPDU()
{
    SDEFFUNC("H323TransportTCP::ReadPDU");

    if( m_tpktState==e_ReadingVersion )
    {
        STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, 0, "() tpktState==e_ReadingVersion" );
        // Make sure is a RFC1006 TPKT
        // Get TPKT length
        BYTE c;
        bool ok = Read(&c, 1);

        if (ok)
        {
            if( GetLastReadCount()>0 )
            {
                switch (c) {
                    case 3:  // Only support version 3
                        m_tpktState=e_ReadingHeader;
                      break;

                    default :  // Unknown version number
                      return SetErrorValues(Miscellaneous, 0x41000000);
                }
            }
        }
        else
        {
            DWORD error = GetLastError();
            if (error!=SChannel::Timeout)
            {
                STRACE2_WARN(g_pH323Module->m_H225TraceId, this, 0, "() error reading tpkt version: " << GetErrorText() );
                return false;
            }
        }
    }

    if( m_tpktState==e_ReadingHeader ) 
    {
        STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, 0, "() tpktState==e_ReadingHeader, bytes left to read " << sizeof(m_tpktHeader)-m_readHeaderBytes );
        // Get TPKT length
        bool ok = Read(m_tpktHeader, sizeof(m_tpktHeader)-m_readHeaderBytes);

        if (ok)
        {
            m_readHeaderBytes+=GetLastReadCount();
            if( sizeof(m_tpktHeader)==m_readHeaderBytes )
            {
                m_tpktState=e_ReadingBody;
                m_bytesLeftToRead = ((m_tpktHeader[1] << 8)|m_tpktHeader[2]) - 4;
                STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, 0, "() tpkt header complete. length=" << m_bytesLeftToRead );
                m_rawData.resize(m_bytesLeftToRead);
                // reset auxiliary values used for reading the header
                m_readHeaderBytes=0;
            }
            else
            {
                STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, 0, "() bytes left to read: " << sizeof(m_tpktHeader)-m_readHeaderBytes );
            }
        }
        else
        {
            DWORD error = GetLastError();
            if (error==SChannel::Timeout)
            {
                m_readHeaderBytes+=GetLastReadCount();
            }
            else
            {
                STRACE2_WARN(g_pH323Module->m_H225TraceId, this, 0, "() error reading tpkt header: " << GetErrorText() );
                return false;
            }
        }
    }

    if( m_tpktState==e_ReadingBody ) 
    {
        STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, 0, "() tpktState==e_ReadingBody, bytes left to read: " << m_bytesLeftToRead  );
        bool ok = Read(&(*(m_rawData.begin()+(m_rawData.size()-m_bytesLeftToRead))), m_bytesLeftToRead);

        if (ok) 
        {
            m_bytesLeftToRead-=GetLastReadCount();
            if( m_bytesLeftToRead==0 )
            {
                m_tpktState=e_PduComplete;
                STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, 0, "() tpktState==e_PduComplete" );
            }
            else
            {
                STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, 0, "() bytes left to read: " << m_bytesLeftToRead );
            }
        }
        else
        {
            DWORD error = GetLastError();
            if (error==SChannel::Timeout)
            {
                m_bytesLeftToRead-=GetLastReadCount();
                STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, 0, "() timeout. bytes left to read: " << m_bytesLeftToRead );
            }
            else
            {
                STRACE2_WARN(g_pH323Module->m_H225TraceId, this, 0, "() error reading tpkt body: " << GetErrorText() );
                return false;
            }
        }
    }

    return true;
}



bool H323TransportTCP::ReadPDU(SByteVector & pdu)
{
  // Make sure is a RFC1006 TPKT
  switch (ReadChar()) {
    case -1 :
      return false;

    case 3 :  // Only support version 3
      break;

    default :  // Unknown version number
      return SetErrorValues(Miscellaneous, 0x41000000);
  }

  // Save timeout
  STimeInterval oldTimeout = GetReadTimeout();

  // Should get all of PDU in 5 seconds or something is seriously wrong,
  SetReadTimeout(5000);

  // Get TPKT length
  BYTE header[3];
  bool ok = ReadBlock(header, sizeof(header));
  if (ok) {
    unsigned int packetLength = ((header[1] << 8)|header[2]) - 4;

    if (packetLength > pdu.size())
        pdu.resize(packetLength);

    ok = ReadBlock(&(*(pdu.begin())), packetLength);
  }

  SetReadTimeout(oldTimeout);

  return ok;
}


bool H323TransportTCP::WritePDU(const SByteVector & pdu)
{
  SWaitAndSignal wait(m_Mutex);

  int packetLength = pdu.size() + 4;

  // Send RFC1006 TPKT length
/*
  BYTE tpkt[4];
  tpkt[0] = 3;
  tpkt[1] = 0;
  tpkt[2] = (BYTE)(packetLength >> 8);
  tpkt[3] = (BYTE)packetLength;

  return Write(tpkt, sizeof(tpkt)) && Write(&(*(pdu.begin())), pdu.size());
*/

  // make just one Write() call as this assures that just one TCP frame is sent instead of two (TPKT and PDU)
  SByteVector tpkt;
  tpkt.reserve(packetLength);
  tpkt.resize(4);
  tpkt[0] = 3;
  tpkt[1] = 0;
  tpkt[2] = (BYTE)(packetLength >> 8);
  tpkt[3] = (BYTE)packetLength;

  tpkt.insert( tpkt.end(), pdu.begin(), pdu.end() );
  return Write(&(*(tpkt.begin())), tpkt.size());
}



bool H323TransportTCP::Connect()
{
    SDEFFUNC("H323TransportTCP::Connect");

  STcpConnection* socket = endpoint.m_tcpConFactory.CreateNew(endpoint, g_pH323Module->m_H323TraceId);

  SIpAddress destAddr = remoteAddress;
  destAddr.m_Port = remotePort;


  channelPointerMutex.StartWrite();
  readChannel = writeChannel = socket;
  channelPointerMutex.EndWrite();

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
      "() Starting connection to " << destAddr);

  channelPointerMutex.StartRead();
  if (!socket->Connect(destAddr) )
  {
    int errnum = socket->GetErrorNumber();
    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
        "() Could not connect to " << destAddr
        << ": " << socket->GetErrorText() << '(' << errnum << ')');
    channelPointerMutex.EndRead();
    channelPointerMutex.StartWrite();
    readChannel = writeChannel = NULL;
    channelPointerMutex.EndWrite();
    delete socket;
    return SetErrorValues(socket->GetErrorCode(), errnum);
  }

  socket->SetOption(TCP_NODELAY, true, IPPROTO_TCP);

  SWaitAndSignal lock(*m_pRefMutex);

  // put socket to the listener in order to read and process messages from wire
  SH323TcpConnection* pTcpCon = dynamic_cast<SH323TcpConnection*>(socket);
  // dynamic_cast will never fail because we generated it

  pTcpCon->SetTransport(this);  

  STRACE2_INFO3( g_pH323Module->m_H323TraceId, this, 0, "() new TlsConnection created: " << pTcpCon );

  endpoint.PutSignallingChannelToListener( socket );
  
  //  readChannel->SetReadTimeout(SMaxTimeInterval); // Set timeout default
  channelPointerMutex.EndRead();
  channelPointerMutex.StartWrite();
  readChannel = writeChannel = NULL;
  channelPointerMutex.EndWrite();

  if (!Open(socket))
    return false;

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
      "() Started connection to " << remoteAddress << ":" << remotePort
      << " (if=" << localAddress << ":" << localPort << ")");

  return true;
}


H323Transport * H323TransportTCP::CreateControlChannel(H323Connection & connection)
{
    SDEFFUNC("H323TransportTCP::CreateCtrlChan");

    // Check for configured interface
    SIPSocket::Address address = connection.GetLocalInterfaceAddress();
    if (!address.IsValid() )
    {
        address = INADDR_ANY;
    }

  // construct H245 socket
  STCPSocket * socket = new STCPSocket;
  if (!socket->Listen(address)) {
    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
        "() TCP Listen for H245 failed: " << socket->GetErrorText());
    delete socket;
    return NULL;
  }

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
      "() TCP Listen for H245 on " << localAddress << ":" << socket->GetPort());

  // Make H323 Connect PDU
  H323SignalPDU connectPDU;
  connectPDU.BuildConnect(connection,
                          connection.GetCallReference(),
                          connection.GetCallIdentifier(),
                          connection.GetConferenceIdentifier(),
                          localAddress, socket->GetPort());


  // Send H323 Connect PDU
  if (!connection.WriteSignalPDU(connectPDU)) {
    delete socket;
    return NULL;
  }

  H323TransportTCP * transport = new H323TransportTCP(endpoint,false);
  transport->h245listener = socket;
  new H245TransportThread(endpoint, connection, *transport);
  return transport;
}


// Extension for SwTCL
H323Transport * H323TransportTCP::CreateSwTCLControlChannel(H323Connection & connection)
{
    SDEFFUNC("H323TranspTCP::CreateSwTCLCtrlCh");

    // Check for configured interface
    SIPSocket::Address address = connection.GetLocalInterfaceAddress();
    if (!address.IsValid() )
    {
        address = INADDR_ANY;
    }

    // construct H245 socket
    STCPSocket * socket = new STCPSocket;
    if (!socket->Listen(address))
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, 0,
            "() TCP Listen for H245 failed: " << socket->GetErrorText());
        delete socket;
        return NULL;
    }

    connection.SetH245LocalAddress(localAddress);
    connection.SetH245LocalPort(socket->GetPort());

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, 0,
        "() TCP Listen for H245 on " << connection.GetH245LocalAddress()
        << ":" << connection.GetH245LocalPort());

    H323TransportTCP * transport = new H323TransportTCP(endpoint,false);
    transport->h245listener = socket;
    new H245TransportThread(endpoint, connection, *transport);
    return transport;
}


bool H323TransportTCP::AcceptControlChannel(H323Connection & connection)
{
    SDEFFUNC("H323TransportTCP::AcceptCtrlChan");

    if (IsOpen())
      return true;

    if (h245listener == NULL)
    {
        SAssertAlways(SLogicError);
        return false;
    }

    int iSeconds = 1;
    int iNumberOfLoops = (int)( (MAX_H245_ACCEPT_WAITING) / iSeconds );

    STCPSocket * h245Socket = new STCPSocket;

    for (int i = 1; i <= iNumberOfLoops; i++)
    {
        // Finish early started H.245 transport thread if 'Early H.245' is canceled
        if (IsCancelEarlyH245())
        {
            delete h245Socket;
            return false;
        }

        if (m_bIsClosed == true)
            break;

        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, 0,
            "() TCP Accept (" << i << ") for H.245 (port: " << h245listener->GetPort()
            << ") waiting " << iSeconds << " seconds...");

        // Wait iSeconds seconds for remote H245 connect
        h245listener->SetReadTimeout(iSeconds*1000);
        if (h245Socket->Accept(*h245listener))
        {
            STRACE2_INFO(g_pH323Module->m_H245TraceId, this, 0,
                "() TCP Accept for H.245 ok, port: " << h245Socket->GetPort());
            return Open(h245Socket);
        }
    }

    if (m_bIsClosed == true)
    {
        STRACE2_INFO(g_pH323Module->m_H245TraceId, this, 0,
            "() Accept for H.245 finished.");
    }
    else
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, 0,
            "() Accept for H.245 failed: " << h245Socket->GetErrorText());
    }
    delete h245Socket;

    if (   !connection.IsConnected()
        || (   connection.FindChannel(RTP_Session::DefaultAudioSessionID, true) == NULL
            || connection.FindChannel(RTP_Session::DefaultAudioSessionID, false) == NULL ) )
    {
        STRACE2_WARN(g_pH323Module->m_H245TraceId, this, 0,
            "() Not connected or no open audio channel found -> clear call.");
        connection.ClearCall(H323Connection::AbnormalTermination);
        return false;
    }
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, 0,
        "() Open audio channel found (e.g. fast start) -> NO clear call.");
    return false;
}



/////////////////////////////////////////////////////////////////////////////

H323TransportUDP::H323TransportUDP(H323EndPoint & end, 
                                   SIPSocket::Address binding,
                                   WORD LocalPort,
                                   bool Multicast)
  : H323TransportIP(end, DefaultRasPort)
{
    SDEFFUNC("H323TransportUDP::H323TransprtUDP");

  SUDPSocket * udp = new SUDPSocket(LocalPort);

  udp->Listen(binding, 5 /* queue size */, LocalPort);

  if (Multicast)
  {
      struct ip_mreq mreq;
      mreq.imr_multiaddr.s_addr = inet_addr(RAS_MULTICAST_ADDRESS);
      mreq.imr_interface.s_addr = binding;

      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // Microsoft has a problem with its ws2_32.lib, so use wsock32.lib
      // instead for multicast support! Set this in the linking options!
      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      int Err = setsockopt(udp->GetHandle(),
                     IPPROTO_IP,
                     IP_ADD_MEMBERSHIP,
                     (char *)&mreq,
                     sizeof(mreq));
      if (Err < 0)
      {
          STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
              "() Cannot open multicast address.");
      }
  }
  
  Open(udp);

//  SAssert(LocalPort == udp->GetPort(), SLogicError);
//  line commented out RB 28.6.2000 - would assert is LocalPort=0, because a free port if determined by
//  SDUPSocket obj itself and set afterwards to localport

  localPort = udp->GetPort();
}


H323TransportUDP::~H323TransportUDP()
{
  Close();
}


void H323TransportUDP::GetLastReceiveAddress(SIPSocket::Address & address, WORD & port)
{
    SUDPSocket * socket = (SUDPSocket *)GetReadChannel();
    socket->GetLastReceiveAddress(address, port);
}


bool H323TransportUDP::SetRemoteAddress(const H323TransportAddress & address)
{
  return address.GetIpAndPort(remoteAddress, remotePort, "udp");
}


bool H323TransportUDP::Connect()
{
  if (remoteAddress == 0 || remotePort == 0)
    return false;

  SUDPSocket * socket = (SUDPSocket *)GetReadChannel();
  socket->SetSendAddress(remoteAddress, remotePort);

  return true;
}


bool H323TransportUDP::ReadPDU(SByteVector & pdu)
{
  // replace original code: GetPointer(10000)
  if (10000 > pdu.size())
      pdu.resize(10000);

  if (!Read(&(*(pdu.begin())), 10000)) {
    pdu.resize(0);
    return false;
  }

  pdu.resize(GetLastReadCount());
  return true;
}


bool H323TransportUDP::WritePDU(const SByteVector & pdu)
{
  return Write(&(*(pdu.begin())), pdu.size());
}


bool H323TransportUDP::DiscoverGatekeeper(H323Gatekeeper & gk,
                                          H323RasPDU & request,
                                          const H323TransportAddress & address)
{
    SDEFFUNC("H323TransportUDP::DiscoverGatek");
    
    SIPSocket::InterfaceTable interfaces;
    
    SSocketList sockets;
    SSocketList::iterator socketsIter;
    
    SSocket::SelectList selection;
    SSocket::SelectList::iterator pos;
    
    SIPSocket::InterfaceTable::size_type i;
    bool               bHavingAddress = false;  
    SIPSocket::Address preferredAddress;
    
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Gatekeeper discovery started");
    
    Close();
    
    remoteAddress = 0;
    remotePort = 0;
    
    SIPSocket::Address destAddr = INADDR_BROADCAST;
    WORD destPort = DefaultRasPort;
    if (!address.empty()) {
        if (!address.GetIpAndPort(destAddr, destPort, "udp")) {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Error decoding address");
            goto Error;
        }
    }
    
    // delete original read channel to avoid a memory leak
    // a new read channel will be allocated below to receiving the GCF message from the gatekeeper
    
    delete GetBaseReadChannel();
    if(writeChannel == readChannel)
    {
        readChannel  = 0;
        writeChannel = 0;
    }
    else
    {
        readChannel  = 0;
    }
    
    // Check for configured interface
    if (   SIPSocket::GetPreferredLocalAddress(preferredAddress) == true
        && preferredAddress.IsValid() )
    {
        bHavingAddress = true;
    }
    
    if (!SIPSocket::GetInterfaceTable(interfaces))
        interfaces.push_back(new SIPSocket::InterfaceEntry("localhost", localAddress, ""));
    
    for (i = 0; i < interfaces.size(); i++) {
        localAddress = interfaces[i]->GetAddress();
        if (localAddress.IsAny() || !localAddress.IsValid() )
            continue;
        
        // Check for already have had that IP address.
        SIPSocket::InterfaceTable::size_type j;
        for (j = 0; j < i; j++) {
            if (localAddress == interfaces[j]->GetAddress())
                break;
        }
        if (j < i)
            continue;
        
        if (   bHavingAddress == true
            && localAddress != preferredAddress)
            continue;
        
        SUDPSocket * socket = new SUDPSocket(localPort);
        sockets.push_back(socket);
        
        if (!socket->Listen(localAddress))
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Error on discover request listen: " << socket->GetErrorText());
            goto Error;
        }
        
        localPort = socket->GetPort();
        
#ifndef __BEOS__
        if (destAddr == INADDR_BROADCAST)
        {
            if (!socket->SetOption(SO_BROADCAST, 1))
            {
                STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                    "() Error allowing broadcast: " << socket->GetErrorText());
                goto Error;
            }
        }
#else
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Broadcast option under BeOS is not implemented yet");
#endif
        
        // Adjust the PD to reflect the interface we are writing to.
        H225_GatekeeperRequest & grq = request;
        SetUpTransportPDU(grq.m_rasAddress, true);
        socket->SetSendAddress(destAddr, destPort);
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS GatekeeperRequest to " << destAddr << ":" <<  destPort);
        
        writeChannel = socket;
        if (request.Write(*this) == H323RasPDU::Success)
            selection.push_back(socket);
        else
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Error writing discovery PDU: " << socket->GetErrorText());
        }
        
        writeChannel = NULL;
        
#ifndef __BEOS__
        if (destAddr == INADDR_BROADCAST)
            socket->SetOption(SO_BROADCAST, 0);
#endif
    }
    
    if (SSocket::Select(selection, endpoint.GetGatekeeperRequestTimeout()) != NoError)
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Error on discover request select");
        goto Error;
    }
    
    SetReadTimeout(0);
    
    for (pos = selection.begin(); pos != selection.end(); ++pos) {
        readChannel = (*pos);
        H323RasPDU response;
        if (response.Read(*this) != H323RasPDU::Success)
        {
            STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
                "() Error on discover request read: " << readChannel->GetErrorText());
            readChannel = NULL;
            goto Error;
        }
        
        do {
            if (gk.HandleRasPDU(response)) {
                SUDPSocket * socket = (SUDPSocket *)readChannel;
                readChannel = NULL;
                
                // Open() is inherited from SIndirectChannel and will delete the socket object itself after closing the channel!
                Open(socket);
                
                // Commented out: Not using source port as dest port
                //socket->GetLastReceiveAddress(remoteAddress, remotePort);
                //socket->SetSendAddress(remoteAddress, remotePort);
                socket->GetLocalAddress(localAddress, localPort);
                for (i = 0; i < interfaces.size(); i++)
                    delete interfaces[i];
                return true;
            }
        } while (response.Read(*this) == H323RasPDU::Success);
    }
    
    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
        "() Discover request failed");

    SIPSocket::GetHostAddress(localAddress);
    readChannel = NULL;
    
Error:
    
    // clean up sockets list to avoid memory leak
    for ( socketsIter = sockets.begin(); socketsIter != sockets.end(); ++socketsIter )
        delete ( *socketsIter );
    
    for (i = 0; i < interfaces.size(); i++)
        delete interfaces[i];
    
    return false;
}


bool H323TransportUDP::DiscoverGatekeeper(H323GkLink & gk,
                                          H323RasPDU & request,
                                          const H323TransportAddress & address)
{
    SDEFFUNC("H323TransportUDP::DiscoverGk-Lk");

  SIPSocket::InterfaceTable interfaces;

  SSocketList sockets;
  SSocketList::iterator socketsIter;

  SSocket::SelectList selection;
  SSocket::SelectList::iterator pos;

  size_t i;
  bool               bHavingAddress = false;  
  SIPSocket::Address preferredAddress = gk.GetLocalInterfaceAddress();

  STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
      "(" << address << ") preferred address: " << preferredAddress << ". Gatekeeper discovery started");

  Close();

  remoteAddress = 0;
  remotePort = 0;

  SIPSocket::Address destAddr = INADDR_BROADCAST;
  WORD destPort = DefaultRasPort;
  if (!address.empty()) {
    if (!address.GetIpAndPort(destAddr, destPort, "udp")) {
      STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
          "() Error decoding address");
      goto Error_gklink;
    }
  }

  // delete original read channel to avoid a memory leak
  // a new read channel will be allocated below to receiving the GCF message from the gatekeeper
  
  delete GetBaseReadChannel();
  if(writeChannel == readChannel)
  {
      readChannel  = 0;
      writeChannel = 0;
  }
  else
  {
      readChannel  = 0;
  }
  
    // Check for configured interface
    if (preferredAddress.IsValid())
    {
        bHavingAddress = true;
    }

  if (!SIPSocket::GetInterfaceTable(interfaces))
    interfaces.push_back(new SIPSocket::InterfaceEntry("localhost", localAddress, ""));

  for (i = 0; i < interfaces.size(); i++) {
    localAddress = interfaces[i]->GetAddress();
    if (localAddress.IsAny() || !localAddress.IsValid())
      continue;

    // Check for already have had that IP address.
    size_t j;
    for (j = 0; j < i; j++) {
      if (localAddress == interfaces[j]->GetAddress())
        break;
    }
    if (j < i)
      continue;

    if (   bHavingAddress == true
        && localAddress != preferredAddress)
      continue;

    SUDPSocket * socket = new SUDPSocket(localPort);
    sockets.push_back(socket);

    if (!socket->Listen(localAddress))
    {
      STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
          "() Error on discover request listen: " << socket->GetErrorText());
      goto Error_gklink;
    }

    localPort = socket->GetPort();

#ifndef __BEOS__
    if (destAddr == INADDR_BROADCAST)
    {
      if (!socket->SetOption(SO_BROADCAST, 1))
      {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Error allowing broadcast: " << socket->GetErrorText());
        goto Error_gklink;
      }
    }
#else
    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
        "() Broadcast option under BeOS is not implemented yet");
#endif

    // Adjust the PD to reflect the interface we are writing to.
    H225_GatekeeperRequest & grq = request;
    SetUpTransportPDU(grq.m_rasAddress, true);
    socket->SetSendAddress(destAddr, destPort);
    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS GatekeeperRequest to " << destAddr << ":" <<  destPort);

    writeChannel = socket;
    if (request.Write(*this) == H323RasPDU::Success)
      selection.push_back(socket);
    else
    {
        STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
            "() Error writing discovery PDU: " << socket->GetErrorText());
    }

    writeChannel = NULL;

#ifndef __BEOS__
    if (destAddr == INADDR_BROADCAST)
      socket->SetOption(SO_BROADCAST, 0);
#endif
  }

  if (SSocket::Select(selection, endpoint.GetGatekeeperRequestTimeout()) != NoError) {
    STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
        "() Error on discover request select");
    goto Error_gklink;
  }

  SetReadTimeout(0);

  for (pos = selection.begin(); pos != selection.end(); ++pos) {
    readChannel = (*pos);
    H323RasPDU response;
    if (response.Read(*this) != H323RasPDU::Success)
    {
      STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
          "() Error on discover request read: " << readChannel->GetErrorText());
      readChannel = NULL;
      goto Error_gklink;
    }

    do {
      if (gk.HandleRasPDU(response)) {
        SUDPSocket * socket = (SUDPSocket *)readChannel;
        readChannel = NULL;

        // Open() is inherited from SIndirectChannel and will delete the socket object itself after closing the channel!
        Open(socket);

        // Commented out: Not using source port as dest port
        //socket->GetLastReceiveAddress(remoteAddress, remotePort);
        //socket->SetSendAddress(remoteAddress, remotePort);
        socket->GetLocalAddress(localAddress, localPort);
        for (i = 0; i < interfaces.size(); i++)
          delete interfaces[i];
        return true;
      }
    } while (response.Read(*this) == H323RasPDU::Success);
  }

  STRACE2_ERROR(g_pH323Module->m_RasTraceId, this, 0,
      "() Discover request failed");
  SIPSocket::GetHostAddress(localAddress);
  readChannel = NULL;

Error_gklink:

  // clean up sockets list to avoid memory leak
  for ( socketsIter = sockets.begin(); socketsIter != sockets.end(); ++socketsIter )
      delete ( *socketsIter );

  for (i = 0; i < interfaces.size(); i++)
      delete interfaces[i];
  
  return false;
}


/////////////////////////////////////////////////////////////////////////////

/*
 * compare function for the connection object container
 * the order doesn't matter for our purposes, but the uniqueness of the used key, 
 * so just take the pointers
 */
bool lessthanSTcpConnection::operator()(const STcpConnection* s1, const STcpConnection* s2) const
{
    return (s1 < s2);
}


H323ListenerTls::H323ListenerTls(H323EndPoint & end,
                                 SIPSocket::Address binding,
                                 WORD tcpPort, WORD tlsPort )
  : H323Listener(end),
    listenerTls(tlsPort),
    listenerTcp(tcpPort),
    localAddress(binding),
	bShutdownListener(false),
    m_bIsTlsServer(tlsPort!=0)
{
}


H323ListenerTls::~H323ListenerTls()
{
  Close();
}

bool H323ListenerTls::AddConnectedConnection(Tools::SChannel* pSocket)
{
    SDEFFUNC("H323ListenerTls::AddConnectedConnection");

    Tools::STcpConnection* pCon = dynamic_cast<Tools::STcpConnection*>(pSocket);

    if( SBAD_READPTR(pCon) )
    {
        STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, 0,
          "() invalid connection pointer pConnection==" << pCon);
        return false;
    }

    // first of all set trace ID string for identifying the SCAMEnpoint objects in the trace
    SIPSocket::Address  addr;
    unsigned short      port;
    pCon->GetPeerAddress(addr, port);

    unsigned long ulConnectionID = STrace::ConvertIP2ConnectionID(addr, m_bTraceFullIP);
    pCon->SetTraceAndConnId(g_pH323Module->m_H323TraceId, ulConnectionID);

    STRACE2_INFO( g_pH323Module->m_H323TraceId, this, ulConnectionID, "() IP: " << addr.AsString() << " = 0x" << std::hex << ntohl((DWORD)addr) << " TraceFullIP: " << m_bTraceFullIP);

    // hk: disable nagles algorithm in TCP socket in order to push data out as soon as write it on the socket;
    //     note: most documentation about TCP socket warn that it is in most cases not recommended to disable
    //           the nagle algorithm, but for signalling paths the flow control of TCP isn't relevant. So 
    //           nagles algorith.
    pCon->SetOption( TCP_NODELAY, true, IPPROTO_TCP );

    pCon->SetNonBlocking(true);
    
    STRACE2_INFO( g_pH323Module->m_H323TraceId, this, ulConnectionID, "() Add new socket and start transport: " << pCon << " Peer Address/Port: " << addr << "/" << port );

    SWaitAndSignal lock(m_TcpConnectionListMutex);

    if( m_TcpConnectionList.insert(pCon).second==true )
    {
        SignalSelectInterrupter();
        return true;
    }

	return false;
}


bool H323ListenerTls::Open()
{
  SDEFFUNC("H323ListenerTls::Open");

  Tools::SIPSocket::Address addr = Tools::SIPSocket::Address::e_Loopback;
  bool selectInterrupterValid = selectInterrupter.Listen(addr);
  selectInterrupter.SetSendAddress(addr, selectInterrupter.GetPort());

  STRACE2_INFO( g_pH323Module->m_H323TraceId, this, 0, 
      "selectInterrupter: " << addr << ":" << selectInterrupter.GetPort());

  return listenerTls.Listen(localAddress, 50) && listenerTcp.Listen(localAddress, 50) && selectInterrupterValid;
}


bool H323ListenerTls::Close()
{
    SDEFFUNC("H323ListenerTls::Close");

    if (listenerTls.IsOpen()||listenerTcp.IsOpen())
	{
		bShutdownListener = true;

        SignalSelectInterrupter();

        WaitForTermination();
		return true;
	}

	return false;
}


void H323ListenerTls::SetUpTransportPDU(H225_ArrayOf_TransportAddress & pdu,
                                        const H225_TransportAddress & first)
{
  WORD port = listenerTls.GetPort();

  if (localAddress != INADDR_ANY) {
    AppendTransportAddress(pdu, localAddress, port);
    return;
  }

  SIPSocket::InterfaceTable interfaces;
  if (!SIPSocket::GetInterfaceTable(interfaces)) {
    SIPSocket::Address ipAddr;
    SIPSocket::GetHostAddress(ipAddr);
    AppendTransportAddress(pdu, ipAddr, port);
    return;
  }

  SIPSocket::Address firstAddress = 0;
  if (first.GetTag() == H225_TransportAddress::e_ipAddress) {
    const H225_TransportAddress_ipAddress & pduAddr = first;
    firstAddress = SIPSocket::Address((DWORD)pduAddr.m_ip);
  }

  SIPSocket::InterfaceTable::size_type i;
  for (i = 0; i < interfaces.size(); i++) {
    SIPSocket::Address ipAddr = interfaces[i]->GetAddress();
    if (ipAddr == firstAddress)
      AppendTransportAddress(pdu, ipAddr, port);
  }

  for (i = 0; i < interfaces.size(); i++) {
    SIPSocket::Address ipAddr = interfaces[i]->GetAddress();
    if (!ipAddr.IsAny() && ipAddr != firstAddress && ipAddr.IsValid() ) 
      AppendTransportAddress(pdu, ipAddr, port);
  }

  for (i = 0; i < interfaces.size(); i++)
      delete interfaces[i];
}


void H323ListenerTls::SetUpTransportPDU(H225_ArrayOf_TransportAddress   & pdu,
                                        const Tools::SIPSocket::Address & addr)
{
    WORD port = listenerTls.GetPort();
    AppendTransportAddress(pdu, addr, port);
    return;
}


bool H323ListenerTls::IsListeningOnAddress(
    const SIPSocket::Address & rAddress) const
{
    if (localAddress != INADDR_ANY)
    {
        if (rAddress == localAddress)
            return true;
        else
            return false;
    }

    SIPSocket::InterfaceTable interfaces;
    if (!SIPSocket::GetInterfaceTable(interfaces))
    {
        SIPSocket::Address ipAddr;
        SIPSocket::GetHostAddress(ipAddr);
        if (rAddress == ipAddr)
            return true;
    }

	for (SIPSocket::InterfaceTable::size_type i = 0; i < interfaces.size(); i++)
    {
        SIPSocket::Address ipAddr = interfaces[i]->GetAddress();
        if (ipAddr == rAddress)
        {
            for (i = 0; i < interfaces.size(); i++)
              delete interfaces[i];
            return true;
        }
    }

    for (unsigned int i = 0; i < interfaces.size(); i++)
      delete interfaces[i];
    
    return false;
}


void H323ListenerTls::Main()
{
    SDEFFUNC("H323ListenerTls::Main");

    int                       iSelectError = 0;
    STime                     EnterSelectTime;
    SSocket::SelectList       selectList;

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, 0, "() Started thread (H323ListenerTls)");

    SIPSocket::Address addrTcp;
    listenerTcp.GetLocalAddress(addrTcp);
    SIPSocket::Address addrTls;
    listenerTls.GetLocalAddress(addrTls);
    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, 0,
        "() Awaiting TCP/Tls connections on " << addrTcp <<":" << listenerTcp.GetPort() << "(TCP) and on " << addrTls <<":" << (m_bIsTlsServer?listenerTls.GetPort():0) << "(TLS)" );

    // hk: don't block in accept statement, the select statement should do the blocking
    listenerTls.SetReadTimeout(0);
    listenerTcp.SetReadTimeout(0);
    selectInterrupter.SetReadTimeout(0);

    while( ( (m_bIsTlsServer?listenerTls.IsOpen():true) && listenerTcp.IsOpen() ) && !bShutdownListener ) 
    {
        /* The select list contains all sockets of current connections +
         * one virgin socket listening for a new connection */
        m_TcpConnectionListMutex.Wait();

        selectList.clear();
        for( ConnectionSet::iterator SckPos = m_TcpConnectionList.begin(); SckPos != m_TcpConnectionList.end(); ++SckPos) 
        {
            selectList += *(*SckPos);
        }

        if( m_bIsTlsServer ) selectList += listenerTls;
        selectList += listenerTcp;
        selectList += selectInterrupter;

        m_TcpConnectionListMutex.Signal();

        /* Select sockets with available data */
        if( selectList.size() > 1) // 1 element => only selectInterrupter present
        {
            SChannel::Errors Error =  SSocket::Select( selectList, 5000 /* 5 s */); 

            if( Error != SChannel::NoError ) 
            {
                STRACE2_WARN( g_pH323Module->m_H323TraceId, this, 0, "() SelectFail: " << SChannel::GetErrorText(Error) );
                if( iSelectError >= 1000 )
                {
                    STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, 0, "() SelectFail 1000st time in sequel -> stop listenerTls");
                    // if this is the second time the select fails break the listenerTls
                    break;
                }
                else
                {
                    iSelectError++;
                }
            }
            else
            {
                iSelectError = 0;
            }
        }
        
        // Read from sockets with new data
        for(SSocket::SelectList::iterator pos = selectList.begin(); pos != selectList.end(); ++pos) 
        {
            STCPSocket* pSocket = dynamic_cast<STCPSocket*>(*pos);

            // If the listening socket is among the selected sockets, then make accept()
            if( pSocket == &listenerTls )
            {
                // create new Socket, because new client=SwyxPhone tries to connect secure H.323 socket
                STcpConnection * signallingChannel = endpoint.m_tlsConFactory.CreateNew(endpoint, g_pH323Module->m_H323TraceId);
                if( signallingChannel == NULL ) 
                {
                    STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, 0, "() Could not create new STlsConnection");
                    continue;
                }

                STRACE2_INFO( g_pH323Module->m_H323TraceId, this, 0, "() accept tls" );
                AcceptTcpConnection(signallingChannel, listenerTls);
                STRACE2_INFO( g_pH323Module->m_H323TraceId, this, 0, "() ConnectionList size (not active ones): " << m_TcpConnectionList.size() );
            }
            else if( pSocket == &listenerTcp )
            {
                // create new Socket, because new client=SwyxPhone tries to connect secure H.323 socket
                STcpConnection * signallingChannel = endpoint.m_tcpConFactory.CreateNew(endpoint, g_pH323Module->m_H323TraceId);
                if( signallingChannel == NULL ) 
                {
                    STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, 0, "() Could not create new STcpConnection");
                    continue;
                }

                STRACE2_INFO( g_pH323Module->m_H323TraceId, this, 0, "() accept tcp" );
                AcceptTcpConnection(signallingChannel, listenerTcp);
                STRACE2_INFO( g_pH323Module->m_H323TraceId, this, 0, "() ConnectionList size (not active ones): " << m_TcpConnectionList.size() );
            }
            else if( pSocket != 0 ) // check for 0 necessary, because it means that the dynmaic_cast to TCPSocket succeeded
            {
                STcpConnection* signallingChannel = dynamic_cast<STcpConnection*>(pSocket);
                
                unsigned long ulAvailableData = 0;
                if( !signallingChannel->DetermineAvailableData(ulAvailableData) || ulAvailableData==0 ) 
                {
                    SIPSocket::Address addr;
                    unsigned short port;
                    signallingChannel->GetPeerAddress(addr, port);

                    // if socket is signaled and zero bytes to be read means, that remote side is has sent FIN to shutdown the TCP connection
                    STRACE2_INFO( g_pH323Module->m_H323TraceId, this, signallingChannel->GetConnectionID(), 
                        "() connection to " << addr.AsString() << ":" << port << " was closed from wire" );
                    signallingChannel->Shutdown(SChannel::ShutdownWrite);
                    STRACE2_INFO( g_pH323Module->m_H323TraceId, this, 0, "() graceful closing socket" );

                    // tell H323 endpoint that TLS connection has been closed
                    if( dynamic_cast<STlsConnection*>(signallingChannel)!=NULL )
                        endpoint.OnSignallingChannelClosed(addr,port);

                    SWaitAndSignal lock(m_TcpConnectionListMutex);

                    STRACE2_INFO( g_pH323Module->m_H323TraceId, this, signallingChannel->GetConnectionID(), "() remove socket from list: " << signallingChannel  );
                    m_TcpConnectionList.erase(signallingChannel);
                    delete signallingChannel;
                }
                else if( !signallingChannel->ProcessRead() )
                {
                    SIPSocket::Address addr;
                    unsigned short port;
                    signallingChannel->GetPeerAddress(addr, port);

                    // tell H323 endpoint that TLS connection has closed
                    if( dynamic_cast<STlsConnection*>(signallingChannel)!=NULL )
                        endpoint.OnSignallingChannelClosed(addr,port);

                    SWaitAndSignal lock(m_TcpConnectionListMutex);

                    STRACE2_INFO( g_pH323Module->m_H323TraceId, this, signallingChannel->GetConnectionID(), 
                        "() remove socket from list: " << signallingChannel << " peer: " << addr.AsString() << ":" << port);

                    m_TcpConnectionList.erase(signallingChannel);
                    delete signallingChannel;
                }
            }
            else
            {
                // read dummy data from selectInterrupter socket
                SUDPSocket* pSocket = dynamic_cast<SUDPSocket*>(*pos);
                if (pSocket == &selectInterrupter)
                {
                    char buf[1024];
                    pSocket->Read(buf, sizeof(buf));
                }
            }
        }

        // call cleanUpConnections from this thread, and not via former, additional 
        // connectionscleaner-thread to avoid race conditions
        endpoint.CleanUpConnections();
    }/*while*/

    // Delete all sockets 

    // delete connection objects from TcpConnectionList
    DeleteAllTcpConnections();

    // close listening socket before deletion, hk: should be closed from other thread on closing the listenerTls
    if(listenerTls.IsOpen() && listenerTls.Close() == false) 
    { 
        STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, 0, "() Error closing socket listener");
    }
     
    // close listening socket before deletion, hk: should be closed from other thread on closing the listener
    if(listenerTcp.IsOpen() && listenerTcp.Close() == false) 
    { 
        STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, 0, "() Error closing socket listenerTcp");
    }
     
    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, 0, "() Finished thread (H323ListenerTls)");
}

void H323ListenerTls::AcceptTcpConnection(Tools::STcpConnection* signallingChannel, Tools::STCPSocket& listenerSocket )
{
    SDEFFUNC("H323ListenerTls::AcceptTcpConnection");

    if( signallingChannel->Accept(listenerSocket) == false) 
    {
        STRACE2_INFO( g_pH323Module->m_H323TraceId, this, 0, "() AcceptFail: " << signallingChannel->GetErrorText());
        
        // delete STcpConnection again
        delete signallingChannel;
        signallingChannel = 0;
    } 
    else 
    {
        bool bIsTls = (&listenerTls==&listenerSocket);

        STRACE2_INFO3( g_pH323Module->m_H323TraceId, this, 0, "() new " << (bIsTls?"Tls":"Tcp") << "Connection accepted" );

        // first of all set trace ID string for identifying the SCAMEnpoint objects in the trace
        SIPSocket::Address  addr;
        unsigned short      port;
        signallingChannel->GetPeerAddress(addr, port);

        unsigned long ulConnectionID = STrace::ConvertIP2ConnectionID(addr, m_bTraceFullIP);

        STRACE2_INFO( g_pH323Module->m_H323TraceId, this, ulConnectionID, "() IP: " << addr.AsString() << " = 0x" << std::hex << ntohl((DWORD)addr) << " TraceFullIP: " << m_bTraceFullIP);

        // hk: disable nagles algorithm in TCP socket in order to push data out as soon as write it on the socket;
        //     note: most documentation about TCP socket warn that it is in most cases not recommended to disable
        //           the nagle algorithm, but for signalling paths the flow control of TCP isn't relevant. So 
        //           nagles algorith.
        signallingChannel->SetOption( TCP_NODELAY, true, IPPROTO_TCP );

        signallingChannel->SetNonBlocking(true);
        
        STRACE2_INFO( g_pH323Module->m_H323TraceId, this, ulConnectionID, "() Add new socket and start transport: " << signallingChannel << " Peer Address/Port: " << addr << "/" << port );

        {
            SWaitAndSignal lock(m_TcpConnectionListMutex);
            m_TcpConnectionList.insert(signallingChannel);
        }

        if( bIsTls )
        {
            endpoint.OnSignallingChannelAccepted(addr,port);
        }
    }
}

Tools::STcpConnection * H323ListenerTls::GetTcpConnectionByRemote(const Tools::SIPSocket::Address & remoteAddress, DWORD remotePort)
{
    SDEFFUNC("H323ListenerTls::GetTcpConnectionByRemote");

    Tools::STcpConnection * pConnection = 0;

    SWaitAndSignal lock(m_TcpConnectionListMutex);

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0, 
        "() search socket for remote IP " << remoteAddress.AsString() << " port " << remotePort << " listsize=" << m_TcpConnectionList.size());

    SIPSocket::Address peerAddr;
    WORD peerPort = 0;
    ConnectionSet::iterator i;
    for (i = m_TcpConnectionList.begin(); i != m_TcpConnectionList.end(); i++)
    {
        pConnection = (*i);
        if (pConnection && pConnection->GetPeerAddress(peerAddr, peerPort))
        {
            STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0, 
                "() check for IP " << peerAddr.AsString() << " port " << peerPort);

            if (peerAddr==remoteAddress && peerPort==remotePort)
            {
                // found appropriate tcp/tls connection, return this one
                STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0, 
                    "() found tlscon " << pConnection << " for remote IP " << remoteAddress.AsString() << " port " << remotePort );

                return pConnection;
            }
        }
    }

    return 0;
}

Tools::STcpConnection * H323ListenerTls::GetTcpConnectionByLocal(const Tools::SIPSocket::Address& localAddress, DWORD localPort)
{
    SDEFFUNC("H323ListenerTls::GetTcpConnectionByLocal");

    Tools::STcpConnection * pConnection = 0;

    SWaitAndSignal lock(m_TcpConnectionListMutex);

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0, 
        "() search socket for local IP " << localAddress.AsString() << " port " << localPort << " listsize=" << m_TcpConnectionList.size());

    SIPSocket::Address interfAddr;
    WORD interfPort = 0;
    ConnectionSet::iterator i;
    for (i = m_TcpConnectionList.begin(); i != m_TcpConnectionList.end(); i++)
    {
        pConnection = (*i);
        if (pConnection && pConnection->GetLocalAddress(interfAddr, interfPort))
        {
            STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0, 
                "() check for IP " << interfAddr.AsString() << " port " << interfPort);

            if (interfAddr==localAddress && interfPort==localPort)
            {
                // found appropriate tcp/tls connection, return this one
                STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0, 
                    "() found tlscon " << pConnection << " for local IP " << localAddress.AsString() << " port " << localPort );

                return pConnection;
            }
        }
    }

    return 0;
}


bool H323ListenerTls::DeleteAllTcpConnections()
{
    SDEFFUNC("H323ListenerTls::DeleteAllTcpConnections");

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0,
      "() DeleteAllTcpConnections, size=" << m_TcpConnectionList.size());

    m_TcpConnectionListMutex.Wait();

    ConnectionSet::iterator tcpConnection = m_TcpConnectionList.begin();
    // we must re-initialize the iterator on each loop, otherwise we would be forced to not do release the mutex which protects the lists
    for( ; tcpConnection != m_TcpConnectionList.end(); tcpConnection = m_TcpConnectionList.begin()) 
    {
        STcpConnection* pTcpConnection = *tcpConnection;
        m_TcpConnectionList.erase( tcpConnection );

        m_TcpConnectionListMutex.Signal();

        STRACE2_INFO( g_pH323Module->m_H323TraceId, this, 0, "() Delete STCPConnection " << pTcpConnection );

        delete pTcpConnection;

        m_TcpConnectionListMutex.Wait();
    }
    // the loop does not stop if the list is not empty
    
    m_TcpConnectionListMutex.Signal();

    return true;
}


/////////////////////////////////////////////////////////////////////////////

H323TransportTls::H323TransportTls(H323EndPoint & end, bool bIsIncoming)
  : H323TransportTCP(end, bIsIncoming), startTls(false)
{
    SDEFFUNC("H323TransportTls::H323TransportTls");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "()");
}


H323TransportTls::H323TransportTls(H323EndPoint & end, WORD port)
  : H323TransportTCP(end, port), startTls(false)
{
    SDEFFUNC("H323TransportTls::H323TransportTls");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "()");    
}


H323TransportTls::~H323TransportTls()
{
    SDEFFUNC("H323TransportTls::~H323TransportTls");

    Close();

    delete h245listener;  // Delete and H245 listener that may be present

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "()");
}


bool H323TransportTls::Close()
{
    SDEFFUNC("H323TransportTls::Close");

    channelPointerMutex.StartWrite();

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "()");

    SIPSocket::Address peerAddr;
    WORD peerPort = 0;
    SH323TlsConnection * pTlsCon = dynamic_cast<SH323TlsConnection *>(GetReadChannel());

    Errors lastErr = SChannel::NoError;

    if (pTlsCon)
    {
        lastErr = pTlsCon->GetErrorCode(); // store lastErr, because GetPeerAddress will overwrite it!
        pTlsCon->GetPeerAddress(peerAddr, peerPort);
    }

    if (IsOpen() && pTlsCon 
                 && lastErr == SChannel::NoError)
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
              "() return tlscon for reuse, local=" << remoteAddress.AsString() << ":" << remotePort
                                    << ", remote=" << localAddress.AsString()  << ":" << localPort );
        
        pTlsCon->RemoveTransport(this);
    }
    else
    {
        // if something went wrong with the socket or we should not reuse it,
        // then don't use it any more - delete it
        if (pTlsCon)
        {
            STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0, "() shut down tlscon " << pTlsCon << " to " << peerAddr.AsString() << ":" << peerPort);

            // shut down tcp connection, the tcp listener will remove this connection from its lists
            pTlsCon->Shutdown(ShutdownReadAndWrite);
        }
    }

    // delete references in the base class so that TlsCon is not deleted after transport object dies
    readChannel = writeChannel = NULL;
    channelPointerMutex.EndWrite();


    return true;
}

bool H323TransportTls::OnOpen()
{
  SDEFFUNC("H323TransportTls::OnOpen");

  SIPSocket * socket = (SIPSocket *)GetReadChannel();

  // Get name of the remote computer for information purposes
  if (!socket->GetPeerAddress(remoteAddress, remotePort))
    return false;

  // get local address of incoming socket to ensure that multi-homed machines
  // use a NIC address that is guaranteed to be addressable to destination
  if (!socket->GetLocalAddress(localAddress, localPort))
    return false;
#ifdef SO_HARD_CLOSE
  // make sure that socket will be closed with RESET (hard close)
  const linger ling = { 1, 0 };
  return socket->SetOption(SO_LINGER, &ling, sizeof(ling));
#else

#ifdef __BEOS__
  return true;
#else
  // switch off nagle algorithm as signalling traffic hardly causes congestion issues.
  socket->SetOption(TCP_NODELAY, true, IPPROTO_TCP);

  // switch on TCP keep alive mechanism (Ticket 080702-26598)
  // set TCPKeepAliveTime to 0 via registry means disable mechanism, 
  // otherwise ms for keep alive timeout
  SConfig Config;
  u_long uKeepAliveTime = Config.GetInteger("TCPKeepAliveTime", 60000);
  bool ret = socket->SetKeepAlive((uKeepAliveTime > 0), uKeepAliveTime);

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
      "() Set TCP keep alive to " << uKeepAliveTime << "ms" << (ret ? " succeeded" : " failed"));


  // make sure do not lose outgoing packets on close
  const linger ling = { 1, 10 };
  return socket->SetOption(SO_LINGER, &ling, sizeof(ling));
#endif
#endif
}



bool H323TransportTls::Connect()
{
    SDEFFUNC("H323TransportTls::Connect");

    if( startTls )
    {
        STcpConnection* socket = endpoint.m_tlsConFactory.CreateNew(endpoint, g_pH323Module->m_H323TraceId);

        SIpAddress destAddr = remoteAddress;
        destAddr.m_Port = remotePort;

        channelPointerMutex.StartWrite();
        readChannel = writeChannel = socket;
        channelPointerMutex.EndWrite();

        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
            "() Starting connection to " << destAddr);

        channelPointerMutex.StartRead();
        if (!socket->Connect(destAddr) )
        {
            int errnum = socket->GetErrorNumber();
            STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
                "() Could not connect to " << destAddr << ": " << socket->GetErrorText() << '(' << errnum << ')');

            channelPointerMutex.EndRead();
            channelPointerMutex.StartWrite();
            readChannel = writeChannel = NULL;
            channelPointerMutex.EndWrite();
            delete socket;
            return SetErrorValues(socket->GetErrorCode(), errnum);
        }

        socket->SetOption(TCP_NODELAY, true, IPPROTO_TCP);

        // put socket to the listener in order to read and process messages from wire
        SH323TlsConnection* pTcpCon = dynamic_cast<SH323TlsConnection*>(socket);
        // dynamic_cast will never return 0 because we generated it

        STRACE2_INFO3( g_pH323Module->m_H323TraceId, this, 0, "() new TlsConnection created: " << pTcpCon );

        // readChannel->SetReadTimeout(SMaxTimeInterval); // Set timeout default
        channelPointerMutex.EndRead();
        channelPointerMutex.StartWrite();
        readChannel = writeChannel = NULL;
        channelPointerMutex.EndWrite();

        if (!Open(socket))
            return false;

        pTcpCon->SetTransport(this);  
        endpoint.PutSignallingChannelToListener( socket );

        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
              "() Started connection local=" << localAddress << ":" << localPort 
                               << " remote=" << remoteAddress << ":" << remotePort);

        return true;
    }

    SWaitAndSignal lock(*m_pRefMutex);

    bool isServer = endpoint.GetIsTlsServer();
    SChannel* pTcpCon = isServer ? endpoint.GetSignallingChannelForReuseByRemote(remoteAddress, remotePort)
                                 : endpoint.GetSignallingChannelForReuseByLocal(localAddress, localPort);

    SH323TlsConnection * socket = dynamic_cast<SH323TlsConnection *>(pTcpCon);

    if (!socket)
    {
        STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
              "() no TLS connection found for " << remoteAddress << ":" << remotePort);

        return false;
    }

    socket->SetTransport(this);  

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
          "() " <<socket<<"->SetTransport(" << this<< ")" );

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
          "() Reusing connection to " << remoteAddress << ":" << remotePort);

    // no autoDelete for Tls connections!
    if (!Open(socket, false))
        return false;

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, 0,
          "() Reused connection local=" << localAddress << ":" << localPort << ")"
                          << " remote=" << remoteAddress << ":" << remotePort);

    return true;
}
