/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: X.224 
 
    Comments: X.224 protocol handler (simplest class)

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/x224.cpp                   $
    $Author:: Martin.hueser                                                $
      $Date:: 20.12.05 18:34                                               $
   $Modtime:: 16.12.05 16:35                                               $
  $Revision:: 2                                                            $
  $Workfile:: x224.cpp                                                     $
$NoKeywords: $
----------------------------------------------------------------------------*/

#include <stlib.h>
#include "x224.h"

using namespace Tools;
using namespace std;

X224::X224()
{
}


void X224::PrintOn(ostream & strm) const
{
  long long indent = 2;
  strm << setprecision(indent) << "{\n"
       << setw(indent) << ' ' << "code=";
  switch (GetCode()) {
    case ConnectRequest :
      strm << "ConnectRequest";
      break;
    case ConnectConfirm :
      strm << "ConnectConfirm";
      break;
    case DataPDU :
      strm << "DataPDU";
  }

  char fillchar = strm.fill();

  strm << '\n'
       << setw(indent) << ' ' << "data: " << data.size() << " bytes\n"
       << hex;

  SByteVector::size_type i = 0;
  while (i < data.size()) {
    strm << setfill(' ') << setw(indent) << ' ' << setfill('0');
    SByteVector::size_type j;
    for (j = 0; j < 16; j++)
      if (i+j < data.size())
        strm << setw(2) << (unsigned)data[i+j] << ' ';
      else
        strm << "   ";
    strm << "  ";
    for (j = 0; j < 16; j++) {
      if (i+j < data.size()) {
        if (__stl_native_std::isprint<char>(data[i+j], locale::classic()))
          strm << data[i+j];
        else
          strm << ' ';
      }
    }
    strm << '\n';
    i += 16;
  }
  strm << dec << setfill(fillchar)
       << setw(indent-1) << '}'
       << setprecision(indent-2);
}


bool X224::Decode(const SByteVector & rawData)
{
  int packetLength = rawData.size();

  int headerLength = rawData[0];
  if (packetLength < headerLength + 1) // Not enough bytes
    return false;

  header.resize(headerLength);
  Tools::copy_n(rawData.begin()+1, headerLength, header.begin());

  packetLength -= headerLength + 1;
  header.resize(packetLength);
  if (packetLength > 0)
    Tools::copy_n(rawData.begin()+headerLength+1, packetLength, data.begin());

  return true;
}


bool X224::Encode(SByteVector & rawData) const
{
  int headerLength = header.size();
  int dataLength = data.size();
  rawData.resize(headerLength + dataLength +1);

  rawData[0] = (BYTE)headerLength;
  Tools::copy_n(header.begin(), headerLength, rawData.begin()+1);

  if (dataLength > 0)
    Tools::copy_n(data.begin(), dataLength, rawData.begin()+headerLength+1);

  return true;
}


void X224::BuildConnectRequest()
{
  data.resize(0);
  header.resize(6);
  header[0] = ConnectRequest;
  header[1] = 0;
  header[2] = 0x7b;
  header[3] = 2;
  header[4] = 0;
  header[5] = 0;
}


void X224::BuildConnectConfirm()
{
  data.resize(0);
  header.resize(6);
  header[0] = ConnectConfirm;
  header[1] = 0;
  header[2] = 0x7b;
  header[3] = 2;
  header[4] = 0;
  header[5] = 0;
}


void X224::BuildData(const SByteVector & d)
{
  header.resize(2);
  header[0] = DataPDU;
  header[1] = 0x80;
  data = d;
}



/////////////////////////////////////////////////////////////////////////////
